/*
 * Tue Jul  1 17:20:25 EDT 2008 <pavlos@cs.umd.edu>
 *
 */

#include "pcapstats.h"

struct global_state {
        char hostid[EST_INFO_IDLEN];

        int verbose;

        int use_estimator;
        struct est_estimator estimator;

        char send_trace_host[EST_INFO_IDLEN];

        char send_host[EST_INFO_IDLEN];
        char send_pcap_name[300];

        char recv_host[EST_INFO_IDLEN];
        char recv_pcap_name[300];

        FILE *burstfp;
        char burst_file[300];

        char filter[300];

} gb = {
        .verbose = 0,
};

struct bursts_info_stats {
        // the bytes follow transport header sizers UDP or MGRP
        int sent_pkts;
        int sent_bytes;

        int recv_pkts;
        int recv_bytes;

        int lost_pkts;
        int lost_bytes;

        int pbk_pkts;
        int pbk_bytes;

        int lost_min;
        int lost_max;
};

struct bursts_info {
        int valid;

        // Sequence number of the next probe, starts from 0
        int burst_seqno;        // seqno=-1 means we cannot assign seqno
        int probe_seqno;        // seqno within burst

        struct timeval start;
        struct timeval end;

        struct bursts_info_stats stats;
};

uint64_t
pcap_pkt_hash(void *data, void *data_key)
{
        return _pcap_pkt_hash(data, data_key);
}

int
pcap_pkt_cmp(void *lhs_data, void *rhs_data)
{
        return _pcap_pkt_cmp(lhs_data, rhs_data);
}

static inline void
burststats_fill_stats_common(struct est_info_stats *stats,
                            const char *tag,
                            const struct timeval *timestamp)
{
        const char *hostid = gb.send_trace_host;
        const char *src_host = gb.send_host;
        const char *dst_host = gb.recv_host;

        // Fill the header: key and timestamp.
        //
        // Infoplot will create a separate data file whose name has the format:
        //
        //      stats-bursts.<src>.<dst>.dat
        //
        // where type="bursts" and subtype=@tag
        //
        est_info_stats_init(stats, "bursts");
        est_fill_info_srcdst(&stats->hdr, src_host, dst_host);
        est_fill_info_timestamp(&stats->hdr, hostid, timestamp);

        snprintf(stats->tag, sizeof(stats->tag), tag);

}

static inline void
bursts_info_stats_update(struct bursts_info_stats *stats,
                         struct pcap_pkt *pkt,
                         struct pcap_pkt_flags *flags,
                         struct bursts_info *binf)
{
        // The probe size includes UDP or MGRP headers but not IP headers
        int probesize = pkt->transport_len;
        int probesize_pbk = flags->pbk_len;

        // update counters
        stats->sent_pkts++;
        stats->sent_bytes += probesize;

        if (pcap_pkt_lost(pkt)) {
                if (stats->lost_bytes == 0)
                        stats->lost_min = binf->probe_seqno;
                stats->lost_max = binf->probe_seqno;

                stats->lost_pkts++;
                stats->lost_bytes += probesize;
        } else {
                stats->recv_pkts++;
                stats->recv_bytes += probesize;
        }

        if (probesize_pbk > 0) {
                stats->pbk_pkts++;
                stats->pbk_bytes += probesize_pbk;
        }

}

static inline void
bursts_info_fill_stats(struct est_info_stats *stats,
                       struct bursts_info *binf)
{
        int avg_probe_size = 0;
        long long burst_usec = atimer_timeval_diff_usec(&binf->end,
                                                        &binf->start);
        
        if (binf->stats.sent_pkts > 0)
                avg_probe_size = binf->stats.sent_bytes / binf->stats.sent_pkts;

        // fill stats-info
        burststats_fill_stats_common(stats, "burst", &binf->start);

        est_info_stats_setint(stats, 0, binf->burst_seqno);
        est_info_stats_setint(stats, 1, burst_usec);

        // num of probes: we start from 0 and this would be the next probe seqno
        est_info_stats_setint(stats, 2, binf->probe_seqno);
        est_info_stats_setint(stats, 3, avg_probe_size);

        est_info_stats_setstr(stats, 4, "s");
        est_info_stats_setint(stats, 5, binf->stats.sent_pkts);
        est_info_stats_setint(stats, 6, binf->stats.sent_bytes);

        est_info_stats_setstr(stats, 7, "l");
        est_info_stats_setint(stats, 8, binf->stats.lost_pkts);
        est_info_stats_setint(stats, 9, binf->stats.lost_bytes);

        est_info_stats_setstr(stats, 10, "p");
        est_info_stats_setint(stats, 11, binf->stats.pbk_pkts);
        est_info_stats_setint(stats, 12, binf->stats.pbk_bytes);

        est_info_stats_setstr(stats, 13, "lminmax");
        est_info_stats_setint(stats, 14, binf->stats.lost_min);
        est_info_stats_setint(stats, 15, binf->stats.lost_max);
}


static inline void
burststats_fill_stats_probe(struct est_info_stats *stats,
                           struct pcap_pkt *pkt,
                           struct pcap_pkt_flags *flags,
                           struct bursts_info *binf)
{
        // The probe size includes UDP or MGRP headers but not IP headers
        int probesize = pkt->transport_len;
        int probesize_pbk = flags->pbk_len;

        char str[100];
        const char *fate = "unknown";

        if (pcap_pkt_outoforder(pkt))
                fate = "oord";
        else if (pcap_pkt_lost(pkt))
                fate = "lost";
        else
                fate = "recv";

        // fill stats-info
        pcap_pkt_oneliner(pkt, str, sizeof(str));
        burststats_fill_stats_common(stats, "probe", &pkt->send_pcap.ts);

        est_info_stats_setint(stats, 0, pkt->send_seqno);
        est_info_stats_settimeval(stats, 1, &binf->start);
        est_info_stats_setint(stats, 2, binf->burst_seqno);
        est_info_stats_setint(stats, 3, binf->probe_seqno);

        est_info_stats_setstr(stats, 4, fate);

        est_info_stats_setint(stats, 5, pkt->send_pcap.len);
        est_info_stats_setint(stats, 6, probesize);
        est_info_stats_setint(stats, 7, probesize_pbk);

        est_info_stats_setbigstr(stats, 8, str);
}

int
bursts_get_next_timestamps(struct timeval *start, struct timeval *end)
{
        int found = 0;
        char str[300];

        if (gb.burst_file[0] == 0)
                return -1;

        if (gb.burstfp == NULL) {
                gb.burstfp = fopen(gb.burst_file, "r");
                if (gb.burstfp == NULL)
                        ERR("error: failed to open burst file %s", gb.burst_file);
        }

        while (fgets(str, sizeof(str), gb.burstfp) != NULL) {
                found = sscanf(str, "%ld.%ld %ld.%ld",
                               &start->tv_sec, &start->tv_usec,
                               &end->tv_sec, &end->tv_usec);
                if (found == 4)
                        break;
        }

        if (found != 4)
                return -1;

        LOG("#");
/*         LOG(""); */
/*         LOG("burst timestams: %ld.%.6ld", end->tv_sec, end->tv_usec); */

        return 0;
}

void
load_match_burststats(struct pcap_stats *stats)
{
        int pkts = 0;
        int loaded = 0;
        int matched = 0;

        if (gb.verbose)
                LOG("Loading send packets..");

        while ((pkts = load_next_send_pkts(stats, PCAPSTATS_LOAD_STEP)) > 0) {
                loaded += pkts;
                if (gb.verbose)
                        LOG("    loaded  %10d total %10d", pkts, loaded);
        }

        if (gb.verbose)
                LOG("Matching recv packets..");

        matched = match_next_recv_pkts(stats);

        if (gb.verbose)
                LOG("    matched %10d total %10d", matched, matched);
}

static inline void
bursts_info_init(struct bursts_info *binf)
{
        int rc = 0;

        memset(binf, 0, sizeof(struct bursts_info));

        rc = bursts_get_next_timestamps(&binf->start, &binf->end);
        binf->valid = (rc == 0);

        if (!binf->valid)
                binf->burst_seqno = -1;
}

static inline int
bursts_info_needs_update(struct bursts_info *binf, struct timeval *pkt_tstamp)
{
        long long diff_usec = atimer_timeval_diff_usec(&binf->end, pkt_tstamp);
        // new packet is beyond the end of the current burst
        int is_new_burst = (diff_usec < 0);

        // cannot read from burst file
        if (!binf->valid)
                return 0;

        return is_new_burst;
}

static inline void
bursts_info_update(struct bursts_info *binf)
{
        int rc = 0;

        binf->burst_seqno++;
        binf->probe_seqno = 0;
        memset(&binf->stats, 0, sizeof(struct bursts_info_stats));

        rc = bursts_get_next_timestamps(&binf->start, &binf->end);
        if (rc) {
                binf->valid = 0;
                binf->burst_seqno = -1;
        }
}

static inline void
bursts_info_advance(struct bursts_info *binf)
{
        binf->probe_seqno++;
}


void
walk_compute_burststats(struct pcap_stats *stats)
{
        int rc = 0;
        anode *node = NULL;

        struct est_info info;
        struct est_info_group igroup;

        int epoch_count = 0;
        long epoch_sec = 0;

        struct bursts_info binf;

        est_info_group_init(&igroup, (gb.use_estimator ? &gb.estimator : NULL));
        bursts_info_init(&binf);

        // no need for special burst_start and burst_end initialization
        // will be initialized in the normal update inside the loop

        rc = alist_iterator_set(&stats->pktlist,
                                ALIST_FORWARD, ALIST_BEFORE_FIRST, NULL);
        BUG_ON(rc);

        while ((node = alist_iterator_next(&stats->pktlist)) != NULL) {
                struct pcap_pkt_flags flags;
                struct pcap_pkt *pkt = node->data;
                
                // first epoch
                if (epoch_count == 0) {
                        epoch_count = 1;
                        epoch_sec = pkt->send_pcap.ts.tv_sec;

                // epoch transition
                } else if (epoch_sec < pkt->send_pcap.ts.tv_sec) {
                        epoch_count++;
                        epoch_sec = pkt->send_pcap.ts.tv_sec;
                }

                pcap_pkt_detect_transports(&flags, pkt);

                // UDP means this is a probe
                if (flags.is_udp) {
                        if (bursts_info_needs_update(&binf, &pkt->send_pcap.ts)) {
                                bursts_info_fill_stats(&info.stats, &binf);
                                est_info_group_add(&igroup, &info, EST_INFO_PRINT);
                                bursts_info_update(&binf);
                        }

                        burststats_fill_stats_probe(&info.stats, pkt,
                                                    &flags, &binf);
                        est_info_group_add(&igroup, &info, EST_INFO_PRINT);

                        bursts_info_stats_update(&binf.stats, pkt, &flags, &binf);
                        bursts_info_advance(&binf);
                }

                if (gb.verbose >= PCAPSTATS_VERBOSE_PKTS)
                        pcap_pkt_log(pkt);
        }

        // update stats of last burst
        bursts_info_fill_stats(&info.stats, &binf);
        est_info_group_add(&igroup, &info, EST_INFO_PRINT);

        // we are done; make sure all remaining INFO events are sent to the estimator
        est_info_group_flush(&igroup);
}

void
compute_burststats(pcap_t *send, pcap_t *recv)
{
        int rc = 0;
        struct pcap_stats stats;

        memset(&stats, 0, sizeof(struct pcap_stats));

        stats.send = send;
        stats.recv = recv;

        rc = alist_init(&stats.pktlist, ALIST_NO_MUTEX, &free);
        BUG_ON(rc);

        // first level = 2^16, second level 2^4 elements
        rc = ahash_init(&stats.pkthash, 16, 2, pcap_pkt_cmp, pcap_pkt_hash);
        BUG_ON(rc);

        load_match_burststats(&stats);
        walk_compute_burststats(&stats);

        ahash_done(&stats.pkthash);
        alist_done(&stats.pktlist);
}

void usage();
void parse_cmdline(int argc, char **argv);

int
main(int argc, char **argv)
{

        pcap_t *send_pcap = NULL;
        pcap_t *recv_pcap = NULL;
        char err[NET_PCAP_ERRBUF_SIZE];

        est_get_host_id(gb.hostid, sizeof(gb.hostid));
	parse_cmdline(argc, argv);

        // estimator connection
        if (gb.use_estimator) {
                int rc = est_open_channel(&gb.estimator.ch, &gb.estimator.addr);
                if (rc) {
                        ERR("error: failed to open channel to estimator on %s (%s)",
                            gb.estimator.unparsed, gb.estimator.addrstr);
                }
        }

        // pcap files
        send_pcap = net_pcap_open_offline_filter(gb.send_pcap_name,
                                                 gb.filter,
                                                 err, sizeof(err));
        if (send_pcap == NULL)
                ERR("error: send trace: %s", err);

        recv_pcap = net_pcap_open_offline_filter(gb.recv_pcap_name,
                                                 gb.filter,
                                                 err, sizeof(err));
        if (recv_pcap == NULL)
                ERR("error: recv trace: %s", err);

        // main pcap processing
        compute_burststats(send_pcap, recv_pcap);

        // cleanup
        pcap_close(send_pcap);
        pcap_close(recv_pcap);
        if (gb.use_estimator)
                est_close_channel(&gb.estimator.ch);

        return 0;
}

static inline void
check_params_error(char *what_not_specified, char option)
{
        printf("error: %s not specified with the -%c option\n",
               what_not_specified, option);
        printf("error: use option '-h' for help\n\n");
        exit(2);
}

void
check_params(void)
{
        if (strlen(gb.send_host) == 0)
                check_params_error("send host", 'S');

        if (strlen(gb.send_host) == 0)
                check_params_error("send trace host", 'T');

        if (strlen(gb.send_pcap_name) == 0)
                check_params_error("send pcap file", 's');
        
        if (strlen(gb.recv_pcap_name) == 0)
                check_params_error("recv pcap file", 'r');
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hr:s:f:E:H:v:S:R:T:B:";
        int found = 0;
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

                case 'v':
                        found = sscanf(optarg, "%u", &gb.verbose);
                        BUG_ON(found != 1);
                        break;

                case 'B':
                        snprintf(gb.burst_file, sizeof(gb.burst_file), optarg);
                        break;

		case 'E':
                        if (gb.use_estimator)
                                ERR("error: too many -%c options", c);
                        gb.use_estimator = 1;
                        est_parse_estimator(optarg, &gb.estimator);
			break;

                case 'H':
                        strncpy(gb.hostid, optarg, sizeof(gb.hostid));
                        LOG_ERR("Overriding hostid: <%s> [actual arg <%s>]",
                            gb.hostid, optarg);
                        break;

                case 'S':
                        snprintf(gb.send_host, sizeof(gb.send_host), optarg);
                        break;

                case 'T':
                        snprintf(gb.send_trace_host, sizeof(gb.send_trace_host), optarg);
                        break;

                case 'R':
                        snprintf(gb.recv_host, sizeof(gb.recv_host), optarg);
                        break;

                case 's':
                        snprintf(gb.send_pcap_name, sizeof(gb.send_pcap_name),
                                 "%s", optarg);
                        break;

                case 'r':
                        snprintf(gb.recv_pcap_name, sizeof(gb.recv_pcap_name),
                                 "%s", optarg);
                        break;

                case 'f':
                        snprintf(gb.filter, sizeof(gb.filter), "%s", optarg);
                        break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
                        ERRFN("invalid options");
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("\n\nOBSOLETE --- UPDATE PLEASE! -- Mon Jun 16 11:53:23 EDT 2008\n");

	printf("\n");
	printf("    pcapstats -- extracts statistics from two tcpdump files\n");
	printf("\n");

	printf("USAGE\n");
	printf("\n");
	printf("    pcapstats -s <send-pcap> -r <recv-pcap> [options]\n");
	printf("\n");

	printf("OPTIONS\n");
	printf("        -s <send-pcap>  pcap file close to sender\n");
	printf("        -r <recv-pcap>  pcap file close to receiver\n");
	printf("        -f <filter-str> tcpdump-type filter string\n");
	printf("\n");
	printf("        -v <level>      verbosity level\n");
	printf("        -v4             one line per packet\n");
	printf("\n");

	exit(1);
}
