/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Fri Nov  2 15:47:52 EDT 2007
 *
 * Testing libsysstat
 *
 * Prints every second statistics about the system.
 * Equivalent to 'mpstat -P ALL 1'
 */

#include <stdio.h>

#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>

/* libcommon */
#include "alog.h"

#include "libsysstat.h"

struct global_state {
	int interval_sec;
        int use_double;
        int use_int;
} gb = {
        .interval_sec = 1,
        .use_double = 1,
        .use_int = 0,
};

void usage(void);
void parse_cmdline(int argc, char **argv);

#define MAX_CPU_NUM     4

void
print_stats_every(int sec)
{
        int curr = 0;

        struct mp_stats_ratio_double *dratio = NULL;
        struct mp_stats_ratio_int *iratio = NULL;
        int ratio_num = 0;
        char ratio_str[1000];
        
        mpstat_init();
        curr = mpstat_init_current();

        ratio_num = mpstat_get_cpu_num() + 1;
        dratio = malloc(sizeof(struct mp_stats_ratio_double) * ratio_num);
        iratio = malloc(sizeof(struct mp_stats_ratio_int) * ratio_num);

        BUG_ON(dratio == NULL);
        BUG_ON(iratio == NULL);

        while (1) {
                struct timeval now;
                // must be at least 26 bytes, since used by ctime_r(3)
                char now_str[50];

                sleep(sec);

                gettimeofday(&now, NULL);
                ctime_r(&now.tv_sec, now_str);

                mpstat_read_stats(curr);
                mpstat_compute_interval_stats(curr, dratio, ratio_num);
                curr = mpstat_advance_current(curr);

                if (gb.use_double) {
                        mpstat_ratio_double_str(ratio_str, sizeof(ratio_str),
                                                MPSTAT_WITH_HEADER,
                                                dratio, ratio_num);
                        LOGNR(">>> double %s", now_str);
                        LOG("%s", ratio_str);
                }
                
                if (gb.use_int) {
                        mpstat_downgrade_array_ratio_double_to_int(
                                                dratio, iratio, ratio_num);
                        mpstat_ratio_int_str(ratio_str, sizeof(ratio_str),
                                             MPSTAT_WITH_HEADER,
                                             iratio, ratio_num);
                        LOGNR(">>> int %s", now_str);
                        LOG("%s", ratio_str);
                }
        }

        free(dratio);
        free(iratio);
}

int
main (int argc, char **argv)
{
	parse_cmdline(argc, argv);
        print_stats_every(gb.interval_sec);

	return 0;
}

void
check_params(void)
{
}

void
parse_cmdline(int argc, char **argv)
{
	char short_options[] = "hi:DI";

	int found = 0;
	char c = '?';

	/* parse command line options */

	while (1) {

		c = getopt(argc, argv, short_options);

		if (c == EOF)
		    break;

		switch (c) {

		case 'i':
                        found = sscanf(optarg, "%u", &gb.interval_sec);
                        if (found != 1)
                                ERR("invalid interval: '%s'", optarg);
			break;

                case 'D':
                        gb.use_double = 1;
                        break;

                case 'I':
                        gb.use_int = 1;
                        break;

		case 'h':
			usage();
			exit(0);
			break;
		    
		case '?':
			break;
		    
		default:
			break;
		}
	}

	check_params();
}

void
usage(void)
{
	printf("USAGE\n");
	printf("\tmpstat1 [-i <interval-sec>]\n");
	printf("\n");
}

