#ifndef __EST_ESTIMATOR_H
#define __EST_ESTIMATOR_H

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>

#include "alist.h"
#include "aptr.h"
#include "trafgen.h"
#include "cpuload.h"
#include "libestimator.h"

#define EST_MAX_TRAFGEN_SENDERS		10
#define EST_INFO_MAX_ENTRIES            5000

#define EST_LOCAL_TRAFFIC_PORT(__estimator_control_port) \
	((__estimator_control_port) + 1)

#define EST_DEFAULT_ID_IFACE            "eth0"

struct estimator_global_state {

        char estimator_id[EST_INFO_IDLEN];
        char estimator_id_iface[50];
	int estimator_control_port;

	int keep_serving;

	int listen_sock;
	int fdmax;
	fd_set fdreads;

	alist channels;

	int quiet;
	int quiet_pathload;
        int timeout_pathload_rcv_sec;

#ifdef PATHLOAD_TESTING
	int sender;
	int receiver;
	int threads;
	int measure;
#endif
	pthread_mutex_t receiver_mutex_door1;
	pthread_mutex_t receiver_mutex_door2;

	/* data for traffic sender threads (one per src,dst pair) */
	pthread_mutex_t trafgen_sender_mutex;
	struct trafgen_sender snd_trafgen[EST_MAX_TRAFGEN_SENDERS];

	/* data for traffic receiver threads (one per type: NORMAL or MGRP) */
	int local_traffic_port;

	int start_rcv_trafgen_normal;
	struct trafgen_receiver rcv_trafgen_normal;

	int start_rcv_trafgen_mgrp;
	struct trafgen_receiver rcv_trafgen_mgrp;

        /* in seconds: if zero, trafgen_info thread is never launched */
        int trafgen_report_every_period;
        /* data for traffic info thread (one for all types of traffic) */
        struct trafgen_info rcv_trafgen_info;

        /* in seconds: if zero, cpuload thread is never launched */
        int cpuload_report_every_period;
        struct cpuload cpuload_data;

        /* info entries */
        alist info_list;
        int info_list_max_entries;
        pthread_mutex_t info_list_mutex;

        /* pointer storage */
        struct aptr_storage pointers;
};

// thread-specific data for threads that process @est_cmd commands
struct est_process_data {
        struct aptr ptr;
        // XXX is this guaranteed to be always valid?
        struct est_channel *ch;
        union est_cmd cmd;
};

static inline struct est_process_data *
est_process_data_alloc_init(struct est_channel *ch, union est_cmd *cmd)
{
        struct est_process_data *pdata = malloc(sizeof(struct est_process_data));

        BUG_ON(pdata == NULL);
        memset(pdata, 0, sizeof(struct est_process_data));

        // the reference counter of @ptr is tied to the memory of @pdata
        aptr_init(&pdata->ptr, pdata);
        pdata->ch = ch;
        // this is a copy of the command
        pdata->cmd = *cmd;

        return pdata;
}

extern struct estimator_global_state gb;

// Returns the address to communicate with the estimator on the localhost
static inline void
estimator_get_loopback_addr(struct sockaddr_in *addr)
{
        memset(addr, 0, sizeof(struct sockaddr_in));
        addr->sin_family = AF_INET;
        addr->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
        addr->sin_port = htons(gb.estimator_control_port);
}
#endif
