/* Pavlos Papageorgiou <pavlos@eng.umd.edu> */
/* Mon Oct 29 13:29:28 EDT 2007 */

#include "info.h"

void
est_process_info_get(struct est_channel *ch, struct est_info_cmd *cmd)
{
        int exhausted = 0;
        int pulled = 0;
        int is_final = 0;

        while (!is_final) {
                int rc = 0;
                int ok = 0;
                struct est_info_reply reply;
                char error[100];
                int popped = 0;
                int remaining = cmd->info_max_entries_to_get - pulled;
                int max_entries = cmp_min(remaining,
                                          EST_INFO_ENTRIES_PER_MSG);
                                          
                memset(&reply, 0, sizeof(reply));

                reply.action = cmd->action;
                popped = info_pop_greedy(reply.info, max_entries);
                reply.info_entries = popped;

                pulled += reply.info_entries;
                exhausted = popped < max_entries;
                is_final = exhausted || pulled == cmd->info_max_entries_to_get;

                ok = est_fill_reply_header(&reply.hdr, EST_TYPE_INFO, is_final);
                BUG_ON(!ok);

                rc = est_send_ctrlmsg(ch, &reply, error, sizeof(error));
                if (rc) {
                        LOGTR("failed to send INFO reply: %s", error);
                }
        }
}

void
est_process_info_put(struct est_channel *ch,
                     struct est_info_cmd *cmd)
{
        int ok = 0;
        int rc = 0;
        int i = 0;
        char error[100];
        struct est_info_reply reply;

        for (i = 0; i < cmd->info_entries; i++) {
                info_alloc_copy_push(&cmd->info[i]);
        }

        /* final reply */
        memset(&reply, 0, sizeof(reply));
        ok = est_fill_reply_header(&reply.hdr, EST_TYPE_INFO, 1);
        BUG_ON(!ok);

        reply.action = cmd->action;

	rc = est_send_ctrlmsg(ch, &reply, error, sizeof(error));
	if (rc) {
		LOGTR("failed to send final INFO reply: %s", error);
	}
}

void
est_process_info(struct est_channel *ch, struct est_info_cmd *cmd)
{
        if (cmd->action == EST_INFO_GET)
                est_process_info_get(ch, cmd);
        else
                est_process_info_put(ch, cmd);
}
