/* Pavlos Papageorgiou <pavlos@eng.umd.edu> */
/* Mon Oct 29 13:29:28 EDT 2007 */

#ifndef __EST_INFO_H
#define __EST_INFO_H

#include "alist.h"

#include "libestimator.h"
#include "estimator.h"

void est_process_info(struct est_channel *ch, struct est_info_cmd *cmd);

/* caller is responsible for freeing non-NULL pointers */
static inline struct est_info *
info_pop()
{
        struct est_info *info = NULL;

        pthread_mutex_lock(&gb.info_list_mutex);
        info = alist_remove_first_r(&gb.info_list);
        pthread_mutex_unlock(&gb.info_list_mutex);
        return info;
}

static inline int
info_pop_greedy(struct est_info *info_array, int info_max)
{
        int popped = 0;

        pthread_mutex_lock(&gb.info_list_mutex);

        while (popped < info_max) {
                struct est_info *info = alist_remove_first_r(&gb.info_list);
                if (info == NULL)
                        break;
                /* copy and free original pointer */
                info_array[popped] = *info;
                free(info);
                popped++;
        }

        pthread_mutex_unlock(&gb.info_list_mutex);

        return popped;
}

/* must be called with info_list_mutex held */
static inline void
_info_pop_discard_overflow()
{
        while (alist_elements_r(&gb.info_list) >= gb.info_list_max_entries) {
                struct est_info *info = alist_remove_first_r(&gb.info_list);

                BUG_ON(info == NULL);
#ifdef DEBUG
                LOG(">>> discarding INFO entry; reached %d entries limit",
                    gb.info_list_max_entries);
                est_info_log(info);
#endif
                free(info);
        }
}

/* list takes ownership of pointer on success; mutex held by list code */
static inline int
info_push(struct est_info *info)
{
        int rc = 0;
        
	if (!info->hdr.timestamp_valid) {
		gettimeofday(&info->hdr.timestamp, NULL);
                strncpy(info->hdr.timestamp_id, gb.estimator_id,
                        sizeof(info->hdr.timestamp_id));
                info->hdr.timestamp_valid = 1;
	}

        if (info->hdr.key.node_id[0] == 0) {
                strncpy(info->hdr.key.node_id, gb.estimator_id,
                        sizeof(info->hdr.key.node_id));
        }

        pthread_mutex_lock(&gb.info_list_mutex);

        _info_pop_discard_overflow();

        rc = alist_insert_last_r(&gb.info_list, info, ALIST_OWNER);
        if (rc) return rc;
        
        pthread_mutex_unlock(&gb.info_list_mutex);
        return 0;
}

static inline void
info_alloc_copy_push(struct est_info *info)
{
        int rc = 0;
        struct est_info *info_alloc = NULL;

        info_alloc = malloc(sizeof(struct est_info));
        BUG_ON(info_alloc == NULL);

        *info_alloc = *info;

        rc = info_push(info_alloc);
        BUG_ON(rc);
}

/* push info/results into global queue */

static inline void
info_push_availbw_results(struct est_results *results,
                          const char *src_id,
                          const char *dst_id,
                          enum est_info_header_type availbw_type)
                                    
{
	struct est_info info;

	est_info_availbw_init(&info.availbw, availbw_type);

        /* need to fill header with nodeid, src, dst */
        strncpy(info.hdr.key.src_id, src_id, sizeof(info.hdr.key.src_id));
        strncpy(info.hdr.key.dst_id, dst_id, sizeof(info.hdr.key.dst_id));
        /* @hdr.key.type filled by _init (above) */
        /* @hdr.key.node_id filled by _copy_push (below) */

        info.availbw.results = *results;
	info_alloc_copy_push(&info);
}
#endif
