#include "aptr.h"
#include "alog.h"
#include "atimer.h"

#include "estimator.h"
#include "probe.h"
#include "pathload.h"

#include "libestimator.h"

void est_process_probestatus(struct est_channel *ch,
                             struct est_probestatus_cmd *cmd)
{
        // find the status pointer that was created and registered
        // by the waiting thread

        struct est_probe_status *status = NULL;
        
        status = aptr_storage_lookup(&gb.pointers, cmd->status_id);
        if (status == NULL) {
                LOGTR("error: failed to lookup status pointer with id %d",
                      cmd->status_id);
                return;
        }

        LOG("sending cond signal to thread waiting for status with id %d",
            cmd->status_id);
                
        pthread_mutex_lock(&status->mutex);
        status->value = cmd->status_value;
        status->results = cmd->status_results;
        pthread_cond_signal(&status->cond);
        pthread_mutex_unlock(&status->mutex);

}

void
est_wait_for_probestatus(struct est_probe_status *status, int timeout_msec)
{
        struct timeval now;
        struct timeval offs = {
                .tv_sec = timeout_msec / 1000,
                .tv_usec = timeout_msec % 1000,
        };
        struct timeval deadline;
        struct timespec deadline_tm;

        // cond_wait takes timeout in absolute time not relative
        gettimeofday(&now, NULL);
        atimer_timeval_add_offset(&deadline, &now, &offs);
        atimer_timeval_to_timespec(&deadline_tm, &deadline);

	pthread_mutex_lock(&status->mutex);
	while (status->value == EST_PROBE_STATUS_UNDEF) {
                int rc = 0;

                LOGTS("++++ waiting for cond_status, timeout at %ld.%.9ld sec",
                      deadline_tm.tv_sec,
                      deadline_tm.tv_nsec);

		if (timeout_msec > 0)
			rc = pthread_cond_timedwait(&status->cond,
					&status->mutex,
					&deadline_tm);
		else
			rc = pthread_cond_wait(&status->cond,
					&status->mutex);

                LOGTS("++++ finished waiting for cond_status, rc %d", rc);

		if (rc == ETIMEDOUT) {
			status->value = EST_PROBE_STATUS_TIMEOUT;
			break;
		}
	}
	pthread_mutex_unlock(&status->mutex);

        LOGFN("status %d %s", status->value,
              est_probe_status_str(status->value));
}


// this is a thread
void *
est_process_peerstart(void *arg)
{
        // the caller has already bumped up the refcount of @pdata
        // we need to release it when we are done
        struct est_process_data *pdata = arg;
        struct est_peerstart_cmd *cmd = &pdata->cmd.peerstart;

        if (cmd->type == EST_PEERTYPE_PATHLOAD)
                pathload_process_peerstart(pdata->ch, cmd, gb.quiet_pathload);
        else
                BUG();

        aptr_release(&pdata->ptr);
        return NULL;
}
