#ifndef __EST_TRAFGEN_H
#define __EST_TRAFGEN_H

#include "alog.h"
#include "libestimator.h"

struct trafgen_payload {
	char sender_id[EST_INFO_IDLEN];
};

struct trafgen_sender {
	int is_valid;
	pthread_t self;

	/* input */
	int use_mgrp;
	int remote_traffic_port;
	/* also key in snd_trafgen array */
	/* the ip/port pair refers to the estimator process; not the traffic */
	struct sockaddr_in src;
	struct sockaddr_in dst;

	/* operational */
	int sock;

	volatile int traffic_active;
	struct est_trafgen_traffic traffic;
	int sleep_usec;

	pthread_mutex_t traffic_mutex;
	pthread_cond_t traffic_cond;

	/* non-zero if it has exited with an error */
	int status;
};

struct trafgen_receiver {
	int is_valid;
	pthread_t self;

	/* input */
        enum est_trafgen_traffic_type type;
	int local_traffic_port;

	/* operational */
	/* non-zero if it has exited with an error */
	int status;
	int sock;
	struct sockaddr_in listen_addr;

        /* statistics */
        int total_packets;
        int total_bytes;
        alist stats_list;
        pthread_mutex_t stats_mutex;
};

#define EST_TRAFGEN_INFO_MAX_RECV       2
#define EST_TRAFGEN_STATS_MAX           10

struct trafgen_info {
	int is_valid;
	pthread_t self;

        /* input */

       /* where to extract the statistics from */
       int recv_count;
       struct trafgen_receiver *recv[EST_TRAFGEN_INFO_MAX_RECV];

       /* period in seconds */
       int report_every_period;
};

static inline void
est_trafgen_sender_init(struct trafgen_sender *trafgen)
{
	memset(trafgen, 0, sizeof(struct trafgen_sender));

	pthread_mutex_init(&trafgen->traffic_mutex, NULL);
	pthread_cond_init(&trafgen->traffic_cond, NULL);

	trafgen->is_valid = 1;
}

static inline void
est_trafgen_receiver_init(struct trafgen_receiver *trafgen)
{
        int rc = 0;

	memset(trafgen, 0, sizeof(struct trafgen_receiver));
	pthread_mutex_init(&trafgen->stats_mutex, NULL);
        rc = alist_init(&trafgen->stats_list, ALIST_NO_MUTEX, &free);
        BUG_ON(rc);

	trafgen->is_valid = 1;
}

static inline void
est_trafgen_info_init(struct trafgen_info *info)
{
	memset(info, 0, sizeof(struct trafgen_info));

	info->is_valid = 1;
}

/* TRAFGEN server side (estimator) */

void est_trafgen_process_traffic(struct est_channel *ch,
			         struct est_trafgen_traffic_cmd *cmd);

#ifdef TRAFGEN_REPORT
void est_trafgen_process_report(struct est_channel *ch,
			        struct est_trafgen_report_cmd *cmd);
#endif

/* helpers */

int est_trafgen_find_empty_index(struct trafgen_sender *trafgen, int count);
int est_trafgen_find_index(struct trafgen_sender *trafgen, int count,
			   struct sockaddr_in *src,
			   struct sockaddr_in *dst);

/* trafgen receiver thread (estimator) */
void *est_trafgen_receiver_thread(void *arg);

/* trafgen info thread (estimator) */
void *est_trafgen_info_thread(void *arg);

#endif
