/*
 * set.h	Simple bit set functions
 *
 * Created:	Tue Feb 28 10:32:06 EST 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */


#ifndef _BITSET_H
#define _BITSET_H

#include "err.h"

#define BITSET_ONE	0x80

struct bitset {
	unsigned char *octets;
	int bits;
};

static inline void
bitset_init(struct bitset *set, int bitsize)
{
	int size = 0;

	memset(set, 0, sizeof(*set));

	size = bitsize / 8;
	if (bitsize % 8)
		size += 1;

	set->octets = malloc(size);

	SYSERR_ON(set->octets == NULL);

	set->bits = size * 8;
}


static inline void
bitset_done(struct bitset *set)
{
	free(set->octets);
	memset(set, 0, sizeof(*set));
}


static inline int
bitset_contains(struct bitset *set, unsigned int val)
{
	int octet = val / 8;
	int shift = val % 8;

	BUG_ON(val >= set->bits);
	
	return set->octets[octet] & (BITSET_ONE >> shift);
}


static inline int
bitset_insert(struct bitset *set, unsigned int val)
{
	int octet = val / 8;
	int shift = val % 8;
	int was_set = 0;

	BUG_ON(val >= set->bits);

	was_set = set->octets[octet] & (BITSET_ONE >> shift);
	set->octets[octet] |= (BITSET_ONE >> shift);

	return was_set;
}


static inline int
bitset_remove(struct bitset *set, unsigned int val)
{
	int octet = val / 8;
	int shift = val % 8;
	int was_set = 0;

	BUG_ON(val >= set->bits);

	was_set = set->octets[octet] & (BITSET_ONE >> shift);
	if (was_set)
		set->octets[octet] ^= (BITSET_ONE >> shift);

	return was_set;
}

static inline int
bitset_toggle(struct bitset *set, unsigned int val)
{
	int octet = val / 8;
	int shift = val % 8;
	int was_set = 0;

	BUG_ON(val >= set->bits);

	was_set = set->octets[octet] & (BITSET_ONE >> shift);
	set->octets[octet] ^= (BITSET_ONE >> shift);

	return was_set;
}

#endif /* _BITSET_H */

