/*
 * err.h	Auxiliary utilities for error handling.
 *
 * Created:	Fri Dec  9 10:00:10 EST 2005
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _ERR_H
#define _ERR_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

static inline void
__syserr(int cond, int sys_errno, const char *file, int line)
{
	if (!cond)
		return;

	fprintf(stderr,
		"%s:%d: syscall errno %d: %s\n",
		file, line, sys_errno,
		strerror(sys_errno));
	exit(1);
}

static inline void
__abort(int cond, const char *abortstr, const char *file, int line)
{
	if (!cond)
		return;

	fprintf(stderr, "%s:%d: %s\n", file, line,
		(abortstr ? abortstr : "aborting.."));
	exit(1);
}


static inline void
__abort_f(int cond, const char *format, ...)
{
	va_list ap;

	if (!cond)
		return;

	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);

	exit(1);
}

#define SYSERR() \
	__syserr(1, errno, __FILE__, __LINE__)

#define SYSERR_ON(cond) \
	__syserr((cond), errno, __FILE__, __LINE__)

#define BUG() \
	__abort(1, "bug", __FILE__, __LINE__);

#define BUG_ON(cond) \
	__abort((cond), "bug_on(" #cond ")", __FILE__, __LINE__);

#define ERR(format, ...) \
	__abort_f(1, "%s:%d " format "\n", __FILE__, __LINE__, ##__VA_ARGS__)

#define ERR_ON(cond, format, ...) \
	__abort_f((cond), "%s:%d " format "\n", __FILE__, __LINE__, ##__VA_ARGS__)

#define LOG(format, ...) \
        printf(format "\n", ## __VA_ARGS__)

#define LOG_TRACE(format, ...) \
        printf("%s:%d " format "\n", __FILE__, __LINE__, ##__VA_ARGS__)

/* #define DEBUG(format, ...)      ((void)(0)) */

#define DEBUG(format, ...) \
	LOG_TRACE(format, ## __VA_ARGS__)


#endif
