/*
 * Created on   : Tue Sep 28 19:12:40 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * ahash.h
 *
 * Abstract hash
 *
 * $Id: ahash.h,v 1.7 2005/01/04 18:25:55 pavlos Exp $
 */

#ifndef __AHASH_H
#define __AHASH_H

#define AHASH_OK                 0
#define AHASH_ERROR             -1
#define AHASH_EEXISTS           -2
#define AHASH_EMISSING          -3

#include <unistd.h>

#include "adar.h"
#include "alist.h"
#include "cmp.h"

typedef struct ahash_node {
    adarray elements;
    alist overflow;
} ahash_node;

typedef struct ahash {

    uint level1_p2elem;
    uint64_t level1_mask;

    uint level2_p2elem;
    uint64_t level2_mask;

    cmp_search_func *search_func;
    cmp_hash64_func *hash_func;

    adarray level1;

} ahash;

typedef struct ahash_pos {
    ahash *hash;
    uint level1_index;
    uint level2_index;
    int found;
    int overflowed;
    anode *overflow_node;
} ahash_pos;


/*
 * Thomas Wang's 64 bit Mix Function
 * Source: http://www.concentric.net/~Ttwang/tech/inthash.htm
 */

static inline uint64_t
ahash_mix64(uint64_t key)
{
    key += ~(key << 32);
    key ^=  (key >> 22);
    key += ~(key << 13);
    key ^=  (key >> 8);
    key +=  (key << 3);
    key ^=  (key >> 15);
    key += ~(key << 27);
    key ^=  (key >> 31);

    return key;
}

#define ahash_level1_index(hash, hkey) \
    ( ((hkey) & (hash)->level1_mask) >> (hash)->level2_p2elem )

#define ahash_level2_index(hash, hkey) \
    ( hkey & hash->level2_mask )

int ahash_init(ahash *hash,
        uint level1_p2elem, uint leve2_p2elem,
        cmp_search_func *search_func, cmp_hash64_func *hash_func);
int ahash_done(ahash *hash);

#define ahash_lookup(hash, key) \
        ahash_lookup_pos((hash), (key), NULL);

void *ahash_lookup_pos(ahash *hash, void *key, ahash_pos *pos);

#define ahash_insert(hash, data, key) \
        ahash_insert_pos((hash), (data), (key), NULL);

int ahash_insert_pos(ahash *hash, void *data, void *key, ahash_pos *pos);

#define ahash_delete(hash, key) \
        ahash_delete_pos((hash), (key), NULL);

int ahash_delete_pos(ahash *hash, void *key, ahash_pos *pos);

void ahash_log(int level, ahash *hash, cmp_log_func *log_func);

int ahash_node_init(ahash_node *node, int nelem);
int ahash_node_done(ahash_node *node);
void ahash_node_free(void *buf);

uint64_t ahash_str64(unsigned char *str);


#endif /* __ADAR_H */

