/*
 * Created on   : Wed Jun 23 10:47:48 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * anode.h
 *
 * Abstract node to be used by abstract data structures.
 *
 * $Id: anode.c,v 1.9 2005/06/17 23:51:29 pavlos Exp $
 */

#include "anode.h"
#include "logerr.h"
#include "mem.h"

#include "rbuf.h"

anode *
anode_create(void *data, int data_owner, rbuf_mgr *rbmgr)
{
    anode *node = NULL;

    if (rbmgr != NULL) {
        
        rbuf *rb = NULL;

        /* request memory from rbuf manager for node */

        rb = rbuf_mgr_alloc(rbmgr, sizeof(anode));
        logerr_bug(rb != NULL);

        node = rb->ptr;
        node->rbowner = rb;

    } else {

        /* allocate memory for node */

        node = malloc(sizeof(anode));
        logerr_return_sys(node != NULL, "Failed to allocate new node", NULL);

        node->rbowner = NULL;
    }

    node->prev = NULL;
    node->next = NULL;
    node->data = data;
    node->data_owner = data_owner;

    return node;
}

int
anode_destroy(anode *node, bytes_free_func *free_func)
{
    int res = 0;

    logerr_passert(node);
    
    res = anode_destroy_data(node, free_func);
    if (res != ANODE_OK)
        return res;

    if (node->rbowner != NULL)
        rbuf_release(node->rbowner);
    else
        free(node);

    return ANODE_OK;
}

int
anode_destroy_data(anode *node, bytes_free_func *free_func)
{
    logerr_passert(node);
    logerr_assert_str(node->data != NULL || !node->data_owner,
                "Cannot be owner of NULL data");
    
    if (node->data_owner) {
        logerr_return_str(free_func != NULL,
                        "Failed to free data: No free function",
                        ANODE_ERROR);
        
        free_func(node->data);
    }

    node->data = NULL;
    node->data_owner = 0;

    return ANODE_OK;
}


int
anode_assign_data(anode *node, void *data, int data_owner)
{
    logerr_passert(node);
    logerr_assert_str(data != NULL || !data_owner,
                "Cannot be owner of NULL data");
    
    node->data = data;
    node->data_owner = data_owner;

    return ANODE_OK;
}


