/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Tue Mar  4 17:01:56 EST 2008
 * 
 * apthread.h
 *
 * Pthread related helper functions
 */

#include <pthread.h>

struct apthread_barrier {
        int count;
        pthread_mutex_t mutex;
        pthread_cond_t cond;
};

static inline void
apthread_barrier_init(struct apthread_barrier *barrier)
{
        memset(barrier, 0, sizeof(*barrier));
        pthread_mutex_init(&barrier->mutex, NULL);
        pthread_cond_init(&barrier->cond, NULL);
}

static inline void
apthread_barrier_engage(struct apthread_barrier *barrier)
{
        pthread_mutex_lock(&barrier->mutex);
        BUG_ON(barrier->count < 0);
        barrier->count += 1;
        pthread_mutex_unlock(&barrier->mutex);
}

static inline void
apthread_barrier_disengage(struct apthread_barrier *barrier)
{
        pthread_mutex_lock(&barrier->mutex);
        BUG_ON(barrier->count <= 0);
        barrier->count -= 1;
        pthread_cond_signal(&barrier->cond);
        pthread_mutex_unlock(&barrier->mutex);

}

static inline void
apthread_barrier_disengage_and_wait(struct apthread_barrier *barrier)
{
        pthread_mutex_lock(&barrier->mutex);
        BUG_ON(barrier->count <= 0);
        barrier->count -= 1;
        pthread_cond_signal(&barrier->cond);
        while (barrier->count > 0) {
                pthread_cond_wait(&barrier->cond, &barrier->mutex);
        }
        pthread_mutex_unlock(&barrier->mutex);
}
