/*
 * aptr -- a hash-based storage of pointers (thread-safe)
 *
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Fri Dec  7 11:41:46 MST 2007
 *
 */

#include "aptr.h"
#include "alog.h"

/* see requirements of cmp_search_func() */
int
aptr_storage_search(void *data, void *key)
{
        int lhs = aptr_ptrkey(data);
        int rhs = *(int *)key;

        if (lhs == rhs)
                return 0;

        return (lhs < rhs ? -1 : 1);
}

/* see requirements of cmp_hash64_func() */
uint64_t
aptr_storage_hash64(void *data, void *key)
{
        /* the key is a non-negative value */
        int key_val = -1;

        /* exactly one must be non-null */
        BUG_ON(data != NULL && key != NULL);

        if (data != NULL) {
                key_val = aptr_ptrkey(data);

        } else if (key != NULL) {
                key_val = *(int *)key;
        }

        BUG_ON(key_val == -1);
        return ahash_mix64(key_val);
}

int
aptr_storage_init(struct aptr_storage *storage)
{
        int rc = 0;

        memset(storage, 0, sizeof(struct aptr_storage));

        /* first level = 2^8, second level 2^3 elements */
        rc = ahash_init(&storage->hash, 8, 3,
                        &aptr_storage_search,
                        &aptr_storage_hash64);
        if (rc)
                return rc;

        pthread_mutex_init(&storage->mutex, NULL);

        return 0;
}

void
aptr_storage_done(struct aptr_storage *storage)
{
        ahash_done(&storage->hash);
}

