/*
 * Pavlos Papageorgiou <pavlos@cs.umd.edu>
 * Fri Nov 16 11:48:10 EST 2007
 * 
 * asys.h
 *
 * System related functions.
 */

#include "asys.h"

int
asys_status_wait_for_change(struct asys_status *status,
                            int timeout_msec,
                            int current_status)
{
        int retval = 0;
        struct timeval now;
        struct timeval offs = {
                .tv_sec = timeout_msec / 1000,
                .tv_usec = timeout_msec % 1000,
        };
        struct timeval deadline;
        struct timespec deadline_tm;

        // cond_wait takes timeout in absolute time not relative
        gettimeofday(&now, NULL);
        atimer_timeval_add_offset(&deadline, &now, &offs);
        atimer_timeval_to_timespec(&deadline_tm, &deadline);

	pthread_mutex_lock(&status->mutex);
	while (status->value == current_status) {
                int rc = 0;

                LOGTS("++++ waiting for cond_status, timeout at %ld.%.9ld sec",
                      deadline_tm.tv_sec,
                      deadline_tm.tv_nsec);

		if (timeout_msec > 0)
			rc = pthread_cond_timedwait(&status->cond,
					&status->mutex,
					&deadline_tm);
		else
			rc = pthread_cond_wait(&status->cond,
					&status->mutex);

                LOGTS("++++ finished waiting for cond_status, rc %d", rc);

                if (rc) {
                        BUG_ON(rc < 0);
                        retval = -rc;
                        break;
                }
	}
	pthread_mutex_unlock(&status->mutex);

        return retval;
}
