/*
 * Created on   : Sat Dec  4 17:53:07 EST 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * bqueue.h
 *
 * Byte Queue
 *
 * $Id: bqueue.h,v 1.3 2005/04/14 02:10:32 pavlos Exp $
 */

#ifndef __BQUEUE_H
#define __BQUEUE_H

#include <sys/uio.h> 

#define BQUEUE_OK               0
#define BQUEUE_ERROR           -1
#define BQUEUE_ENOSPACE        -2
#define BQUEUE_EEMPTY          -3
#define BQUEUE_ENOTHING        -4

#define BQUEUE_FRONT            1
#define BQUEUE_BACK             2

typedef struct bqueue {

    void *buf;
    int bufsize;

    int bytes;
    void *first;
    void *last;

} bqueue;

int bqueue_init(bqueue *bq, int bufsize);
int bqueue_done(bqueue *bq);

int bqueue_flush(bqueue *bq);

#define bqueue_isempty(bq)      ((bq)->bytes == 0)
#define bqueue_occupied(bq)     ((bq)->bytes)
#define bqueue_free(bq)         ((bq)->bufsize - (bq)->bytes)
#define bqueue_size(bq)         ((bq)->bufsize)

int bqueue_reserve(bqueue *bq, int dsize, int where,
        struct iovec *iov, unsigned int *iovlen);

#define bqueue_push_front(bq, data, dsize) \
        bqueue_push((bq), (data), (dsize), BQUEUE_FRONT)

#define bqueue_push_back(bq, data, dsize) \
        bqueue_push((bq), (data), (dsize), BQUEUE_BACK)

int bqueue_push(bqueue *bq, void *data, int dsize, int where);

int bqueue_peek(bqueue *bq, int dsize, int from,
        struct iovec *iov, int *iovlen);

#define bqueue_pop_front(bq, data, dsize) \
        bqueue_pop((bq), (data), (dsize), BQUEUE_FRONT)

#define bqueue_pop_back(bq, data, dsize) \
        bqueue_pop((bq), (data), (dsize), BQUEUE_BACK)

int bqueue_pop(bqueue *bq, void *data, int dsize, int from);

void bqueue_log(int level, bqueue *bq);

#endif /* __BQUEUE_H */

