/*
 * Created on   : Sat Dec 11 16:43:29 EST 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * buf.h
 *
 * Buffer related utilities
 * 
 * $Id: buf.h,v 1.3 2005/01/03 23:15:01 pavlos Exp $Buffer related utilities
 */

#ifndef __BUF_H
#define __BUF_H

#include "bytes.h"
#include "aqueue.h"

#define BUF_OK                  0
#define BUF_ERROR              -1

typedef struct bufview bufview;

typedef struct bufmem {

    uint8_t *ptr;
    size_t exported_size;
    size_t ptr_size;

    int ptr_owner;

    int inuse;

    bufview *views;
    int view_count;

} bufmem;


struct bufview {

    int attached;

    bufmem *mem;

    uint8_t *ptr;
    size_t size;

    bufview *prev;
    bufview *next;
};

typedef struct bufmgr {
    
    aqueue avail;
    aqueue inuse;

} bufmgr;

int bufmem_init_alloc(bufmem *mem, size_t size, int view_count);
int bufmem_init_ptr(bufmem *mem, uint8_t *ptr, size_t size,
       int view_count, int owner);
int bufmem_done(bufmem *mem);
void bufmem_free(void *mem);

int bufmem_reset(bufmem *mem, int view_count);
int bufmem_cmp(const void *mem1, const void *mem2);

#define bufmem_use(mem) \
        do { \
            logerr_bug((mem)->inuse >= 0); \
            (mem)->inuse++; \
        } while (0)

#define bufmem_release(mem) \
        do { \
            logerr_bug((mem)->inuse >= 0); \
            (mem)->inuse--; \
        } while (0)

/* bufview */

int bufview_init(bufview *view);
int bufview_done(bufview *view);
int bufview_reset(bufview *view);

/* bufmgr */

int bufmgr_init(bufmgr *mgr);
int bufmgr_done(bufmgr *mgr);

bufmem  *bufmgr_getmem(bufmgr *mgr, size_t size);
bufview *bufmgr_getview(bufmgr *mgr, size_t size);

#endif /* __BUF_H */


