/*
 * Created on   : Sat Jul 24 18:16:44 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * bytes.h
 *
 * Useful routines for search and compare operations
 *
 * $Id: cmp.h,v 1.9 2005/06/28 00:46:11 pavlos Exp $
 */

#ifndef __CMP_H
#define __CMP_H

#define CMP_NO_CMP_FUNC       NULL

/*
 * Input  : void * - pointer to data,     Data
 *          void * - pointer to data key, Key
 *
 * Output : -1  Data <before>    Key
 *           0  Data <equal>     Key
 *           1  Data <after>     Key
 */
typedef int  (cmp_search_func)          (void *data, void *key);

#define cmp_before(r) \
        ((r) == -1)

#define cmp_after(r) \
        ((r) == 1)

#define cmp_equal(r) \
        ((r) == 0)

#define cmp_before_or_equal(r) \
        ((r) == -1 || (r) == 0)

#define cmp_after_or_equal(r) \
        ((r) == 1 || (r) == 0)

/*
 * Input  : void * - pointer to data, A
 *          void * - pointer to data, B
 *
 * Output : -1  A <before>    B
 *           0  A <equal>     B
 *           1  A <after>     B
 *          -2  A <not-equal> B
 *              [forces every element to be checked in sorted lists]
 */
typedef int  (cmp_cmp_func)             (const void *, const void *);

/*
 * Input  : void * - pointer user specified data (to keep state for example)
 *          void * - pointer to node data
 *
 * Output :  1  - success, continue with next element in the list
 *           0  - success, stop processing elements in the list
 *          -1  - error, stop processing elements in the list
 */
typedef int  (cmp_foreach_func)         (void *, void *);

typedef void (cmp_log_func) (int level, char *prefix, void *data, void *arg);

/*
 * Input  : void *data - pointer to user data
 *          void *key  - pointer to user data key
 *          [only one argument should be non-null]
 *
 * Output : unsigned 64 bit integer - Hashed value of the user data key.
 */
typedef uint64_t (cmp_hash64_func)      (void *data, void *key);

#define cmp_max(a,b)                    ((a) > (b) ? (a) : (b))
#define cmp_min(a,b)                    ((a) < (b) ? (a) : (b))

#define cmp_ascend(a,b)                 ((a) == (b) ? 0 : (a) < (b) ? -1 : 1)
#define cmp_descend(a,b)                ((a) == (b) ? 0 : (a) > (b) ? -1 : 1)

#endif /* __CMP_H */

