#include "logerr.h"
#include "aqueue.h"
#include "mem.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>


#define ITEMS           20
#define TIMES           1

int global_i = 1234;

typedef struct data {
    int a;
    int b;
} data;

#define datalog(dp)     (printf("DATA: %d - %d\n", (dp)->a, (dp)->b))

void
enqueue(aqueue *pq)
{
    int i = 0;
    int res = 0;
    int use_pointer = !bit_test(pq->mode, AQUEUE_COPY_DATA);

    printf("---- Enqueuing\n");

    for (i = 0; i < ITEMS; i++) {
        data d;
        d.a = rand() % 100;
        d.b = d.a+1;
        data *pd = NULL;

        if (use_pointer) {
            pd = mem_malloc(sizeof(data));
            logerr_passert(pd);

            *pd = d;
            res = aqueue_push_ptr_back(pq, pd, 1);

        } else {
            res = aqueue_push_data_back(pq, &d);
        }

        if (res == AQUEUE_FULL)
            logerr_assert_str(0, "Queue is full");
        logerr_assert_str(res == AQUEUE_OK, "main error");
        printf("Enqueued: %d - %d\n", d.a, d.b);
    }

}

void
enqueuev(aqueue *pq)
{
    int i = 0;
    int res = 0;
    int use_pointer = !bit_test(pq->mode, AQUEUE_COPY_DATA);

    printf("---- Enqueuing\n");

    for (i = 0; i < ITEMS; i++) {
        data d;

        struct iovec vector[2];
        int count = 2;

        vector[0].iov_base = &d.a;
        vector[0].iov_len  = sizeof(d.a);

        vector[1].iov_base = &d.b;
        vector[1].iov_len  = sizeof(d.b);

        d.a = rand() % 100;
        d.b = d.a+1;
        data *pd = NULL;

        if (use_pointer) {
            pd = mem_malloc(sizeof(data));
            logerr_passert(pd);

            *pd = d;
            res = aqueue_push_ptr_back(pq, pd, 1);

        } else {
            res = aqueue_pushv_data_back(pq, vector, count);
        }

        if (res == AQUEUE_FULL)
            logerr_assert_str(0, "Queue is full");
        logerr_assert_str(res == AQUEUE_OK, "main error");
        printf("Enqueued: %d - %d\n", d.a, d.b);
    }

}

void
dequeue(aqueue *pq)
{
    data d;
    int res = AQUEUE_OK;
    int use_pointer = !bit_test(pq->mode, AQUEUE_COPY_DATA);
    data *pd = NULL;

    printf("---- Dequeuing\n");

    pd = aqueue_peek_front(pq);
    if (pd != NULL)
        printf("Peeking FRONT : %d - %d\n", pd->a, pd->b);

    pd = aqueue_peek_back(pq);
    if (pd != NULL)
        printf("Peeking BACK  : %d - %d\n", pd->a, pd->b);

    while (1) {

        if (use_pointer) {
            res = aqueue_pop_ptr_front(pq, &pd);
            if (res == AQUEUE_EMPTY)
                break;

            logerr_passert(pd);
            d = *pd;
            mem_free(pd);

        } else {
            res = aqueue_pop_data_front(pq, &d);
            if (res == AQUEUE_EMPTY)
                break;
        }


        logerr_assert_str(res == AQUEUE_OK, "main error");
        printf("Dequeued: %d - %d\n", d.a, d.b);
    }
}

void
dequeuev(aqueue *pq)
{
    data d;
    int res = AQUEUE_OK;
    int use_pointer = !bit_test(pq->mode, AQUEUE_COPY_DATA);
    data *pd = NULL;

    struct iovec vector[2];
    int count = 2;

    vector[0].iov_base = &d.a;
    vector[0].iov_len  = sizeof(d.a);

    vector[1].iov_base = &d.b;
    vector[1].iov_len  = sizeof(d.b);

    printf("---- Dequeuing\n");

    pd = aqueue_peek_front(pq);
    if (pd != NULL)
        printf("Peeking FRONT : %d - %d\n", pd->a, pd->b);

    pd = aqueue_peek_back(pq);
    if (pd != NULL)
        printf("Peeking BACK  : %d - %d\n", pd->a, pd->b);

    while (1) {

        if (use_pointer) {
            res = aqueue_pop_ptr_front(pq, &pd);
            if (res == AQUEUE_EMPTY)
                break;

            logerr_passert(pd);
            d = *pd;
            mem_free(pd);

        } else {
            res = aqueue_popv_data_front(pq, vector, count);
            if (res == AQUEUE_EMPTY)
                break;
        }


        logerr_assert_str(res == AQUEUE_OK, "main error");
        printf("Dequeued: %d - %d\n", d.a, d.b);
    }
}


int
cmp(const void *pa, const void *pb)
{
    data a = *(data *)pa;
    data b = *(data *)pb;

    if (a.a < b.a)
        return -1;

    if (a.a > b.a)
        return 1;

    if (a.b < b.b)
        return -1;

    if (a.b > b.b)
        return 1;

    return  0;
}

int
main(int argc, char **argv)
{

    int res = 0;
    aqueue q;
    int i = 0;

    data d1;
    data d2;

    d1.a = 10;
    d1.b = 100;


    datalog(&d1);
    datalog(&d2);
    


    /* bounded queue */

    logerr_log(1, "-----> Bounded Queue");

    srand(getpid());

    res = aqueue_data_init(&q, 0, sizeof(data));

    logerr_assert_str(res == AQUEUE_OK, "main error");

    for (i = 0; i < TIMES; i++) {
        enqueue(&q);
        dequeue(&q);
    }

    res = aqueue_done(&q);
    logerr_assert_str(res == AQUEUE_OK, 0);

    /* sorted and bounded queue */

    logerr_log(1, "-----> Sorted Queue");

    srand(getpid());

    res = aqueue_init(&q, AQUEUE_SORTED | AQUEUE_BOUNDED | AQUEUE_COPY_DATA,
                ITEMS+1, -1, -1, sizeof(data),
                NULL, &cmp, NULL);

    logerr_assert_str(res == AQUEUE_OK, "main error");

    for (i = 0; i < TIMES; i++) {
        enqueue(&q);
        dequeue(&q);
    }

    res = aqueue_done(&q);
    logerr_assert_str(res == AQUEUE_OK, 0);

    /* sorted queue with preallocated elements that recycle */

    logerr_log(1, "-----> Sorted/Prealloc/Recycle Queue");

    srand(getpid());

    res = aqueue_init(&q,
                AQUEUE_SORTED | AQUEUE_PREALLOC | AQUEUE_RECYCLE,
                ITEMS+1, ITEMS+1, ITEMS+1, sizeof(data),
                &mem_free_func, &cmp, NULL);

    logerr_assert_str(res == AQUEUE_OK, "main error");

    for (i = 0; i < TIMES; i++) {
        enqueue(&q);
        enqueue(&q);
        dequeue(&q);
    }

    res = aqueue_done(&q);
    logerr_assert_str(res == AQUEUE_OK, 0);

    /* sorted queue with preallocated elements that recycle --  COPY_DATA */

    logerr_log(1, "-----> Sorted/Prealloc/Recycle Queue and COPY_DATA");

    srand(getpid());

    res = aqueue_init(&q,
                AQUEUE_SORTED |
                AQUEUE_PREALLOC |
                AQUEUE_RECYCLE | 
                AQUEUE_COPY_DATA,
                ITEMS+1, ITEMS+1, ITEMS+1, sizeof(data),
                &mem_free_func, &cmp, NULL);

    logerr_assert_str(res == AQUEUE_OK, "main error");

    for (i = 0; i < TIMES; i++) {
        enqueuev(&q);
        dequeue(&q);
        enqueue(&q);
        dequeuev(&q);
    }

    res = aqueue_done(&q);
    logerr_assert_str(res == AQUEUE_OK, 0);



    printf("Done\n");

    return 0;
}

