#include "logerr.h"
#include "aqueue.h"
#include "mem.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>


#define ITEMS           20
#define TIMES           1

int global_i = 1234;


void
enqueue(aqueue *pq)
{
    int i = 0;
    int res = 0;
    int use_pointer = !bit_test(pq->mode, AQUEUE_COPY_DATA);

    printf("---- Enqueuing\n");

    for (i = 0; i < ITEMS; i++) {
        int d;
        d = rand() % 100;
        int *pd = NULL;

        if (use_pointer) {
            pd = mem_malloc(sizeof(int));
            logerr_passert(pd);

            *pd = d;
            res = aqueue_push_ptr_back(pq, pd, 1);

        } else {
            res = aqueue_push_data_back(pq, &d);
        }

        if (res == AQUEUE_FULL)
            logerr_assert_str(0, "Queue is full");
        logerr_assert_str(res == AQUEUE_OK, "main error");
        printf("Enqueued: %d\n", d);
    }

}

void
dequeue(aqueue *pq)
{
    int d = -1;
    int res = AQUEUE_OK;
    int use_pointer = !bit_test(pq->mode, AQUEUE_COPY_DATA);
    int *pd = NULL;

    printf("---- Dequeuing\n");

    while (1) {

        if (use_pointer) {
            res = aqueue_pop_ptr_front(pq, &pd);
            if (res == AQUEUE_EMPTY)
                break;

            logerr_passert(pd);
            d = *pd;
            mem_free(pd);

        } else {
            res = aqueue_pop_data_front(pq, &d);
            if (res == AQUEUE_EMPTY)
                break;
        }


        logerr_assert_str(res == AQUEUE_OK, "main error");
        printf("Dequeued: %d\n", d);
    }
}

int
cmp(const void *pa, const void *pb)
{
    int a = *(int *)pa;
    int b = *(int *)pb;

    if (a < b)
        return -1;

    if (a == b)
        return 0;

    if (a > b)
        return 1;

    return -2;
}

int
main(int argc, char **argv)
{

    int res = 0;
    aqueue q;
    int i = 0;

    /* bounded queue */

    logerr_log(1, "-----> Bounded Queue");

    srand(getpid());

    res = aqueue_data_init(&q, 0, sizeof(int));

    logerr_assert_str(res == AQUEUE_OK, "main error");

    for (i = 0; i < TIMES; i++) {
        enqueue(&q);
        dequeue(&q);
    }

    res = aqueue_done(&q);
    logerr_assert_str(res == AQUEUE_OK, 0);

    /* sorted and bounded queue */

    logerr_log(1, "-----> Sorted Queue");

    srand(getpid());

    res = aqueue_init(&q, AQUEUE_SORTED | AQUEUE_BOUNDED | AQUEUE_COPY_DATA,
                ITEMS+1, -1, -1, sizeof(int),
                NULL, &cmp, NULL);

    logerr_assert_str(res == AQUEUE_OK, "main error");

    for (i = 0; i < TIMES; i++) {
        enqueue(&q);
        dequeue(&q);
    }

    res = aqueue_done(&q);
    logerr_assert_str(res == AQUEUE_OK, 0);

    /* sorted queue with preallocated elements that recycle */

    logerr_log(1, "-----> Sorted/Prealloc/Recycle Queue");

    srand(getpid());

    res = aqueue_init(&q,
                AQUEUE_SORTED | AQUEUE_PREALLOC | AQUEUE_RECYCLE,
                ITEMS+1, ITEMS+1, ITEMS+1, sizeof(int),
                &mem_free_func, &cmp, NULL);

    logerr_assert_str(res == AQUEUE_OK, "main error");

    for (i = 0; i < TIMES; i++) {
        enqueue(&q);
        enqueue(&q);
        dequeue(&q);
    }

    res = aqueue_done(&q);
    logerr_assert_str(res == AQUEUE_OK, 0);

    /* sorted queue with preallocated elements that recycle --  COPY_DATA */

    logerr_log(1, "-----> Sorted/Prealloc/Recycle Queue and COPY_DATA");

    srand(getpid());

    res = aqueue_init(&q,
                AQUEUE_SORTED |
                AQUEUE_PREALLOC |
                AQUEUE_RECYCLE | 
                AQUEUE_COPY_DATA,
                ITEMS+1, ITEMS+1, ITEMS+1, sizeof(int),
                &mem_free_func, &cmp, NULL);

    logerr_assert_str(res == AQUEUE_OK, "main error");

    for (i = 0; i < TIMES; i++) {
        enqueue(&q);
        enqueue(&q);
        dequeue(&q);
    }

    res = aqueue_done(&q);
    logerr_assert_str(res == AQUEUE_OK, 0);



    printf("Done\n");

    return 0;
}

