
#define IDSTR               "server"
#define ACTIONSTR           "receiving"

#include "ipc_ex.h"

#include "atimer.h"

atimer_group tg;
int packet = 0;
int timers = 0;

void
read_in_one_exact_buffer(ipc_channel *c)
{
    int res = 0;
    rec r;

    logerr_log(2, "Waiting to recv [read_in_one_exact_buffer]");

    res = ipc_recv(c, &r, sizeof(rec));
    logerr_fatal_str(res > 0, "server failed");
    logerr_log(1, "server: received record with %d bytes", res);

    prec(&r);
}

void
server(void)
{
    int timers_processed = 0;
    int res = 0;
    ipc_channel client;
    fd_set rfds;
    sigset_t unblock_sigmask;
    struct timeval timeout;

    res = ipc_init(&client);
    logerr_fatal_str(res == IPC_OK, "server failed");

    res = ipc_create_fifo(FIFO_SEND, 0700, IPC_CREATE|IPC_USE_EXISTING);
    logerr_fatal_str(res == IPC_OK, "client failed");

    res = ipc_open_read_fifo(&client, FIFO_SEND, 1);
    logerr_fatal_str(res == IPC_OK, "server failed");

    res = sigemptyset(&unblock_sigmask);
    logerr_assert_sys(res != -1, "Failed to empty new sigmask");

    res = sigaddset(&unblock_sigmask, SIGALRM);
    logerr_assert_sys(res != -1, "Failed to set signal in new sigmask");

    timeout.tv_sec = 5;
    timeout.tv_usec = 0;

    while (timers_processed < timers) {

        int got_timer = 0;
        int fd_ready = 0;

        FD_ZERO(&rfds);
        FD_SET(client.read.fd, &rfds);

        logerr_log(100, "");
        logerr_log(100, "");
        logerr_log(100, "before select pid(%d)", getpid());

        res = asignal_select(client.read.fd+1, &rfds, NULL, NULL,
                NULL,
                &unblock_sigmask);

        logerr_assert_sys(res != -1 || errno == EINTR, "Select failed");
        logerr_log(100, "after select res(%d)", res);

        got_timer = ((res == 0) || (res == -1 && errno == EINTR));
        fd_ready = (res == 1);

        if (got_timer)
            logerr_log(100, "Checking timers");

        while (got_timer) {
            
            char data[100];
            struct timeval now;
            struct timeval diff;
            atimer_header tcur;
            atimer_header tnext;
            int got_timer = 0;

            res = atimer_get_expired(&tg, data, &tcur, &tnext, NULL);

            logerr_log(1, "Getting expired!");

            logerr_fatal_str(res >= ATIMER_OK, "server failed");

            got_timer = res == ATIMER_OK;

            if (!got_timer)
                break;

            res = gettimeofday(&now, NULL);
            logerr_assert_sys(res != -1, "gettimeofday failed");

            atimer_timeval_get_offset(&now, &tcur.expires, &diff);

            timers_processed++;
            logerr_log(1, "Processing timer %d: %s", timers_processed, data);
            atimer_timeval_log(1, "  ---- > NOW  ", &now);
            atimer_timeval_log(1, "  ---- > DIFF ", &diff);
            atimer_header_log(1, "    [current] ", &tcur);
            if (!atimer_zero_timeval(&tnext.expires))
                atimer_header_log(1, "    [next]    ", &tnext);
        }

        if (fd_ready && FD_ISSET(client.read.fd, &rfds)) {
            read_in_one_exact_buffer(&client);
            packet++;
            res = atimer_set_timer_msec(&tg, "READ PACKET", 1500+packet);
            logerr_fatal_str(res == ATIMER_OK, "main failed");
        }

    }

    res = ipc_done(&client);
    logerr_fatal_str(res == IPC_OK, "server failed");
}

int
main(void)
{
    int res = 0;
    struct timeval failsafe;

    failsafe.tv_sec = 0;
    failsafe.tv_usec = 10000;

    printf("Begin %s\n", IDSTR);

    atimer_set_recurring_alarm(&failsafe);

    res = atimer_init(&tg, 100, 10, NULL);
    logerr_fatal_str(res == ATIMER_OK, "main failed");

    atimer_sleep_sec(2);

    res = atimer_set_timer_msec(&tg, "FIRST", 2000);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_msec(&tg, "SECOND", 5000);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_msec(&tg, "THIRD", 10000);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_msec(&tg, "FOURTH", 15000);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_msec(&tg, "FIFTH", 20000);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_sec(&tg, "SEC-2", 2);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_sec(&tg, "SEC-5", 5);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_msec(&tg, "MSEC-800", 800);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_msec(&tg, "MSEC-200", 200);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_usec(&tg, "USEC-800", 800);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_usec(&tg, "USEC-200", 200);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_nsec(&tg, "NSEC-10000", 10001);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_nsec(&tg, "NSEC-800", 800);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = atimer_set_timer_nsec(&tg, "NSEC-200", 200);
    logerr_fatal_str(res == ATIMER_OK, "main failed");
    timers++;

    res = asignal_block(SIGALRM, NULL);
    logerr_fatal_str(res == ASIGNAL_OK, "main failed");

    server();

    res = atimer_done(&tg);
    logerr_fatal_str(res == ATIMER_OK, "main failed");

    printf("Done\n");
    return 0;
}
