#include "bytes.h"
#include "bqueue.h"
#include "logerr.h"

typedef struct {
    u_char b1;
    u_char b2;
} twobytes;

void
push2(bqueue *bq, int start, int end, int where)
{
    int i = 0;
    int res = 0;
    twobytes bb;

    for (i = start; i <= end; i++) {
        bb.b1 = i;
        bb.b2 = i | 0xf0;
        res = bqueue_push(bq, &bb, sizeof(bb), where);
        if (res != sizeof(bb)) {
            logerr_log(-1, "res(%d)", res);
            logerr_fatal_str(res == sizeof(bb), "Failed to push");
        }
        logerr_log(1, "Pushed %.2x %.2x", bb.b1, bb.b2);
    }

}

void
pop2(bqueue *bq, int elem, int where)
{
    int i = 0;
    int res = 0;
    twobytes bb;

    for (i = 0; i < elem; i++) {
        res = bqueue_pop(bq, &bb, sizeof(bb), where);
        logerr_fatal_str(res == sizeof(bb), "Failed to pop");
        logerr_log(1, "Popped %.2x %.2x", bb.b1, bb.b2);
    }
}

void
push(bqueue *bq, int start, int end, int where)
{
    int i = 0;
    int res = 0;
    u_char byte = 0;

    for (i = start; i <= end; i++) {
        byte = i;
        res = bqueue_push(bq, &byte, sizeof(byte), where);
        logerr_fatal_str(res == sizeof(byte), "Failed to push");
        logerr_log(1, "Pushed %.2x res(%d)", byte, res);
    }

}

void
pop(bqueue *bq, int elem, int where)
{
    int i = 0;
    int res = 0;
    u_char byte = 0;

    for (i = 0; i < elem; i++) {
        res = bqueue_pop(bq, &byte, sizeof(byte), where);
        logerr_fatal_str(res == sizeof(byte), "Failed to pop");
        logerr_log(1, "Popped %.2x", byte);
    }
}



int
main()
{
    bqueue bq;

    bqueue_init(&bq, 32);

    push2(&bq, 1, 15, BQUEUE_BACK);
    bqueue_log(1, &bq);

    pop2(&bq, 5, BQUEUE_FRONT);
    bqueue_log(1, &bq);

    push2(&bq, 32, 36, BQUEUE_BACK);
    bqueue_log(1, &bq);

    bqueue_done(&bq);

    return 0;
}
