#include "logerr.h"
#include "aqueue.h"
#include "msem.h"
#include <unistd.h>
#include <stdio.h>

typedef struct thread_state {
    pthread_t thread;
    aqueue *q;
    int sleep;
    int id;
    msem *nempty;
    msem *nstored;
} thread_state;

void *
producer(void *arg)
{
    int i = 0;
    int res = 0;
    thread_state *state = arg;

    for (i = 0; i < 20; i++) {
        int *pi = malloc(sizeof(int));

        *pi = state->id * 100000 + i;

        res = msem_wait(state->nempty);
        logerr_assert_str(res == MSEM_OK, "Failed while waiting");

        res = aqueue_push_ptr_back(state->q, pi, 1);
        logerr_assert_str(res == AQUEUE_OK, "Failed to insert element");

        printf("[thread %d]: Inserted element %d\n", state->id, *pi);

        res = msem_post(state->nstored);
        logerr_assert_str(res == MSEM_OK, "Failed while posting");
    }

    return NULL;
}

void *
consumer(void *arg)
{
    thread_state *state = arg;
    char idstr[50];
    int max_times = 2;
    int times = max_times;
    int *pi = NULL;

    snprintf(idstr, 50, "[thread %d]:", state->id);

    while (times-- > 0) {

        int res = 0;
        void *data = NULL;

        printf("%s Start Servicing\n", idstr);

        res = msem_trywait(state->nstored);
        logerr_assert_str(res == MSEM_OK || res == MSEM_EAGAIN,
                        "Failed trywait");

        if (res == MSEM_EAGAIN) {
            printf("%s Sleeping\n", idstr);
            sleep(state->sleep);
            printf("%s Awake\n", idstr);
            continue;
        }

        res = aqueue_pop_ptr_front(state->q, &data);
        logerr_assert_str(res == AQUEUE_OK, "Failed to dequeue");

        pi = data;
        printf("%s Removed list element: %d\n", idstr, *pi);
        free(pi);
        times = max_times;

        res = msem_post(state->nempty);
        logerr_assert_str(res == MSEM_OK, "Failed while posting");

    }

    printf("%s Stop Servicing\n", idstr);
    return NULL;
}


int
main(int argc, char **argv)
{

    int res = 0;
    aqueue thequeue;
    msem t_nempty;
    msem t_nstored;

    thread_state prod1;
    thread_state prod2;
    thread_state cons1;
    thread_state cons2;

    res = aqueue_ptr_init(&thequeue, 1, &free);
    logerr_assert_str(res == AQUEUE_OK, "main failed");

    res = msem_init(&t_nempty, 0, 5);
    logerr_assert_str(res == MSEM_OK, "main failed");

    res = msem_init(&t_nstored, 0, 0);
    logerr_assert_str(res == MSEM_OK, "main failed");

    prod1.id = 1;
    prod1.q  = &thequeue;
    prod1.nempty  = &t_nempty;
    prod1.nstored = &t_nstored;

    prod2.id = 2;
    prod2.q  = &thequeue;
    prod2.nempty  = &t_nempty;
    prod2.nstored = &t_nstored;

    cons1.id = 10;
    cons1.q  = &thequeue;
    cons1.nempty  = &t_nempty;
    cons1.nstored = &t_nstored;
    cons1.sleep = 2;

    cons2.id = 20;
    cons2.q  = &thequeue;
    cons2.nempty  = &t_nempty;
    cons2.nstored = &t_nstored;
    cons2.sleep = 1;

    pthread_create(&prod1.thread, NULL, &producer, &prod1);
    pthread_create(&prod2.thread, NULL, &producer, &prod2);
    pthread_create(&cons1.thread, NULL, &consumer, &cons1);
    pthread_create(&cons2.thread, NULL, &consumer, &cons2);

    pthread_join(prod1.thread, NULL);
    pthread_join(prod2.thread, NULL);
    pthread_join(cons1.thread, NULL);
    pthread_join(cons2.thread, NULL);

    res = aqueue_done(&thequeue);
    logerr_assert_str(res == AQUEUE_OK, "main failed");

    res = msem_done(&t_nempty);
    logerr_assert_str(res == MSEM_OK, "main failed");

    res = msem_done(&t_nstored);
    logerr_assert_str(res == MSEM_OK, "main failed");

    printf("Done\n");

    return 0;
}

