#!/bin/sh
#
# Runs all the examples
#

function run
{
        echo ">>> Running $1"
        ./$* && echo ">>> PASSED $1" || echo ">>> FAILED $1"
        echo ">>> Done. Pausing before next test"
        echo
        sleep 1
}

function run_simple
{
        run alist_ex
        run alist_thread_ex
        run aqueue_ex
        run aqueue2_ex
        run adar_ex
        run ahash_ex
        run ahash_stress
        run asleep
        run atimer_ex
        run bqueue_ex
        run msem_thread_ex
        run asys_ex
}

function run_client_server
{
        client=$1
        server=$2
        waitfor=$3

        # test client/server
        echo
        echo ">>> Starting $server"
        ./$server &
        pid=$!
        sleep 1
        run $client
        sleep $waitfor
        kill -9 $pid
        echo ">>> Done with client-server test. Pausing before next test"
        echo
        sleep 1
}

function run_shared
{
        test=$1

        echo ">>> Create shared memory"
        run $test /tmp.$test 1

        echo ">>> Write to shared memory"
        run $test /tmp.$test 3

        echo ">>> Read from shared memory"
        run $test /tmp.$test 4

        echo ">>> Destroy shared memory"
        run $test /tmp.$test 2

        echo ">>> Done with shared test. Pausing before next test"
        echo
        sleep 1
}

function run_all
{
        run_simple
        run_client_server ipc_client ipc_server 1
        run_client_server "udgram -c" "udgram -s" 10
        run_client_server "ustream -c" "ustream -s" 1
        run_client_server "ustream2 -c" "ustream2 -s" 1
        run_shared shared
        run_shared shared2
}

if [ -n "$*" ]; then
        run $*
else
        run_all
fi
