/*
 * Example on posix shared memory
 */

#include "logerr.h"

#include <stdio.h>

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <stdlib.h>

extern int shm_open(const char *, int, mode_t);
extern int shm_unlink(const char *);

#define FILE_MODE       0700

int
main(int argc, char **argv)
{
    int code = 0;
    int res = 0;
    int length = 10;
    int fd;
    char *ptr;
    struct stat stat;

    char *filename = NULL;

    srand(getpid());

    logerr_assert_str(argc >= 3, "Args not right");

    filename = argv[1];
    code = atoi(argv[2]);

    switch (code) {

    case 1:

        fd = shm_open(filename, O_RDWR|O_CREAT|O_EXCL, FILE_MODE);
        logerr_return_sys(fd != -1, "main failed", 1);
        
        res = ftruncate(fd, length);
        logerr_return_sys(res != -1, "main failed", 1);

        ptr = mmap(NULL, length, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
        logerr_return_sys(ptr != MAP_FAILED, "main failed", 1);

        /*
        res = shm_unlink(filename);
        logerr_return_sys(res != -1, "main failed", 1);

        printf("Unlinked\n");

        printf("Sleeping\n");
        sleep(10);
        */

        {
            int i;
            for (i = 0; i <length; i++)
                printf ("Reading: loc[%d] = %d\n", i, *(ptr+i));
        }

        res = close(fd);
        logerr_return_sys(res != -1, "main failed", 1);
        printf("Closed\n");

        break;

    case 2:

        res = shm_unlink(filename);
        logerr_return_sys(res != -1, "main failed", 1);

        break;

    case 3:
    case 4:

        fd = shm_open(filename, (code == 4) ? O_RDONLY : O_RDWR, FILE_MODE);
        logerr_return_sys(fd != -1, "main failed", 1);

        res = fstat(fd, &stat);
        logerr_return_sys(res != -1, "main failed", 1);

        ptr = mmap(NULL, stat.st_size,
                        (code == 4) ? PROT_READ : PROT_WRITE,
                        MAP_SHARED, fd, 0);
        logerr_return_sys(ptr != MAP_FAILED, "main failed", 1);
                        
        res = close(fd);
        logerr_return_sys(res != -1, "main failed", 1);

        if (code == 3) {
            int i;
            int r = abs((rand() / 2)) * 10;
            r = r % 100;
            printf ("---> Starting from number %d\n", r);
            for (i = 0; i <stat.st_size; i++) {
                *(ptr+i) = r+i;
                printf ("Writing: loc[%d] = %d\n", i, *(ptr+i));
            }
        } else {
            int i;
            for (i = 0; i <stat.st_size; i++)
                printf ("Reading: loc[%d] = %d\n", i, *(ptr+i));
        }

        break;
        
    default:

        logerr_assert_str(0, "Invalid code");
        break;

    }


    printf("Exiting normally\n");
    return res;
}


