/*
 * Example on posix shared memory
 */

#include "logerr.h"
#include "mshared.h"

/* #include <stdio.h> */
/* #include <unistd.h> */
/* #include <stdlib.h> */

extern int shm_open(const char *, int, mode_t);
extern int shm_unlink(const char *);

int
main(int argc, char **argv)
{
    int code = 0;
    int res = 0;
    int length = 10;
    char *ptr;
    mshared mem;

    char *filename = NULL;

    srand(getpid());

    logerr_assert_str(argc >= 3, "Args not right");

    filename = argv[1];
    code = atoi(argv[2]);

    switch (code) {

    case 1:

        printf("Creating\n");

        res = mshared_create(&mem, filename, 0777, length);
        logerr_return_str(res == MSHARED_OK, "main failed", 1);
        ptr = mshared_ptr(&mem);
        
        {
            int i;
            for (i = 0; i <length; i++)
                printf ("Reading: loc[%d] = %d\n", i, *(ptr+i));
        }

        break;

    case 2:

        res = mshared_open(&mem, filename, 1);
        logerr_return_str(res == MSHARED_OK, "main failed", 1);

        printf("Unlinking\n");

        res = mshared_unlink(&mem);
        logerr_return_str(res == MSHARED_OK, "main failed", 1);

        break;

    case 3:

        res = mshared_open(&mem, filename, 0);
        logerr_return_str(res == MSHARED_OK, "main failed", 1);
        ptr = mshared_ptr(&mem);

        {
            int i;
            int r = abs((rand() / 2)) * 10;
            r = r % 100;
            printf("Writing\n");
            printf ("---> Starting from number %d\n", r);
            for (i = 0; i <length; i++) {
                *(ptr+i) = r+i;
                printf ("Writing: loc[%d] = %d\n", i, *(ptr+i));
            }
        }
        break;

    case 4:

        res = mshared_open(&mem, filename, 1);
        logerr_return_str(res == MSHARED_OK, "main failed", 1);
        ptr = mshared_ptr(&mem);

        {
            int i;
            printf("Reading\n");
            for (i = 0; i <length; i++)
                printf ("Reading: loc[%d] = %d\n", i, *(ptr+i));
        }
        break;
        
    default:

        logerr_assert_str(0, "Invalid code");
        break;

    }


    res = mshared_done(&mem);
    logerr_return_str(res == MSHARED_OK, "main failed", 1);

    printf("Exiting normally\n");
    return 0;
}


