#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "logerr.h"

#define DEF_SERVER_PATH         "/tmp/udgram.server"
#define DEF_CLIENT_PATH         "/tmp/udgram.client"

typedef struct global_options {
    int server;
    char server_path[500];
    char client_path[500];
} global_options;


void
server(global_options *op)
{
    int res = 0;
    int sockfd = -1;
    int seq = 0;
    struct sockaddr_un addr;
    unsigned char buf[200];

    logerr_log(1, "Server");

    sockfd = socket(AF_LOCAL, SOCK_DGRAM, 0);
    logerr_fatal_sys(sockfd != -1, "");

    unlink(op->server_path);
    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, op->server_path, sizeof(addr.sun_path));

    res = bind(sockfd, (struct sockaddr *)&addr, sizeof(addr));
    logerr_fatal_sys(res != -1, "");
    
    seq = 0;
    while (1) {
        
        sleep(1);

        res = read(sockfd, buf, sizeof(buf));
        logerr_fatal_sys(res >= 0, "read error");
        
        if (res == 0) {
            logerr_log(1, "Peer closed");
            break;
        }

        seq += res;
        logerr_log(1, "read pkt(%d) seq(%d)", res, seq);

    } 

    res = close(sockfd);
    logerr_fatal_sys(res != -1, "");
}

void
client(global_options *op)
{
    int i = 0;
    int seq = 0;
    int res = 0;
    int sockfd = -1;
    int num = 0;
    struct sockaddr_un addr;
    unsigned char buf[500];

    logerr_log(1, "Client");

    memset(buf, 0xf0, sizeof(buf));

    sockfd = socket(PF_LOCAL, SOCK_DGRAM, 0);
    logerr_fatal_sys(sockfd != -1, "");

    unlink(op->client_path);
    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, op->client_path, sizeof(addr.sun_path));

    res = bind(sockfd, (struct sockaddr *)&addr, sizeof(addr));
    logerr_fatal_sys(res != -1, "");

    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, op->server_path, sizeof(addr.sun_path));

    res = connect(sockfd, (struct sockaddr *)&addr, sizeof(addr));
    logerr_fatal_sys(res != -1, "");

    logerr_log(1, "Established connection.");

    num = 10;
    seq = 0;
    for (i = 0; i < num; i++) {

        res = write(sockfd, buf, sizeof(buf));
        logerr_fatal_sys(res >= 0, "write error");

        seq += res;
        logerr_log(1, "wrote pkt(%d) seq(%d)", res, seq);

    }

    res = close(sockfd);
    logerr_fatal_sys(res != -1, "");
}

void
main_parse_cmdline(global_options *op, int argc, char **argv)
{
    static char short_options[] = "scp:r:";
    char c = 0;

    logerr_passert(op);

    memset(op, 0, sizeof(global_options));

    /* initialize options */

    op->server = -1;
    strncpy(op->server_path, DEF_SERVER_PATH, sizeof(op->server_path));
    strncpy(op->client_path, DEF_CLIENT_PATH, sizeof(op->client_path));

    /* parse command line options */

    while (1) {

        c = getopt(argc, argv, short_options);

        if (c == EOF)
            break;

        switch (c) {

        case 's':
            op->server = 1;
            break;

        case 'c':
            op->server = 0;
            break;

        case 'p':

            strncpy(op->server_path, optarg, sizeof(op->server_path));
            logerr_log(1, "Using server pathname '%s'", optarg);

            break;

        case 'r':

            strncpy(op->client_path, optarg, sizeof(op->client_path));
            logerr_log(1, "Using client pathname '%s'", optarg);

            break;

        case '?':

            logerr_abort(0, "failed to parse command line");
            break;
            
        default:
            logerr_abort(0, "do not know hot to process arg '%c'", c);
            break;

        }
    }
    
}

int
main (int argc, char **argv)
{
    global_options options;

    logerr_log(1, "PID(%d)", getpid());

    main_parse_cmdline(&options, argc, argv);

    if (options.server == 1)
        server(&options);
    else if (options.server == 0)
        client(&options);

    exit (0);
}


