/*
 * Created on   : Tue Jun 15 15:04:04 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * logerr.c
 *
 * Error logging routines
 *
 * $Id: logerr.c,v 1.21 2005/12/12 16:39:56 pavlos Exp $
 */

#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#include "logerr.h"

logerr_global_state logerr_globals = {-999, 999, 0, ""};

void
logerr_setlevel(int level)
{
    logerr_globals.min_level = -999;
    logerr_globals.max_level = level;
}

void
logerr_setlevel_only(int level)
{
    logerr_globals.min_level = level;
    logerr_globals.max_level = level;
}

void
logerr_setlevel_range(int min, int max)
{
    logerr_globals.min_level = min;
    logerr_globals.max_level = max;
}

void
logerr_set_idstr(const char *str)
{
    int count = LOGERR_MAX_IDSTR_LEN - 3;

    logerr_globals.idstr[0] = '[';
    strncpy(logerr_globals.idstr+1, str, count);
    logerr_globals.idstr[count] = 0;

    count = strlen(logerr_globals.idstr);
    logerr_globals.idstr[count++] = ']';
    logerr_globals.idstr[count++] = ' ';
    logerr_globals.idstr[count] = 0;

    logerr_globals.idlen = count;
}

int
logerr_aux(uint16_t flags, const void *p, const char *source,
                const char *file, unsigned int line)
{
    const char *msg = p;

    if (bit_match(flags, LOGERR_RETRY_EINTR) && errno == EINTR) {
        /* XXX Just to catch first occurence. Remove. */
        logerr_assert_str(0, "EINTR caught! -- just a test, remove NOW!");
        return 0;
    }

    if (logerr_globals.idlen > 0)
        fprintf(stderr, "%s", logerr_globals.idstr);

    if (logerr_globals.idlen > 0)
        fprintf(stderr, "%s", logerr_globals.idstr);
    fprintf(stderr, "%s:%d: ", file, line);
    if (msg != NULL)
       fprintf(stderr, "%s\n", msg);
    else
       fprintf(stderr, "\n");

    if (source != NULL) {
        if (logerr_globals.idlen > 0)
            fprintf(stderr, "%s", logerr_globals.idstr);
        fprintf(stderr, "    failed test: '%s'\n", source);
    }

    if (bit_test(flags, LOGERR_ERRNO)) {
        if (logerr_globals.idlen > 0)
            fprintf(stderr, "%s", logerr_globals.idstr);
        fprintf(stderr, "    errno %d: %s\n", errno, strerror(errno));
    }

    return 1;
}


int
logerr_assert_aux(int valid, uint16_t flags, const void *p,
                const char *source, const char *file, unsigned int line)
{
    if (valid)
        return 0;

    logerr_aux(flags, p, source, file, line);
    if (logerr_globals.idlen > 0)
        fprintf(stderr, "%s", logerr_globals.idstr);
    fprintf(stderr, "Assert at %s:%d: %s\n", file, line, source);
    exit(1);

    return 1;
}


void
logerr_fatal_aux(void)
{
    exit(1);
}

/*
 * pos = LOGERR_SINGLE, LOGERR_FIRST, LOGERR_MIDDLE, LOGERR_LAST
 */
void
logerr_log_aux(int level, int pos,
        const char *file, unsigned int line,
        const char *format, ...)
{
    va_list ap;

    if (!logerr_loglevel_permits(level))
        return;

    if (file != NULL && bit_test(pos, LOGERR_FIRST))
        fprintf(stderr, "%-13s:%-4d: ", file, line);

    if (logerr_globals.idlen > 0 && bit_test(pos, LOGERR_FIRST))
        fprintf(stderr, "%s", logerr_globals.idstr);

    va_start(ap, format);

    vfprintf(stderr, format, ap);

    if (bit_test(pos, LOGERR_LAST)) {
        fprintf(stderr, "\n");
        fflush(stderr);
    }

    va_end(ap);
}

void
logerr_abort(int exitcode, const char *format, ...)
{
    va_list ap;

    fprintf(stderr, "%serror: ", logerr_globals.idstr);

    va_start(ap, format);

    vfprintf(stderr, format, ap);
    fprintf(stderr, "\n");
    fflush(stderr);

    va_end(ap);

    exit(exitcode);
}

