/*
 * Created on   : Tue Jul 20 15:28:32 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * mem.h
 *
 * Wrappers for memory allocation functions to keep
 * track of allocated memory.
 * 
 * $Id: mem.h,v 1.4 2004/07/31 23:37:35 pavlos Exp $
 */

#ifndef __MEM_H
#define __MEM_H

#include <stdlib.h>
#include <sys/uio.h>

#define MEM_LOG_LEVEL           99

#define MEM_OK                  0
#define MEM_ENOSPACE           -1

#ifdef NDEBUG

#define mem_malloc(size)        malloc((size))
#define mem_calloc(elem,elsize) calloc((elem), (elsize))
#define mem_free(ptr)           free((ptr))

#else /* NDEBUG */

#define mem_malloc(size) \
        mem_malloc_aux((size), __FILE__, __LINE__)

#define mem_calloc(elem,elsize) \
        mem_calloc_aux((elem), (elsize),__FILE__, __LINE__)

#define mem_free(ptr) \
        mem_free_aux((ptr), __FILE__, __LINE__)

#endif /* NDEBUG */

void * mem_malloc_aux(size_t size,
        const char *file, unsigned int line);

void * mem_calloc_aux(size_t nelem, size_t elsize,
        const char *file, unsigned int line);

void mem_free_aux(void *ptr,
        const char *file, unsigned int line);

void mem_free_func(void *ptr);

int mem_copy_from_vector(void *buf, size_t size,
                const struct iovec *vector, int count);
int mem_copy_to_vector(void *buf, size_t size,
                const struct iovec *vector, int count);

#endif /* __MEM_H */

