/*
 * Created on   : Wed Jun 16 14:39:13 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * msem.c
 *
 * Wrapper functions for POSIX memory-based (unnamed) semaphores that
 * can be shared across threads and processes.
 *
 *    + Uses System V semaphores to implement the POSIX semaphores if
 *      the compile flag MSEM_USE_SYSV is defined.
 *
 * $Id: msem.c,v 1.4 2004/07/21 22:00:34 pavlos Exp $
 */

#ifdef MSEM_USE_SYSV

#else  /* MSEM_USE_SYSV */

#endif /* MSEM_USE_SYSV */

#include "msem.h"
#include "logerr.h"


#ifdef MSEM_USE_SYSV

/* XXX NOT IMPLEMENTED -- see stevens UNIX IPC book */

#else  /* MSEM_USE_SYSV */

int
msem_init(msem *sem, int pshared, unsigned int value) {

    int res = 0;

    logerr_passert(sem);

    res = sem_init(sem, pshared, value);
    logerr_return_sys(res != -1,
                "POSIX semaphore initialization failed",
                errno);

    return MSEM_OK;
}


int
msem_done(msem *sem) {

    int res = 0;

    logerr_passert(sem);

    res = sem_destroy(sem);
    logerr_return_sys(res != -1,
                "POSIX semaphore destroy failed",
                errno);

    return MSEM_OK;
}


int
msem_post(msem *sem) {

    int res = 0;

    logerr_passert(sem);

    res = sem_post(sem);
    logerr_return_sys(res != -1,
                "POSIX semaphore post failed",
                errno);

    return MSEM_OK;
}


int
msem_getvalue(msem *sem, int *pval) {

    int res = 0;

    logerr_passert(sem);

    res = sem_getvalue(sem, pval);
    logerr_return_sys(res != -1,
                "POSIX semaphore getvalue failed",
                errno);

    return MSEM_OK;
}

int
msem_wait(msem *sem) {

    int res = 0;

    logerr_passert(sem);

    logerr_return_rsys(
                (res = sem_wait(sem)) != -1,
                "POSIX semaphore wait failed",
                errno);

    return MSEM_OK;
}

int
msem_trywait(msem *sem) {

    int res = 0;

    logerr_passert(sem);

    logerr_return_rsys(
                !((res = sem_trywait(sem)) == -1 && errno != EAGAIN),
                "POSIX semaphore wait failed",
                errno);

    return (res != -1) ? MSEM_OK : MSEM_EAGAIN;
}

#endif /* MSEM_USE_SYSV */

