/*
 * Created on   : Wed Jun 16 14:39:13 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * msem.h
 *
 * Wrapper functions for POSIX memory-based (unnamed) semaphores that
 * can be shared across threads and processes.
 *
 *    + Uses System V semaphores to implement the POSIX semaphores if
 *      the compile flag MSEM_USE_SYSV is defined.
 *
 * $Id: msem.h,v 1.4 2004/07/21 22:00:34 pavlos Exp $
 */

#ifndef __MSEM_H
#define __MSEM_H

#include <errno.h>

#ifdef MSEM_USE_SYSV

#else  /* MSEM_USE_SYSV */

#include <semaphore.h>

#endif /* MSEM_USE_SYSV */


/* positive return codes reserved for errno */

#define MSEM_EAGAIN             EAGAIN
#define MSEM_OK                 0
#define MSEM_ERROR             -1

#ifdef MSEM_USE_SYSV

typedef struct msem {
    int id;
} msem;

#else  /* MSEM_USE_SYSV */

typedef sem_t msem;

#endif /* MSEM_USE_SYSV */

int msem_init(msem *sem, int pshared, unsigned int value);
int msem_done(msem *sem);
int msem_post(msem *sem);
int msem_getvalue(msem *sem, int *pval);
int msem_wait(msem *sem);
int msem_trywait(msem *sem);

#endif /* __MSEM_H */

