/*
 * Created on   : Tue Jun 29 16:49:17 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * mshared.h
 *
 * Utilities for sharing memory among processes
 * Front-end to POSIX shared memory
 *
 * $Id: mshared.h,v 1.3 2004/06/30 23:36:38 pavlos Exp $
 */

#ifndef __MSHARED_H
#define __MSHARED_H

#include <unistd.h>
#include <sys/types.h>

#define mshared_ptr(MEM)        (logerr_passert((MEM)), (MEM)->ptr)
#define mshared_size(MEM)       (logerr_passert((MEM)), (MEM)->size)

#define MSHARED_OK               1
#define MSHARED_ERROR           -1
#define MSHARED_ENOMEM          -2
#define MSHARED_ENAME           -3
#define MSHARED_EINIT           -4
#define MSHARED_EOPEN           -5
#define MSHARED_ECLOSE          -6
#define MSHARED_EMAP            -7
#define MSHARED_EUNMAP          -8
#define MSHARED_ERESIZE         -9
#define MSHARED_EFSTAT          -10
#define MSHARED_EUNLINK         -11


typedef struct mshared {
    int fd;
    size_t size;
    int open;
    int mapped;
    int unlinked;
    char *name;
    void *ptr;
} mshared;

int mshared_create(mshared *mem, char *name, mode_t mode, size_t size);
int mshared_open(mshared *mem, char *name, int read_only);

int mshared_unlink(mshared *mem);
int mshared_done(mshared *mem);

#endif /* __MSHARED_H */

