/*
 * Created on   : Fri Jul  1 17:01:31 EDT 2005
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * plog.c
 *
 * Simple packet logging
 *
 * $Id: plog.c,v 1.2 2005/11/02 00:19:17 pavlos Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "plog.h"
#include "logerr.h"

int
plog_init(plog *log, char *filename)
{
    memset(log, 0, sizeof(plog));

    log->fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, 00644);

    if (log->fd == -1) {
        logerr_log(-1, "Failed to open packet logfile '%s'.", filename);
        logerr_fatal_sys(log->fd != -1, "file open error: ");
    }
    
    return PLOG_OK;
}


int
plog_done(plog *log)
{
    close(log->fd);
    return PLOG_OK;
}

plog *
plog_create(char *filename)
{
    plog *log = malloc(sizeof(plog));

    logerr_bug(log != NULL);
    plog_init(log, filename);
    
    return log;
}

void
plog_destroy(plog *log)
{
    logerr_bug(log != NULL);

    plog_done(log);
    free(log);
}


int
plog_packet_multi(plog *log, plog_pkt *pkt, int count)
{
    int res = 0;
    int i = 0;

    for (i = 0; i < count; i++) {
        res = write(log->fd, pkt+i, sizeof(plog_pkt));
        sysbug_on(res == -1);
        logerr_bug(res == sizeof(plog_pkt));
    }

    return PLOG_OK;
}

void
plog_pkt_log(plog_pkt *pkt, FILE *outfile)
{
    unsigned int seqno = pkt->seqno;
    unsigned int pktsize = pkt->pktsize;
    int flag_cnt = 0;

    fprintf(outfile, "%ld.%.6ld %u %u ",
            pkt->tstamp.tv_sec,
            pkt->tstamp.tv_usec,
            seqno,
            pktsize);

    if (bit_test(pkt->flags, PLOG_FLAG_SEND))
        fprintf(outfile, "%sSEND", (flag_cnt++ > 0 ? "|" : ""));
    else if (bit_test(pkt->flags, PLOG_FLAG_RECV))
        fprintf(outfile, "%sRECV", (flag_cnt++ > 0 ? "|" : ""));

    fprintf(outfile, "\n");
}

void
plog_log2ascii(FILE *outfile, char *infilename)
{
    int res = 0;
    int fd = -1;
    plog_pkt pkt;

    fd = open(infilename, O_RDONLY, 0);

    if (fd == -1) {
        logerr_log(-1, "Failed to open packet logfile '%s'.", infilename);
        logerr_fatal_sys(fd != -1, "file open error: ");
    }

    while ((res = read(fd, &pkt, sizeof(pkt))) != 0) {
        
        if (res == -1 && errno == EINTR)
            continue;

        sysbug_on(res == -1);
        logerr_bug(res == sizeof(pkt));

        plog_pkt_log(&pkt, outfile);

    }
    
    close(fd);
}

