/*
 * Created on   : Fri Jul  1 17:01:31 EDT 2005
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * plog.h
 *
 * Simple packet logging
 *
 * $Id: plog.h,v 1.1 2005/07/06 23:18:48 pavlos Exp $
 */

#ifndef __PLOG_H
#define __PLOG_H

#include <sys/time.h>
#include <arpa/inet.h>

#define PLOG_OK         0

#define PLOG_FLAG_SEND  0x0001
#define PLOG_FLAG_RECV  0x0002

typedef struct plog_pkt {
    struct timeval tstamp;
    uint32_t seqno;
    uint16_t pktsize;
    uint16_t flags;
} plog_pkt;

typedef struct plog {
    int fd;
} plog;


int plog_init(plog *log, char *filename);
int plog_done(plog *log);

plog * plog_create(char *filename);
void plog_destroy(plog *log);

#define plog_packet(_l, _p) \
        plog_packet_multi((_l), (_p), 1);

#define plog_packet_simple(_l, _tstamp, _seqno, _pktsize, _flags) \
    do { \
        plog_pkt _p; \
        memset(&_p, 0, sizeof(plog_pkt)); \
        _p.tstamp = *(_tstamp); \
        _p.seqno = (_seqno); \
        _p.pktsize = (_pktsize); \
        _p.flags = (_flags); \
        plog_packet_multi((_l), &_p, 1); \
    } while(0);

int plog_packet_multi(plog *log, plog_pkt *pkt, int count);

void plog_pkt_log(plog_pkt *pkt, FILE *outfile);

void plog_log2ascii(FILE *outfile, char *infilename);

#endif /* __PLOG_H */

