/*
 * Created on   : Wed Oct  6 12:41:41 EDT 2004
 * Created by   : Pavlos Papageorgiou <pavlos@eng.umd.edu>
 * 
 * procman.c
 *
 * Process management functions
 * 
 * $Id: procman.c,v 1.4 2005/07/20 23:07:48 pavlos Exp $
 */

#include "procman.h"

#include <sys/types.h>

#include "logerr.h"

#if _POSIX_PRIORITY_SCHEDULING > 0

int
procman_schedule(int policy, int priority)
{
    int res = 0;
    struct sched_param schedParam;

    if (getuid() != 0)
        return PROCMAN_ENOTROOT;

    schedParam.sched_priority = priority;

    res = sched_setscheduler(0, policy, &schedParam);

    logerr_log(-1, "changed scheduler: policy %d priority %d",
        policy, priority);

    return (res == 0) ? PROCMAN_OK : PROCMAN_ERROR;
}

#endif

#if _POSIX_MEMLOCK > 0

int
procman_mlock_this(int flags, const void *addr, size_t len)
{
    int res = 0;

    if (getuid() != 0)
        return PROCMAN_ENOTROOT;

    if (addr != NULL)
        res = mlock(addr, len);
    else
        res = mlockall(flags);

    return (res == 0) ? PROCMAN_OK : PROCMAN_ERROR;
}

#endif

