// Kindly reused from:
// http://www.macdonald.egate.net/CompSci/knowlton/quiksort.c

/* quiksort.c */
/* An implementation of a quicksort algorithm. */

#include "quicksort.h"

/* partitions input_array, returning middle index. Used by procedure quicksort. */
static inline
int quicksort_partition(long long input_array[], int top, int bottom)
{
  long long x = input_array[top];
  int i = top - 1;
  int j = bottom + 1;
  long long temp;

  do
  {
    /* move j towards top to the next element less than or equal to x. */
    do
    {
      j--;
    } while(x > input_array[j]);

    /* move j towards bottom to the next element greater than or equal to x. */
    do
    {
      i++;
    } while(x < input_array[i]);

    if (i < j)
    {
      /* switch elements at positions i and j. */
      temp = input_array[i];
      input_array[i] = input_array[j];
      input_array[j] = temp;
    }
  } while(i < j);	/* loop ends when i and j have met in the middle. */

  return j;	/* j and above represents top partition, below j is bottom partition. */
}

/* Quicksort procedure. Sorts an array of ints in descending order. */
void quicksort(long long input_array[], int top, int bottom)
{
  int middle;
  
  if (top < bottom)
  {
    middle = quicksort_partition(input_array, top, bottom);
    quicksort(input_array, top, middle);	/* sort top partition */
    quicksort(input_array, middle + 1, bottom); /* sort bottom partition */
  }
}
