#include "libestimator.h"

void est_copy_info_entries(struct est_info *dst,
                           struct est_info *src,
                           int count)
{
        int i = 0;
        for (i = 0; i < count; i++) {
                dst[i] = src[i];
        }
}

int est_push_info_entries(struct est_channel *ch,
                          struct est_info *info_array,
                          int info_array_entries)
{
        int rc = 0;
        int pushed = 0;

        while (pushed < info_array_entries) {
                int ok = 0;
                char error[100];
                struct est_info_cmd cmd;
                struct est_info_cmd reply;
                int remaining = info_array_entries - pushed;

                memset(&cmd, 0, sizeof(cmd));
                ok = est_fill_cmd_header(&cmd.hdr, EST_TYPE_INFO);
                BUG_ON(!ok);

                cmd.action = EST_INFO_PUT;
                cmd.info_entries = cmp_min(remaining, EST_INFO_ENTRIES_PER_MSG);
                est_copy_info_entries(cmd.info,
                                      info_array + pushed,
                                      cmd.info_entries);

                rc = est_send_ctrlmsg(ch, &cmd, error, sizeof(error));
                if (rc) {
                        LOG("failed to send INFO/PUT cmd: %s", error);
                        goto error;
                }

                rc = est_recv_ctrlmsg(ch, &reply, sizeof(reply),
                                      error, sizeof(error));
                if (rc) {
                        LOG("failed to recv INFO/PUT reply: %s", error);
                        goto error;
                }

                ok = (reply.hdr.type & EST_TYPE_INFO &&
                      est_is_reply_final(&reply.hdr) &&
                      reply.action == cmd.action);
                if (!ok) {
                        rc = EST_EINVAL;
                        LOG("reply was not an INFO/PUT");
                        goto error;
                }

                pushed += cmd.info_entries;
        }
        return 0;

error:
        return rc;
}

int est_pull_info_entries(struct est_channel *ch,
                          int info_array_max_entries_to_get,
                          struct est_info *info_array,
                          int *info_array_entries)
{
        int rc = 0;
        int ok = 0;
        char error[100];
        struct est_info_cmd cmd;
        int exhausted = 0;
        int pulled = 0;
        int reply_count = 0;

        memset(&cmd, 0, sizeof(cmd));
        ok = est_fill_cmd_header(&cmd.hdr, EST_TYPE_INFO);
        BUG_ON(!ok);

        cmd.action = EST_INFO_GET;
        cmd.info_max_entries_to_get = info_array_max_entries_to_get;
        
        rc = est_send_ctrlmsg(ch, &cmd, error, sizeof(error));
        if (rc) {
                LOG("failed to send INFO/GET cmd: %s", error);
                goto error;
        }

        while (!exhausted) {
                struct est_info_cmd reply;
                int remaining = info_array_max_entries_to_get - pulled;

                rc = est_recv_ctrlmsg(ch, &reply, sizeof(reply),
                                      error, sizeof(error));
                if (rc) {
                        LOG("failed to recv INFO/GET reply: %s", error);
                        goto error;
                }

                reply_count++;
                ok = (reply.hdr.type & EST_TYPE_INFO &&
                      reply.action == cmd.action);
                if (!ok) {
                        rc = EST_EINVAL;
                        ERR("reply was not an INFO/GET");
                        goto error;
                }
                /* sender sent more than we asked with initial cmd */
                if (reply.info_entries > remaining) {
                        LOG("we asked for %d entries, got %d",
                            remaining, reply.info_entries);
                        LOG("this is reply %d, initial max %d",
                            reply_count, info_array_max_entries_to_get);
                        BUG();
                }

                est_copy_info_entries(info_array + pulled,
                                      reply.info,
                                      reply.info_entries);
                pulled += reply.info_entries;
                exhausted = est_is_reply_final(&reply.hdr);
        }

        *info_array_entries = pulled;
        return 0;

error:
        return rc;
}
