/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 al_int32 with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
              estimation tool
   Author   : Manish Jain ( jain@cc.gatech.edu )
              Constantinos Dovrolis (dovrolis@cc.gatech.edu )
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
              program of the US department 
--------------------------------------------------*/

/*
 * $header$
 */

#include <stdio.h>
#include <pthread.h>

static inline void
doexit(int rc)
{
	/* must not conflict with PTHREAD_CANCELED */
	if (rc == -1)
	    rc = -999;

	printf("%s\n", __FUNCTION__);
	pthread_exit((void *)rc);
}

#ifdef MGRP
#include "bhtp.h"
#include "udpex.h"
#include "probe.h"
#endif

#if SIZEOF_LONG == 4
  typedef long l_int32 ;
  typedef unsigned long l_uint32 ;
#elif SIZEOF_INT == 4
  typedef int l_int32 ;
  typedef unsigned int l_uint32 ;
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <errno.h>
#include <ctype.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include <float.h>
#include <fcntl.h>
#include <sys/utsname.h>
#ifdef THRLIB
#include <pthread.h>
#endif

/* Code numbers sent from receiver to sender using the TCP control stream */
#define  CTR_CODE           0x80000000
#define  SEND_FLEET         0x00000001
#define  RECV_FLEET         0x00000002
#define  CONTINUE_STREAM    0x00000003
#define  FINISHED_STREAM    0x00000007
#define  TERMINATE          0x00000005
#define  ABORT_FLEET        0x00000006
#define  SEND_TRAIN         0x00000008 
#define  FINISHED_TRAIN     0x00000009
#define  BAD_TRAIN          0x0000000a
#define  GOOD_TRAIN         0x0000000b

#define UDP_BUFFER_SZ       400000    /* bytes */
#define TREND_ARRAY_LEN     50 

#define NUM_STREAM          12
#define MAX_STREAM_LEN      400
#define STREAM_LEN          100       /* # of packets */
#ifdef MGRP
/* min packet size can be lower if overhead > 28, fe when using bhtp.
 *
 * We do this so that we have the same ip packet size when
 * using small packets (bhtp and non-bhtp case), so that
 * it is easier to compare in experiments.
 */
#define MIN_PKT_SZ          (200 - gb->overhead + 28)
#else
#define MIN_PKT_SZ          200       /* bytes */
#endif


#define MAX_TRAIN           5
#define TRAIN_LEN           50

#ifdef MGRP
static inline int
get_udp_ip_overhead(int sock)
{
	int overhead = 20;	// IP
	int active = udpex_socket(sock);
	FILE *fp = stdout;

	if (active)
		overhead += sizeof(struct bhtphdr) +	// MGRP
			    sizeof(struct udpexhdr);	// UDPex
	else
		overhead += 8;	// UDP


	fprintf(fp,
		"MGRP is %s for socket %d, ip/udp overhead %d bytes\n",
		(active ? "ON" : "OFF"),
		sock,
		overhead);

	return overhead;
}

static inline void
set_payload_int(char *buf, int index, l_int32 val)
{
	memcpy(buf+index*sizeof(l_int32), &val, sizeof(l_int32));
}

/* just to get rid of compiler warnings -- quick and dirty fix */
static inline int
l32int(l_int32 val)
{
	return val;
}

static inline int
longint(long int val)
{
	return val;
}

static inline int
susecint(suseconds_t val)
{
	return val;
}
/*
  Compute the time difference in microseconds
  between two timeval measurements
*/
static inline double time_to_us_delta(struct timeval tv1, struct timeval tv2)
{
  double time_us;
  time_us = (double)
    ((tv2.tv_sec-tv1.tv_sec)*1000000+(tv2.tv_usec-tv1.tv_usec));
  return time_us;
}
#endif
