/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
            estimation tool
   Author   : Manish Jain ( jain@cc.gatech.edu )
              Constantinos Dovrolis (dovrolis@cc.gatech.edu )
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
            program of the US department 
--------------------------------------------------*/

/*
 * $Header: /net/cvs/bwtest/pathload/pathload_rcv.c,v 1.139 2006/05/19 22:30:13 jain Exp $
 */

#include "libpathload.h"
#include "pathload_gbls.h"
#include "pathload_rcv.h"

void pathload_rcv_parse_cmdline(int argc, char* argv[], struct pathload_rcv_options *op)
{
  int found = 0;
  l_int32 errflg=0;
  l_int32 c ;

  if ( argc == 1 ) errflg++ ;
#ifdef MGRP
  while ((c = getopt(argc, argv, "t:s:hw:vHqo:O:N:VxT:U:")) != EOF)
#else
  while ((c = getopt(argc, argv, "t:s:hw:vHqo:O:N:V")) != EOF)
#endif
    switch (c) 
    {
      case 't':
        op->requested_delay = atoi(optarg);
        break;
      case 's':
        strcpy(op->sender_hostname,optarg);
        break;
      case 'w':
        op->bw_resol = atof(optarg);
        break;
      case 'q':
        op->Verbose=0;
        op->verbose=0;
        break;
      case 'v':
        op->Verbose=1;
        break;
      case 'O':
        op->file=1;
        strcpy(op->filename,optarg);
        strcpy(op->mode,"a");
        break;
      case 'o':
        op->file=1;
        strcpy(op->filename,optarg);
        strcpy(op->mode,"w");
        break;
      case 'H':
      case 'h':
        rcv_help() ;
        break ;
      case 'N':
        op->netlog=1; 
        strcpy(op->netlogfile,optarg);
        break;
      case 'V':
        op->VVerbose = 1;
        break ;
#ifdef MGRP
      case 'T':
	found = sscanf(optarg, "%u", &op->rcv_tcp_port);
	if (found != 1) {
		fprintf(stderr, "invalid TCP send port '%s'", optarg);
		exit(-1);
	}
        break ;
      case 'U':
	found = sscanf(optarg, "%u", &op->rcv_udp_port);
	if (found != 1) {
		fprintf(stderr, "invalid UDP recv port '%s'", optarg);
		exit(-1);
	}
        break ;
      case 'x':
        op->rcv_use_kernel_probes = 1;
        break ;
#endif
      case '?':
        errflg++;
        break ;
    }
  if (errflg)
  {
    fprintf(stderr, "usage: pathload_rcv [-q|-v] [-o <filename>] [-N <filename>]\
[-w <bw_resol>] [-h|-H] -s <sender>\n");
    exit (0);
  }
}

int main(int argc, char* argv[])
{
	int rc = 0;
	struct pathload_rcv_options op;

	pathload_rcv_options_init(&op);
	pathload_rcv_parse_cmdline(argc, argv, &op);

	pathload_rcv(&op);
	return rc;
}

