/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
              estimation tool
   Author   : Manish Jain ( jain@cc.gatech.edu )
              Constantinos Dovrolis (dovrolis@cc.gatech.edu )
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
              program of the US department 
--------------------------------------------------*/

#include "libpathload.h"
#include "libestimator.h"

/*
 * $Header: /net/cvs/bwtest/pathload/pathload_rcv.h,v 1.72 2006/05/19 20:21:15 jain Exp $
 */

#define SCALE_FACTOR                2
#define NOTREND                     1
#define INCREASING                  2
#define GREY                        3

#define MEDIUM_LOSS_RATE            3
#define HIGH_LOSS_RATE              15
#define MIN_PARTITIONED_STREAM_LEN  5       /* number of packet */
#define MIN_STREAM_LEN              36      /* # of packets */
#define PCT_THRESHOLD               .55
#define PDT_THRESHOLD               .4
#define AGGREGATE_THRESHOLD         .6

#define NUM_RETRY_RATE_MISMATCH     0
#define NUM_RETRY_CS                1 
#define MAX_LOSSY_STREAM_FRACTION   50

#define MIN_TIME_INTERVAL           7         /* microsecond */
#define MAX_TIME_INTERVAL           200000    /* microsecond */

l_int32 aggregate_trend_result() ;
double time_to_us_delta(struct timeval tv1, struct timeval tv2);
double get_avg(double data[], l_int32 no_values);
double pairwise_comparision_test (double array[] , l_int32 start , l_int32 end);
double pairwise_diff_test(double array[] ,l_int32 start , l_int32 end);
l_int32 rate_adjustment(l_int32 flag);
l_int32 eliminate_sndr_side_CS (double sndr_time_stamp[], l_int32 split_owd[] ) ;
l_int32 eliminate_rcvr_side_CS ( double rcvr_time_stamp[] , double[], double[],l_int32,l_int32,l_int32 *,l_int32 * ) ;
l_int32 eliminate_b2b_pkt_ic ( double rcvr_time_stamp[] , double[], double[],l_int32,l_int32,l_int32 *,l_int32 * ) ;
void adjust_offset_to_zero(double owd[], l_int32 last_pkt_id);
l_int32 recv_fleet() ;
void print_contextswitch_info(l_int32 num_sndr_cs[], l_int32 num_rcvr_cs[],l_int32 discard[],l_int32 stream_cnt);
void *ctrl_listen(void *) ;
void radj_notrend() ;
void radj_increasing() ;
void radj_greymin() ;
void radj_greymax() ;
l_int32 calc_param();
void get_sending_rate() ;
double  get_adr() ;
l_int32 recv_train(l_int32 , struct timeval *, l_int32 );
l_int32 recvfrom_latency(struct sockaddr_in rcv_udp_addr);
double grey_bw_resolution() ;
void get_pct_trend(double[] , l_int32[], l_int32 ) ;
void get_pdt_trend(double[] , l_int32[], l_int32 ) ;
void get_trend(double owdfortd[],l_int32 pkt_cnt );
l_int32 get_sndr_time_interval(double snd_time[],double *sum);
void sig_alrm();
void terminate_gracefully(struct timeval exp_start_time);
l_int32 check_intr_coalescence(struct timeval time[],l_int32, l_int32 * );
void rcv_order_float(float unord_arr[],float ord_arr[],l_int32 start, l_int32 num_elems);
void rcv_order_dbl(double unord_arr[],double ord_arr[], l_int32 start,l_int32 no_elems) ;
void rcv_order_int(l_int32 unord_arr[], l_int32 ord_arr[], l_int32 no_elems);
l_int32 max(l_int32,l_int32 ) ;
void rcv_send_ctr_mesg(char *ctr_buff, l_int32 ctr_code) ;
l_int32 rcv_recv_ctr_mesg(l_int32 ctr_strm, char *ctr_buff) ;
l_int32 equal(double a , double b);
l_int32 less_than(double a , double b);
l_int32 grtr_than(double a , double b);
void netlogger();
void print_time(FILE *fp, l_int32 time);
void rcv_help();
void sig_sigusr1() ;
void sig_sigusr1_second() ;

struct rcv_pathload_globals {

	/* Characteristics of Packet Stream */
	l_int32 time_interval ;              /* in us */
	l_uint32 transmission_rate ; /* in bps */
	l_int32 cur_pkt_sz ;                 /* in bytes */
	l_int32 max_pkt_sz ;                 /* in bytes */
	l_int32 rcv_max_pkt_sz ;             /* in bytes */
	l_int32 snd_max_pkt_sz ;             /* in bytes */
	l_int32 num_stream ;
	l_int32 stream_len ;                 /* in packets */

	l_int32 verbose ;
	l_int32 Verbose ;
	l_int32 VVerbose ;

	/* modification not protected by MGRP defines */
	int overhead;

#ifdef MGRP
	int rcv_tcp_port;
	int rcv_udp_port;
	int rcv_use_kernel_probes;
#endif

	l_int32 exp_flag  ;
	l_int32 grey_flag  ;
	double tr ;
	double adr ;
	double tr_max ;
	double tr_min ;
	double grey_max  , grey_min  ;
	l_int32 sock_tcp, sock_udp;
	l_int32 exp_fleet_id  ;
	float bw_resol;
	l_uint32 min_time_interval, snd_latency, rcv_latency ;
	l_int32 repeat_fleet ;
	l_int32 counter ; /* # of consecutive times act-rate != req-rate */
	l_int32 converged_gmx_rmx ;
	l_int32 converged_gmn_rmn ;
	l_int32 converged_rmn_rmx ;
	l_int32 converged_gmx_rmx_tm ;
	l_int32 converged_gmn_rmn_tm ;
	l_int32 converged_rmn_rmx_tm ;
	double cur_actual_rate , cur_req_rate ;
	double prev_actual_rate , prev_req_rate ;
	double max_rate , min_rate ;
	l_int32 max_rate_flag , min_rate_flag ; 
	l_int32 bad_fleet_cs , bad_fleet_rate_mismatch ;
	l_int32 retry_fleet_cnt_cs , retry_fleet_cnt_rate_mismatch ;
	FILE *pathload_fp, *netlog_fp  ;
	double pct_metric[50], pdt_metric[50];
	l_int32 trend_idx ;
	double snd_time_interval ;
	l_int32 min_rsleep_time ; 
	l_int32 num ; 
	l_int32 slow;
	l_int32 interrupt_coalescence;
	l_int32 ic_flag ;
	l_int32 netlog ;
	char hostname[256];
	l_int32 repeat_1, repeat_2;
	l_int32 lower_bound;
	l_int32 increase_stream_len;
	l_int32 requested_delay ;
	struct timeval  exp_start_time ;

        struct est_probe_status *status;

        int use_estimator_rpc;
        struct est_channel estimator_channel;
};

extern pthread_key_t rcv_globals_key;

#define rcv_globals() __rcv_globals(__FILE__, __LINE__)

static inline struct rcv_pathload_globals *
__rcv_globals(char *file, int line)
{
	struct rcv_pathload_globals *gb = NULL;

	gb = pthread_getspecific(rcv_globals_key);
	if (gb == NULL) {
		printf("%s:%d: rcv thread-specific global pointer is NULL\n", file, line);
		exit(-1);
	}
	// printf("%s:%d: %p:\n", file, line, gb);
	return gb;
}


#ifdef MGRP
static inline __u16
get_rcv_tcp_port()
{
	struct rcv_pathload_globals *gb = rcv_globals();
	__u16 port = gb->rcv_tcp_port;
	return port;
}

static inline __u16
get_rcv_udp_port()
{
	struct rcv_pathload_globals *gb = rcv_globals();
	__u16 port = gb->rcv_udp_port;
	return port;
}
#endif


#ifdef THRLIB
typedef struct 
{
    pthread_t ptid ;
    l_int32 finished_stream ;
    l_int32 stream_cnt;
    struct rcv_pathload_globals *gb;
} thr_arg;
#endif
