/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
              estimation tool
   Author   : Manish Jain (jain@cc.gatech.edu)
              Constantinos Dovrolis (dovrolis@cc.gatech.edu)
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
              program of the US department 
--------------------------------------------------*/

#include "libpathload.h"
#include "pathload_gbls.h"
#include "pathload_snd.h"

void pathload_snd_parse_cmdline(int argc, char* argv[], struct pathload_snd_options *op)
{
  int c ;
  int errflg=0;
#ifdef MGRP
  int found = 0;
#endif

#ifdef MGRP
  while ((c = getopt(argc, argv, "ihHqxT:U:")) != EOF)
#else
  while ((c = getopt(argc, argv, "ihHq")) != EOF)
#endif
    switch (c) 
    {
      case 'H':
      case 'h':
        snd_help() ;
        break ;
      case 'i':
        op->iterate=1;
        break;
      case 'q':
        op->quiet=1;
        break;
#ifdef MGRP
      case 'T':
	found = sscanf(optarg, "%u", &op->snd_tcp_port);
	if (found != 1) {
		fprintf(stderr, "invalid TCP send port '%s'\n", optarg);
		exit(-1);
	}
        break ;
      case 'U':
	found = sscanf(optarg, "%u", &op->snd_udp_port);
	if (found != 1) {
		fprintf(stderr, "invalid UDP recv port '%s'\n", optarg);
		exit(-1);
	}
        break ;
      case 'x':
        op->snd_use_kernel_probes=1;
        break;
#endif
      case '?':
        errflg++;
    }
  if (errflg)
  {
    fprintf(stderr, "usage: pathload_snd [-q] [-H|-h]\n");
    exit(-1);
  }

}

int main(int argc, char* argv[])
{
	int rc = 0;
	struct pathload_snd_options op;

	pathload_snd_options_init(&op);
	pathload_snd_parse_cmdline(argc, argv, &op);

	pathload_snd(&op);
	return rc;
}

