/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
              estimation tool
   Author   : Manish Jain ( jain@cc.gatech.edu )
              Constantinos Dovrolis (dovrolis@cc.gatech.edu )
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
              program of the US department 
--------------------------------------------------*/

#include "libestimator.h"

/* 
 * $Header: /net/cvs/bwtest/pathload/pathload_snd.h,v 1.38 2006/05/19 19:12:27 jain Exp $
 */

int send_fleet() ;
int send_train();
int snd_send_ctr_mesg(char *ctr_buff, l_int32 ctr_code);
l_int32 snd_recv_ctr_mesg( char *ctr_buff);
l_int32 send_latency() ;
void min_sleeptime() ;
void snd_order_int(int unord_arr[], int ord_arr[], int num_elems);
double time_to_us_delta(struct timeval tv1, struct timeval tv2);

void snd_order_dbl(double unord_arr[], double ord_arr[],int start, int num_elems);
void snd_order_float(float unord_arr[], float ord_arr[],int start, int num_elems);
void snd_order_int(int unord_arr[], int ord_arr[], int num_elems);
void snd_help() ;

struct snd_pathload_globals {

	/* Characteristics of Packet Stream */
	l_int32 time_interval ;              /* in us */
	l_uint32 transmission_rate ; /* in bps */
	l_int32 cur_pkt_sz ;                 /* in bytes */
	l_int32 max_pkt_sz ;                 /* in bytes */
	l_int32 rcv_max_pkt_sz ;             /* in bytes */
	l_int32 snd_max_pkt_sz ;             /* in bytes */
	l_int32 num_stream ;
	l_int32 stream_len ;                 /* in packets */

#ifdef MGRP
	int snd_tcp_port;
	int snd_udp_port;
	int snd_use_kernel_probes;
#endif

	l_int32 fleet_id_n ;
	l_int32 fleet_id  ;
	int sock_udp, sock_tcp, ctr_strm, send_buff_sz;
	unsigned int rcv_tcp_adrlen;
	struct sockaddr_in snd_udp_addr, snd_tcp_addr, rcv_udp_addr, rcv_tcp_addr;
	l_int32 min_sleep_interval ; /* in usec */
	l_int32 min_timer_intr ; /* in usec */
	int gettimeofday_latency ;

	int quiet ;
	int iterate ;

/* 	l_int32 verbose ; */
/* 	l_int32 Verbose ; */
/* 	l_int32 VVerbose ; */

	struct est_probe_status *status;
        
        int use_estimator_rpc;
        struct est_channel estimator_channel;

	int bind_failed;
};

extern pthread_key_t snd_globals_key;

#define snd_globals() __snd_globals(__FILE__, __LINE__)

static inline struct snd_pathload_globals *
__snd_globals(char *file, int line)
{
	struct snd_pathload_globals *gb = NULL;

	gb = pthread_getspecific(snd_globals_key);
	if (gb == NULL) {
		printf("%s:%d: snd thread-specific global pointer is NULL\n", file, line);
		exit(-1);
	}
	// printf("%s:%d: %p:\n", file, line, gb);
	return gb;
}


#ifdef MGRP
static inline __u16
get_snd_tcp_port()
{
	struct snd_pathload_globals *gb = snd_globals();
	__u16 port = gb->snd_tcp_port;
	return port;
}

static inline __u16
get_snd_udp_port()
{
	struct snd_pathload_globals *gb = snd_globals();
	__u16 port = gb->snd_udp_port;
	return port;
}
#endif


