#include "libpathload.h"

void
final_result(const int packet_size, const int packet_overhead,
             const long usec, const double max_p, const double min_p,
             const char *result,
             struct est_results *results,
             enum est_probe_status_value *status_value,
             int probe_bytes, int probe_pkts,
             int allow_piggyback,
             double adjust_stream,
             double adjust_fleet,
             enum pathload_final_result_adjust adjust,
             char *results_str, int results_strlen)

{
        double max = max_p;
        double min = min_p;
	struct timeval now;
	int payload_bytes = packet_size;
	int ip_bytes = payload_bytes + packet_overhead;
        int mgrp_on = packet_overhead > 28;
        struct pathload_results results_p;
        long tv_sec = 0;
        long tv_usec = 0;

	if (max == 0)
		max = min;
	else if (min == 0)
		min = max;

	gettimeofday(&now, NULL);
        tv_sec = now.tv_sec;
        tv_usec = now.tv_usec;
	snprintf(results_str, results_strlen,
                 "final-result: %s: %ld.%.6ld "
                 "payload %d ip %d sec %.2f max %.2f min %.2f result %s "
                 "adjust %.2f %.2f %s",
		 (mgrp_on ? "mgrp" : "orig"),
		 tv_sec, tv_usec,
		 payload_bytes, ip_bytes,
		 0.000001 * usec, max, min, result,
                 adjust_stream, adjust_fleet,
                 pathload_get_adjust_policy(adjust));

		
	/* fill the results structure */

        results_p.ended = now;
        results_p.mgrp = mgrp_on;
        results_p.piggyback = allow_piggyback;
        results_p.payload_bytes = payload_bytes;
        results_p.ip_bytes = ip_bytes;
        results_p.probe_bytes = probe_bytes;
        results_p.probe_pkts = probe_pkts;

        if (strcmp(result, "normal") == 0)
                results_p.type = PATHLOAD_RESULT_NORMAL;
        else if (strcmp(result, "coalescence") == 0)
                results_p.type = PATHLOAD_RESULT_COALESCENCE;
        else
                results_p.type = PATHLOAD_RESULT_UNKNOWN;

        results_p.usec = usec;

        // adjust the final result (store the original)
        // based on piggybacking feedback

        results_p.orig_min = min;
        results_p.orig_max = max;

        if (adjust == PATHLOAD_ADJUST_NONE) {
                results_p.min = min;
                results_p.max = max;

        } else if (adjust == PATHLOAD_ADJUST_STREAM) {
                results_p.min = est_adjust(min, adjust_stream);
                results_p.max = est_adjust(max, adjust_stream);

        } else if (adjust == PATHLOAD_ADJUST_FLEET) {
                results_p.min = est_adjust(min, adjust_fleet);
                results_p.max = est_adjust(max, adjust_fleet);

        } else {
                ERRFN("unknown adjustment type: %s", adjust);
        }

        results_p.adjust_stream = adjust_stream;
        results_p.adjust_fleet = adjust_fleet;

        /* heuristic to discard zero, invalid results
         *
         * in practice I see that pathload produces sometimes
         * results very close to zero that are invalid
         */
/*         if (max * min < 0.001) { */
/*                 results_p.type = PATHLOAD_RESULT_UNDEF; */
/*         } */

        if (results_p.type == PATHLOAD_RESULT_NORMAL)
                *status_value = EST_PROBE_STATUS_SUCCESS;
        else
                *status_value = EST_PROBE_STATUS_FAILED;

        pathload_results_to_est_results(&results_p, results);
}


