/*
 * pt.h		Packet Train -- definitions
 *
 * Created:	Tue Oct  4 14:30:29 EDT 2005
 *
 * Version:	$Id: pt.h,v 1.18 2005/12/16 23:05:20 pavlos Exp $
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 * truncated versions of original files found in:
 * /homes/pavlos/work/phd/projects/tagent/src/ptrain
 * taken on Tue Dec 20 10:58:36 EST 2005.
 */

#ifndef _PT_H
#define _PT_H

#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/ioctl.h>

#include "alog.h"

#define PT_TRAIN_MAX			200
#define PT_PKTSIZE_MAX			1500
#define PT_MAGIC_BYTE			0xaf

/* NULL values */

#define PT_NULL_IP		0xffffffff
#define PT_NULL_IP_ID		((uint32_t)-1)

#define PT_NULL_EPOCH		((uint32_t)-1)
#define PT_NULL_SEQNO		((uint32_t)-1)
#define PT_NULL_STAMP		((uint16_t)-1)
#define PT_NULL_DIFF_STAMP	((uint16_t)-1)
#define PT_NULL_BASE_STAMP	((uint64_t)-1)

#define PT_NULL_TGAP		((uint32_t)-1)
#define PT_NULL_TLEN		((uint16_t)-1)
#define PT_NULL_TSEQNO		((uint16_t)-1)

#define PT_NULL_PKTSIZE		((uint16_t)-1)

#define PT_FILENAME_STR		500
#define PT_IFACE_STR		50

#define PT_KSTAMP_MAX		8		/* multiples of 4 please */
#define PT_KSTAMP_MAGIC		0xafbfcfef

/* must match the top of pt_slog -- only this is captured by pcap */
typedef struct pt_slog_pcap pt_slog_pcap;
struct pt_slog_pcap {

	uint32_t magic;

	uint32_t src_epoch;
	uint32_t src_seqno;

	uint32_t train_gap;
	uint16_t train_len;
	uint16_t train_seqno;

	uint8_t delay_stage;
	uint8_t delay_usec;

	uint8_t kstamp_max;
	uint8_t kstamp_num;

	uint64_t src_base_stamp;

	uint16_t src_kstamp[PT_KSTAMP_MAX];
};

typedef struct pt_slog pt_slog;
struct pt_slog {

	uint32_t magic;

	uint32_t src_epoch;
	uint32_t src_seqno;

	uint32_t train_gap;
	uint16_t train_len;
	uint16_t train_seqno;

	uint8_t delay_stage;
	uint8_t delay_usec;

	uint8_t kstamp_max;
	uint8_t kstamp_num;

	uint64_t src_base_stamp;

	uint16_t src_kstamp[PT_KSTAMP_MAX];
	uint16_t src_pstamp;
	uint16_t src_ustamp2;
	uint16_t src_ustamp1;

	uint16_t src_sleep_diff;

	uint64_t src_load_epoch;

	uint32_t src_ipid;
	uint32_t dst_ipid;
	uint16_t pkt_size;
	uint8_t pad[6];
};

typedef struct pt_slog_v pt_slog_v;
struct pt_slog_v {
	pt_slog *slogs;
	int len;
};

typedef struct pt_header pt_header;
struct pt_header {
	/* must be on top and match pt_slog_cap */
	pt_slog slog;
};

extern const pt_slog PT_NULL_SLOG;

static inline void
pt_slog_clear(pt_slog *slogs, int len)
{
	int i = 0;
	for (i = 0; i < len; i++)
		slogs[i] = PT_NULL_SLOG;
}

#endif /* _PT_H */
