/*
 * rate.c	Log utilities
 *
 * Created:	Tue Oct  4 16:24:50 EDT 2005
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#include "rate.h"

void
rate_init(rate *r)
{
	memset(r, 0, sizeof(*r));
	rate_reset_total(r);
}

void
rate_reset_total(rate *r)
{
	int res = 0;

	r->total_bytes = 0;
	res = gettimeofday(&r->total_begin, NULL);
	SYSERR_ON(res);

	r->seg_bytes = 0;
	r->seg_begin = r->total_begin;
}

void
rate_reset_seg(rate *r, struct timeval *now)
{
	int res = 0;

	r->seg_bytes = 0;

	if (now == NULL) {
		res = gettimeofday(&r->seg_begin, NULL);
		SYSERR_ON(res);
	} else {
		r->seg_begin = *now;
	}

}

void
rate_report_seg(rate *r, int reset, FILE *fd, char *prefix)
{
	struct timeval now;
	long long report_usec = 0;

	double usec = 0;
	double rate = 0;

	usec = tick_now_udiff_past(&r->seg_begin, &now);
	report_usec = tick_tv_udiff(&r->seg_begin, &r->total_begin);

	rate = (8 * r->seg_bytes) * (1000000 / usec);
	
	fprintf(fd, "%s [%lld - %.0f sec]: %.0f Kbps\n",
		(prefix ? prefix : ""),
		report_usec/1000000,
		(report_usec + usec)/1000000,
		rate/1000);

	if (reset)
		rate_reset_seg(r, &now);
}

void
rate_report_total(rate *r, FILE *fd, char *prefix)
{
	struct timeval now;

	double usec = 0;
	double rate = 0;

	usec = tick_now_udiff_past(&r->total_begin, &now);
	rate = (8 * r->total_bytes) * (1000000 / usec);
	
	fprintf(fd, "%s [0 - %.0f sec]: %.0f Kbps\n",
		(prefix ? prefix : ""),
		usec/1000000,
		rate/1000);
}


