/*
 * A libary wrapper of the SYSSTAT utilities by Sebastien Godard
 * Currently contains the mpstat code: per-processor statistics
 *
 *      http://perso.orange.fr/sebastien.godard/
 *
 * (C) 2007 by Pavlos Papageorge (pavlos <at> cs.umd.edu)
 * (C) 2000-2007 by Sebastien GODARD (sysstat <at> orange.fr)
 *
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   *
 ***************************************************************************
 */

#ifndef _LIBSYSSTAT_H
#define _LIBSYSSTAT_H

/*
 * Typicall usage:
 *
 *      mpstat_init()
 *
 *      int curr = mpstat_init_current();
 *      struct mp_stats_ratio ratio[MAX_CPU_NUM+1];
 *
 *      while (1) {
 *              sleep(1);
 *              mpstat_read_stats(curr);
 *              mpstat_compute_interval_stats(curr, ratio, MAX_CPU_NUM+1);
 *              curr = mpstat_advance_current(curr);
 *
 *              // do something with the stats
 *              // ratio[0] : ratio stats across CPUs
 *              // ratio[i] : ratio stats for CPU i
 *      }
 */

#define MPSTAT_NO_HEADER        0
#define MPSTAT_WITH_HEADER      1

struct mp_stats_ratio_double {
   double cpu_idle;
   double cpu_iowait;
   double cpu_user;
   double cpu_nice;
   double cpu_system;
   double cpu_hardirq;
   double cpu_softirq;
   double cpu_steal;
   double irq;
};

/* same as mp_stats_ratio but multiplied by 1000 */
struct mp_stats_ratio_int {
   int cpu_idle;
   int cpu_iowait;
   int cpu_user;
   int cpu_nice;
   int cpu_system;
   int cpu_hardirq;
   int cpu_softirq;
   int cpu_steal;
   int irq;
};

void mpstat_init();

void mpstat_read_stats(int curr);

void mpstat_compute_total_stats(
                int curr, 
                struct mp_stats_ratio_double *ratio,
                int ratio_num);

void mpstat_compute_interval_stats(
                int curr, 
                struct mp_stats_ratio_double *ratio,
                int ratio_num);

int mpstat_advance_current(int curr);
int mpstat_init_current();

int mpstat_get_cpu_num();

void mpstat_ratio_double_str(char *str, int strlen, int with_header,
                             struct mp_stats_ratio_double *ratio,
                             int ratio_num);

void mpstat_ratio_int_str(char *str, int strlen, int with_header,
                          struct mp_stats_ratio_int *ratio,
                          int ratio_num);

void mpstat_ratio_int_oneline_str(char *str, int len,
                                  struct mp_stats_ratio_int *ratio,
                                  int idx);

void mpstat_downgrade_ratio_double_to_int(
                        const struct mp_stats_ratio_double *from_ratio_double,
                        struct mp_stats_ratio_int *to_ratio_int);

static inline void mpstat_downgrade_array_ratio_double_to_int(
                        const struct mp_stats_ratio_double *from_ratio_double,
                        struct mp_stats_ratio_int *to_ratio_int,
                        int ratio_num)
{
        int i = 0;
        for (i = 0; i < ratio_num; i++) {
                mpstat_downgrade_ratio_double_to_int(
                        &from_ratio_double[i],
                        &to_ratio_int[i]);
        }
}
#endif
