#!/bin/env python

import os
import sys
import re
import getopt

def get_script_path():
	script = sys.argv[0]
	path = re.sub('\/[^/]+$', '', script)
	if (path == script):
		path = '.'
	return path


class GlobalState:
	def __init__(self):

		self.tcp = False
		self.udp = False
		self.mss = 0
		self.reuse = False
		self.probe = False
		self.queue = False
		self.fraglen = 0

		self.path = get_script_path()
		self.libtsocket = "%s/../build/libtsocket.so" % self.path
		self.first_cmd_arg = 1

gb = GlobalState()

def main():
	parse_cmdline()

	cargv = ['env']
	cargv.append("LD_PRELOAD=%s" % gb.libtsocket);

	if gb.tcp:
		cargv.append("TSOCKET_TCP=1")
	if gb.udp:
		cargv.append("TSOCKET_UDP=1")
	if gb.probe:
		cargv.append("TSOCKET_PROBE=1")
	if gb.queue:
		cargv.append("TSOCKET_QUEUE=1")
	if gb.mss > 0:
		cargv.append("TSOCKET_MSS=%d" % gb.mss)
	if gb.fraglen > 0:
		cargv.append("TSOCKET_FRAGLEN=%d" % gb.fraglen)
	if gb.reuse:
		cargv.append("TSOCKET_REUSE=1")
	
	for a in sys.argv[gb.first_cmd_arg:]:
		cargv.append(a)
	
	print cargv;
	os.execv('/bin/env', cargv)


def parse_cmdline():
	needs_help = False
	short = "htum:rpqf:"
	long = ["help", "tcp", "udp", "mss=", "reuse", "probe", "queue",
		"fraglen="]

	try:
		opts, args = getopt.getopt(sys.argv[1:], short, long)
		for opt, val in opts:

			if opt in ("-h", "--help"):
				needs_help = True
				break

			if opt in ("-t", "--tcp"):
				gb.tcp = True
				gb.first_cmd_arg += 1

			if opt in ("-u", "--udp"):
				gb.udp = True
				gb.first_cmd_arg += 1

			if opt in ("-p", "--probe"):
				gb.probe = True
				gb.first_cmd_arg += 1

			if opt in ("-q", "--queue"):
				gb.queue = True
				gb.first_cmd_arg += 1

			if opt in ("-m", "--mss"):
				gb.mss = int(val)
				gb.first_cmd_arg += 1
				if opt == "-m":
					gb.first_cmd_arg += 1
	
			if opt in ("-f", "--fraglen"):
				gb.fraglen = int(val)
				gb.first_cmd_arg += 1
				if opt == "-f":
					gb.first_cmd_arg += 1
	
	except Exception, what:
		print what
		usage()
		sys.exit(2)
	
	if needs_help:
		usage()
		sys.exit(2)
		
def usage():
	print """
NAME
	tsocket.py - Preload the libtsocket library

USAGE

	tsocket.py [options] command <command-args>

	-h --help

	-t --tcp
	-u --udp
	-m --mss <TCP_MAXSEG>
	-r --reuse

EXAMPLES

	tsocket.py --tcp iperf -s
	tsocket.py --tcp iperf -c <server>
	"""

if __name__ == "__main__":
	main()

