/*
 * pad.c	Pseudo Transport Unit (TUN) that sits on top of BHTP
 *
 * Created:	Fri Aug 25 17:19:15 MDT 2006
 *
 * Authors:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 * This is not a fully featured TUN such as udpex.c and tcp.c.
 * It exists just to export the pad_tun_rcv() operation that
 * consumes the padding from incoming packets.
 *
 * The idea is to treat the padding exactly as any other TUN protocol,
 * instead of a special case in the bhtp_rcv_tun().
 *
 * >>>>>>>>> OBSOLETE: No longer used since padding is not a TUN anymore
 * >>>>>>>>> Tue Apr 15 11:10:39 MDT 2008
 */

/* Must come before log.h. Otherwise we get compiler warning  */
#define LOG_PREFIX "pad: "
#include "log.h"

#include "pad.h"
#include "bhtp_abi.h"

struct pad_state {
	struct bhtp_tun_prot tun_prot;
	struct bhtp_tun_ops tun_ops;
};

int
pad_tun_rcv(struct sk_buff *skb)
{
	DEBUG("tun_rcv: consuming padding TUN %d", bhtp_qsk(skb)->tun_len);
	kfree_skb(skb);
	
	return 0;
	// HERE: try this out to see the discarded probes
}

struct pad_state pad = {

	/* for BHTP registration */
	.tun_prot = {
		.proto		= BHTP_PROTO_PAD,
		.ops		= &pad.tun_ops,
	},

	.tun_ops = {
		.rcv		= pad_tun_rcv,
	},
};


int __init
pad_init(void)
{
	int rc = 0;

	LOG("loading");

	/* register with BHTP */
	rc = bhtp_register_tun_prot(&pad.tun_prot);
	if (rc)
		goto error;
	LOG("registered with BHTP");

	LOG("loaded");

	return 0;

error:
	return rc;
}

void __exit
pad_exit(void)
{
	LOG("unloading");

	bhtp_unregister_tun_prot(&pad.tun_prot);
	LOG("de-registered from BHTP");

	LOG("unloaded");
}

