/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Sun Mar 26 19:36:51 EST 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _BHTP_PROTO_H
#define _BHTP_PROTO_H

#include <net/ip.h>
#include <net/protocol.h>

#include "bhtp.h"
#include "bhtp_abi.h"

// defined in req.h
struct bhtp_flow_req;

struct bhtp_state {
	struct kobject kobj;
	/* FIXME: bhtp_state_init does not initialize everything */
	int done;

	struct net_protocol protocol;
	struct task_struct *sched_thread;

	u32 last_flowid;
	struct list_head flows;

	struct bhtp_tun_prot **tun_prots;

	spinlock_t lock;

	atomic_t seqno;

	atomic_t req_waiting;
	wait_queue_head_t req_wq;
	struct list_head req_list;
	spinlock_t req_lock;
        struct hrtimer req_hrtimer;
	
	atomic_t reqid;

#ifdef MGRP_TSKBSTATS
        struct mgrp_tskbstats *tskbstats;
#endif

	/* module parameters */
	int latency;
	int maxpbk_percent;
	int maxdelay;
        int buckets;
        int policy;
};


extern struct bhtp_state bhtp;

int bhtp_state_init(void);
void bhtp_state_done(void);

ktime_t bhtp_process_req(void);

int mgrp_destroy_all_flows(void);

void mgrp_reset_fragid_in_all_flows(void);
void mgrp_reset_counters_in_all_flows(void);

void bhtp_send_req(struct bhtp_flow_req *req);

static inline struct bhtp_tun_prot *
bhtp_lookup_proto(int proto)
{
	if (proto < 0 || proto > BHTP_PROTO_MAX)
		return NULL;
	
	return bhtp.tun_prots[proto];
}


#endif

