/*
 * Timer requests to send skbs that enter the MGRP buffer if not piggybacked.
 *
 * Wed Aug 27 07:52:53 MDT 2008
 */

#include "req.h"

void
mgrp_drop_req(void)
{
        struct list_head *l = NULL;
        struct list_head *tmp = NULL;
        struct bhtp_flow_req *req = NULL;

        spin_lock_bh(&bhtp.req_lock);
        list_for_each_safe(l, tmp, &bhtp.req_list) {
                req = list_entry(l, struct bhtp_flow_req, list);
                list_del(&req->list);
                atomic_dec(&bhtp.req_waiting);
                LOG("drop_req: DROPPING request, waiting %d",
                        atomic_read(&bhtp.req_waiting));
                kfree(req);
        }
        spin_unlock_bh(&bhtp.req_lock);

        // wait for handler to finish **after you empty the list**
        // in case it was still servicing a request
        MDEBUG("hrtimer_cancel start");
        hrtimer_cancel(&bhtp.req_hrtimer);
        MDEBUG("hrtimer_cancel end");
}

// timer handler for servicing requests
enum hrtimer_restart
mgrp_req_hrtimer(struct hrtimer *timer)
{
        enum hrtimer_restart restart_status = HRTIMER_NORESTART;
        struct bhtp_flow_req *req = NULL;
        struct bhtp_flow_req *req_dequeued = NULL;

        int has_next = 0;
        ktime_t next_expires;

        // do not dequeue yet
        req = mgrp_peek_expired_req();
        if (req == NULL)
                goto done;

        bhtp_send_req(req);

        req_dequeued = mgrp_dequeue_req_and_peek_next(&has_next, &next_expires);
        BUG_ON(req != req_dequeued);
        kfree(req);

        if (has_next) {
                timer->expires = next_expires;
                restart_status = HRTIMER_RESTART;
        }

        // The only place where requests are added is in mgrp_enqueue_req()
        //
        // Since we dequeue only after we have serviced the request we
        // minimize the time that the enqueue function will have to wait
        // for us to finish (it waits if it finds the queue empty).
        //
        // The whole point is that either us or the enqueue function need
        // to schedule the next timer. Needs attention not to miss a timer.
        //
        // The key here is that we dequeue the serviced request and the
        // expiration time of the next request in one atomic step.
        // That way the enqueue function should be certain that if it
        // finds the queue not-empty, then it is this function that
        // will reschedule the timer.
done:
        return restart_status;
}
