/*
 * BHTP		Bottom Half Transport Protocol
 *
 * Created:	Thu Oct 26 13:53:03 MDT 2006
 *
 * Author:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#include "main.h"
#include "sysfs.h"

/* for reference see: */
/* drivers/md/md.c */
/* rdev_sysfs_ops */
/* rdev_ktype */
/* rdev_attr_show */
/* rdev_attr_store: also has example for root only access */

struct bhtp_sysfs_entry {
	struct attribute attr;
	ssize_t (*show)(struct bhtp_state *state, char *);
	ssize_t (*store)(struct bhtp_state *state, const char *, size_t);
};


static ssize_t
bhtp_sysfs_attr_show(struct kobject *kobj, struct attribute *attr, char *page)
{
	struct bhtp_sysfs_entry *entry = container_of(attr, struct bhtp_sysfs_entry, attr);
	struct bhtp_state *state = container_of(kobj, struct bhtp_state, kobj);

	if (!entry->show)
		return -EIO;
	return entry->show(state, page);
}

static ssize_t
bhtp_sysfs_attr_store(struct kobject *kobj, struct attribute *attr,
	      const char *page, size_t length)
{
	struct bhtp_sysfs_entry *entry = container_of(attr, struct bhtp_sysfs_entry, attr);
	struct bhtp_state *state = container_of(kobj, struct bhtp_state, kobj);

	if (!entry->store)
		return -EIO;
/* 	if (!capable(CAP_SYS_ADMIN)) */
/* 		return -EACCES; */
	return entry->store(state, page, length);
}

static ssize_t
bhtp_sysfs_latency_show(struct bhtp_state *state, char *buf)
{
	/* nothing yet */
	printk("show");
	return 0;
}

static ssize_t
bhtp_sysfs_latency_store(struct bhtp_state *state, const char *buf, size_t len)
{
	/* nothing yet */
	printk("store");
	return 0;
}

static void
bhtp_sysfs_release(struct kobject *kobj)
{
	struct bhtp_state *state = container_of(kobj, struct bhtp_state, kobj);
	LOG("sysfs_release: done %d", state->done);
}

static struct bhtp_sysfs_entry bhtp_sysfs_latency =
__ATTR(latency, S_IWUSR, bhtp_sysfs_latency_show, bhtp_sysfs_latency_store);

static struct attribute *bhtp_sysfs_default_attrs[] = {
	&bhtp_sysfs_latency.attr,
};

static struct sysfs_ops bhtp_sysfs_ops = {
	.show		= bhtp_sysfs_attr_show,
	.store		= bhtp_sysfs_attr_store,
};

static struct kobj_type bhtp_sysfs_ktype = {
	.release	= bhtp_sysfs_release,
	.sysfs_ops	= &bhtp_sysfs_ops,
	.default_attrs	= bhtp_sysfs_default_attrs,
};

int
bhtp_sysfs_register(void)
{
	int rc = 0;

	memset(&bhtp.kobj, 0, sizeof(bhtp.kobj));

	bhtp.kobj.parent = &THIS_MODULE->mkobj.kobj;
	bhtp.kobj.ktype = &bhtp_sysfs_ktype;

	rc = kobject_set_name(&bhtp.kobj, "bhtpfoo");
	if (rc)
		return rc;

	rc = kobject_register(&bhtp.kobj);
	if (rc)
		return rc;

	LOG("((((((((()))))))))sysfs_register: success");
	return rc;
}

void
bhtp_sysfs_unregister(void)
{
	kobject_unregister(&bhtp.kobj);
}

