#ifndef _BHTP_TCP_H
#define _BHTP_TCP_H

/* Before including these two files we need to define
 * BHTP_KERNEL so that we can see definitions that are
 * only used when BHTP enabled. This of course means
 * the the kernel must also be compiled with BHTP_KERNEL.
 *
 * We have added CONFIG_TCP_MGRP in Kconfig that appends to
 * CFLAGS_EXTRA the BHTP define. This is for the local
 * compilation of the module.
 */

#if __KERNEL__

/* so that code is included also in kernel includes */
#define MGRP 1

#include <net/tcp.h>
#include <linux/tcp.h>

#include "bhtp_abi.h"

struct tcp_bhtp_sock {
	int setup;
	struct bhtp_flow *bflow;
	int flags;
	struct bhtp_tun tun;
	int fragment;
	int frag_len_max;
};

extern struct inet_connection_sock_af_ops ipv4_specific;
extern struct inet_connection_sock_af_ops ipv4_specific_mgrp;

static inline const struct inet_connection_sock_af_ops *
tcp_orig_icsk_af_ops(struct sock *sk)
{
	return tcp_sk(sk)->mgrp.orig_icsk_af_ops;
}

static inline void
tcp_bhtp_frag_state_debug(struct tcp_bhtp_sock *tbs)
{
	DEBUGFN("tcp_bhtp fragmentation %s len %d",
		(tbs->fragment ? "ON" : "OFF"),
		tbs->frag_len_max);
}

// XXX HACK to fix BUGS/tcp-packets-marked-as-mgrp/README
static inline int
mgrp_looks_like_tcp(struct sk_buff *skb) {
	struct tcphdr *th = (struct tcphdr *)skb->transport_header;
        int tcphdr_len = (4 * th->doff);

        if (skb->transport_header != skb->data) {
                ERRFN("transport header not at top of skb");
                return 0;
        }
        // good enough
        return (skb->len == tcphdr_len);
}

static inline void
mgrp_sendto_tcp_directly(struct sk_buff *skb)
{
        int rc = tcp_v4_rcv(skb);
        if (rc) {
                ERRFN("dropping tcp_direct packet len %d", skb->len);
                kfree_skb(skb);
        }
}

#else

/* FIXME: this is a duplicate entry; normally defined in include/linux/tcp.h */
/* 	  To be used in setsockopt() from user-space to turn TCP over MGRP */
/* 	  (see tsocket.c) */
#define TCP_MGRP	99

#endif // __KERNEL

/* for do_tcp_bhtp_setsockopt() and do_tcp_bhtp_getsockopt() */
#define TCP_MGRP_FRAGMENT		1
#define TCP_MGRP_FRAG_LEN_MAX		2

#endif

