#ifndef _MGRP_TSKBSTATS_H
#define _MGRP_TSKBSTATS_H

// control the size of the global structure bhtp.tskbstats
// that is allocated at runtime using vmalloc()
// allocate enough to store all tcp packets for one emulab experiment run
#define MGRP_TSKBSTATS_MAX  500000

#define MGRP_TSKBCODE_QUEUED    'q'
#define MGRP_TSKBCODE_PROBENOW  'p'
#define MGRP_TSKBCODE_SENDNOW   's'

struct mgrp_tskbstats_entry {
        struct timeval ts;
        int skb_len;
        ktime_t latency;
        struct bhtp_qskb qskb;
        char code;
};

struct mgrp_tskbstats {
        spinlock_t lock;
        int count;
        int overflow;
        struct mgrp_tskbstats_entry entries[MGRP_TSKBSTATS_MAX];
};

static inline void
mgrp_tskbstats_init(struct mgrp_tskbstats *stats)
{
        memset(stats, 0, sizeof(struct mgrp_tskbstats));
	spin_lock_init(&stats->lock);
}

static inline void
mgrp_tskbstats_reset(struct mgrp_tskbstats *stats)
{
        stats->count = 0;
        stats->overflow = 0;
}

static inline int
mgrp_tskbstats_oneliner_header(char *str, int len)
{
        return snprintf(str, len,
                        "#timestamp "
                        "code "
                        "len "
                        "TUN "
                        "len "
                        "fragmax "
                        "hdr "
                        "dat "
                        "pad "
                        "gap "
                        "WAITS "
                        "latency");

}

static inline int
mgrp_tskbstats_oneliner(char *str, int len, struct mgrp_tskbstats_entry *entry)
{
        return snprintf(str, len,
                        "%ld.%.6ld %c %d tun %d %d %d %d %d %d waits %lld",
                        entry->ts.tv_sec,
                        entry->ts.tv_usec,
                        entry->code,
                        entry->skb_len,
                        entry->qskb.tun_len,
                        entry->qskb.tun_frag_len_max,
                        entry->qskb.tun_hdr,
                        entry->qskb.tun_data,
                        entry->qskb.tun_pad,
                        entry->qskb.tun_gap,
                        ktime_to_ns(entry->latency));
}

static inline void
mgrp_tskbstats_add(struct mgrp_tskbstats *stats,
                   struct mgrp_tskbstats_entry *entry)
{
        struct mgrp_tskbstats_entry *new_entry = &stats->entries[stats->count];

        if (stats->count >= MGRP_TSKBSTATS_MAX) {
                stats->overflow++;
                if (stats->overflow % 1000 == 0)
                        ERRFN("mgrp_tskbstats is full; entries overflowed %d",
                              stats->overflow);
                return;
        }

        *new_entry = *entry;
        stats->count++;
}
#endif
