/*
 * UDPex	Extended UDP, Transport unit built on top of BHTP
 *
 * Created:	Thu Aug 18 19:15:52 EDT 2005
 *
 * Version:	$Id: bhtp.h,v 1.3 2005/10/15 19:26:18 pavlos Exp $
 *
 * Authors:	Pavlos Papageorgiou <pavlos@eng.umd.edu>
 *
 */

#ifndef _UDPEX_H
#define _UDPEX_H

#if __KERNEL__

#include <linux/in.h>
#include <linux/net.h>
#include <linux/ioctl.h>
#include <linux/skbuff.h>

#else

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/types.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <unistd.h>
/* #include <netinet/ip.h> */

#endif /* __KERNEL__ */

#include "bhtp.h"

#ifdef __cplusplus
extern "C" {
#endif

/* do_udpex_setsockopt() */
/* bhtp_recv_timestamp() */
/* udpex_recvmsg and udpex_sendmsg ancillary data */
#define UDPEX_TSTAMP		1
#define UDPEX_PROBE		2
#define UDPEX_PROBE_RESULTS	3

#define UDPEX_PIGGYBACKED(uh) \
	((uh)->plen > (uh)->len)

#define UDPEX_PADDED(uh) \
	((uh)->plen > (uh)->ulen)

#define UDPEX_WASTED_PADDING(uh) \
	((uh)->len - (uh)->ulen)

struct udpexhdr {
	__u16	source;
	__u16	dest;
/* 	__u16	len; */
/* 	__u16	check; */
/* 	__u16	plen; */
/* 	__u16	ulen; */
};

/* ancillary data for udpex_sendmsg (SOL_MGRP, UDPEX_PROBE) */
struct udpex_probe {
	int barrier;
	int gap_usec;
	int pad;
	/* make sense only when barrier = 0 */
	int can_wait_for_usec;
	int tell_results;
        int piggyback_not_allowed;
};

struct udpex_probe_results {
	int sent;
	/* all durations in nanosec */
	/* duration barrier was down: repeated user sends */
	long long t1;
	/* delay after barrier was lifted and probes sent: waiting for BHTP */
	long long t2;
	/* time it took to send all probes once ready to send */
	long long t3;

        // total packets/bytes sent
        // (counts everything below IP)
        int total_pkts;
        int total_bytes;

        // packets/bytes that were piggybacked
        // (counts portion of padding that was reused)
        int pbk_pkts;
        int pbk_bytes;

        // packets/bytes that were piggybacked and that would
        // have gone in between probes if there was no pbk
        int pbk_inwin_pkts;
        int pbk_inwin_bytes;

        // actual start/end time of the sent probe train (nsec)
        long long send_start;
        long long send_end;

        // estimated duration of probe train (before sending) (nsec)
        long long send_precomputed_duration;
};

static inline char *
udpex_probe_results_str(char *str, int strlen, struct udpex_probe_results *res)
{
	snprintf(str, strlen,
                "sent %d t1 %8lld t2 %8lld t3 %8lld "

                "\n\t"
                "total %d %d "
                "pbk %d %d "
                "pbk_inwin %d %d "

                "\n\t"
                "duration actual %lld precomputed %lld ",

		res->sent,
		res->t1,
		res->t2,
		res->t3,

                res->total_pkts,
                res->total_bytes,
                res->pbk_pkts,
                res->pbk_bytes,
                res->pbk_inwin_pkts,
                res->pbk_inwin_bytes,

                res->send_end - res->send_start,
                res->send_precomputed_duration);
		
	return str;
}

#define UDPEX_IOC_MAGIC		'h'

#define UDPEX_IOC_ISACTIVE	_IO(UDPEX_IOC_MAGIC, 0x80)
#define UDPEX_IOC_PROBE		_IO(UDPEX_IOC_MAGIC, 0x81)
#define UDPEX_IOC_QUEUE		_IO(UDPEX_IOC_MAGIC, 0x82)
#define UDPEX_IOC_PROBE_RESULTS _IOW(UDPEX_IOC_MAGIC, 0x83, struct udpex_probe_results)

#if __KERNEL__
static inline void
udpex_probe_print(struct udpex_probe *p)
{
	LOG("udpex-probe: barrier %d gap_usec %d pad %d",
		p->barrier,
		p->gap_usec,
		p->pad);
}
#else
static inline void
udpex_probe_print(struct udpex_probe *p)
{
	printf("udpex-probe: barrier %d gap_usec %d pad %d\n",
		p->barrier,
		p->gap_usec,
		p->pad);
}
#endif

#if __KERNEL__

struct udpex_options {
	/* module parameters */
	int probe_show;
	int probe_timeout;
	int probe_accumulated;
	// prints detailed debug for every UDP packet
	// dropped due to port not found
	int debug_port_unreachable;
};

extern struct udpex_options udpex_op;

static inline int
udpex_sane_header(struct udpexhdr *uh)
{
/* 	if (unlikely(uh->ulen < sizeof(struct udpexhdr))) */
/* 		return 0; */
/*  */
/* 	if (unlikely(uh->ulen > uh->len)) */
/* 		return 0; */
/* 	 */
/* 	if (unlikely(uh->len > uh->plen)) */
/* 		return 0; */

	return 1;
}

#define UDPEX_QSKB_MAGIC	0xbf

struct udpex_qskb {
	short magic;
};

static inline int
udpex_qsk_valid(struct udpex_qskb *qskb) {
	return (qskb->magic == UDPEX_QSKB_MAGIC);
}

static inline struct udpex_qskb *
udpex_qsk_novalidate(struct sk_buff *skb) {
	struct udpex_qskb *qskb = (struct udpex_qskb *)skb->cb;
	BUG_ON(sizeof(struct udpex_qskb) > sizeof(skb->cb));
	return qskb;
}

static inline struct udpex_qskb *
udpex_qsk(struct sk_buff *skb) {
	struct udpex_qskb *qskb = udpex_qsk_novalidate(skb);
	BUG_ON(!udpex_qsk_valid(qskb));
	return qskb;
}

static inline struct udpex_qskb *
udpex_qsk_init(struct sk_buff *skb) {
	struct udpex_qskb *qskb = udpex_qsk_novalidate(skb);
	memset(qskb, 0, sizeof(*qskb));
	qskb->magic = UDPEX_QSKB_MAGIC;
	return qskb;
}

#else

static inline void
udpex_print_hdr(struct udpexhdr *uh)
{
	printf("%hu > %hu", ntohs(uh->source), ntohs(uh->dest));
}

static inline int
udpex_socket(int sock)
{
	int rc = 0;

	if (sock < 0)
		return 0;

	rc = ioctl(sock, UDPEX_IOC_ISACTIVE);
	return (rc == 0);
}

static inline int
udpex_is_active()
{
	int sock = 0;
	int is_active = 0;

	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0)
		return 0;

	is_active = udpex_socket(sock);
	close(sock);

	return is_active;
}

static inline int
udpex_ioctl(int cmd)
{
	int sock = 0;
	int rc = 0;


	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP);
	if (sock < 0)
		return sock;

	rc = ioctl(sock, cmd);
	close(sock);

	return rc;
}


#ifdef __cplusplus
}
#endif    

#endif /* __KERNEL__ */

#endif /* _UDPEX_H */

