#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <poll.h>
#include <unistd.h>
#include <errno.h>

#include "probe.h"
#include "mgrp/mgrp.h"

#define SOURCE_PORT 4200
#define DEST_PORT 4201
#define PROBE_LEN 400
#define HEADER_LEN 42

struct sockaddr_in destination;

void setTarget(const char *dstaddr)
{
}


void sendProbe(int sd, int probe_seq, int n_probes)
{
    char *buffer = calloc(1, PROBE_LEN);
    int *intbuf = (int *)buffer;
    int i;
    intbuf[0] = probe_seq;
    
    for (i = 0; i < n_probes; ++i)
    {
        int order = (i 
                     ? ( (i == n_probes - 1) ? PROBE_LAST : PROBE_MIDDLE)
                     : PROBE_LAST);
        int len = HEADER_LEN;
        int pad = PROBE_LEN - len;
        int gap = 0;
        int sbytes = 0;
        fprintf(stderr, "probe_seq=%d n_probes=%d i=%d, order=%d gap=%d, buffer=%p, len=%d, pad=%d, pbk=%d\n",
                probe_seq, n_probes, i, order, gap, buffer, len, pad, PROBE_PBK_ON
                );
        sbytes = probe_send_train_pkt(sd, 
                                      order, 
                                      gap, 
                                      buffer, 
                                      len,
                                      pad,
                                      PROBE_PBK_ON
                                      );
        fprintf(stderr, "sbytes=%d\n", sbytes);
        if (sbytes < 0) return;
    }
}

void sender(char *dstaddr)
{
    int sd = -1;
    int probe_sequence = 0;
    int n_probes = 3;
    struct sockaddr_in source;
    int i = 0;

    memset(&destination, 0, sizeof(struct sockaddr_in));
    destination.sin_family = AF_INET;
    destination.sin_port = htons(DEST_PORT);
    
    struct hostent *hp = gethostbyname2(dstaddr, AF_INET);
    if (hp) 
    {
        memcpy(&destination.sin_addr, hp->h_addr_list[0], hp->h_length);
    } else {
        fprintf(stderr, "gethostbyname2(%s) call failed: %d/%s\n", dstaddr, errno, strerror(errno));
    }

    printf("Target set to %s, %x:%hu\n\n",
           dstaddr, destination.sin_addr.s_addr, destination.sin_port
           );

    sd = mgrp_socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP, NULL);
    if (sd == -1) {
        printf("socket call failed: %d/%s\n", errno, strerror(errno));
        return;
    }

    memset(&source, 0, sizeof(struct sockaddr_in));
    source.sin_family = AF_INET;
    source.sin_port = htons(SOURCE_PORT);
    if (0 != (bind(sd, (const struct sockaddr*)&source, sizeof(struct sockaddr_in)))) {
        printf("bind() failed: %d/%s\n", errno, strerror(errno));
        return;
    }

    if (0 != (connect(sd, (struct sockaddr *)&destination, sizeof(destination)))) {
        printf("connect() failed: %d/%s\n", errno, strerror(errno));
        return;
    }

    for (i = 0; i < 10; i++)
    {
        sendProbe(sd, probe_sequence, n_probes);
        ++probe_sequence;
    }
    return;
}

    
void receiver()
{

    int sd = -1;
    struct pollfd pfd[1]; 

    int on = 1;
    int rv = -1;

    char buffer[PROBE_LEN];

    sd = mgrp_socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP, NULL);
    if (sd == -1) {
        fprintf(stderr, "socket call failed: %d/%s\n", errno, strerror(errno));
        goto err;
    }

    rv = setsockopt(sd, SOL_MGRP, UDPEX_TSTAMP, &on, sizeof(on));
    if (-1 == rv) {
	fprintf(stderr, "error in setsockopt(SOL_MGRP, UDPEX_TSTAMP): %d/%s", errno, strerror(errno));
	goto err;
    }

    struct sockaddr_in sin;
    memset(&sin, 0, sizeof(struct sockaddr_in));
    sin.sin_family = AF_INET;
    sin.sin_port = htons(DEST_PORT);
    if (0 != (bind(sd, (const struct sockaddr*)&sin, sizeof(struct sockaddr_in)))) {
        fprintf(stderr, "bind() failed: %d/%s\n", errno, strerror(errno));
        goto err;
    }

    while (1)
    {
        /*** see tagent/tools/pathload/pathload_recv_func.c:bhtprecv() ***/
        ssize_t rbytes;
        
        struct bhtp_timestamps bts;
        struct timeval tv;

        struct timeval tv2;

        fprintf(stderr, "Calling probe_recv_train_pkt()\n");
        rbytes = probe_recv_train_pkt(sd, buffer, PROBE_LEN, &bts);
        if (rbytes < HEADER_LEN)
        {
            fprintf(stderr, "error on receive - incomplete probe ( %d/%s", errno, strerror(errno));
            exit(-1);
        }
        fprintf(stderr, "Received %d bytes\n", rbytes);
        
        bhtp_tstamp_to_timeval(&bts.rcv, &tv);
        bhtp_tstamp_to_timeval(&bts.snd, &tv2);
        /*
         * really simple output format.  we don't do any processing
         * in the receiver - just print stuff out.  all estimation
         * is done in an external script.
         */
        printf("Recv'd packet %d at %u.%6u, %u.%6u\n",
               ((int*)buffer)[0], (unsigned int)tv.tv_sec, (unsigned int)tv.tv_usec,
               (unsigned int)tv2.tv_sec, (unsigned int)tv2.tv_usec);
    }

 err:
    close (sd);
    return;
}

void usage() {
    printf("program [r | s destination]\n\n");
    exit(-1);
}

int main(int argc, char **argv) {
    char *which = argv[1];
    if (argc < 2) usage();
    if (strcmp(which, "s") == 0){
        printf("SENDER");
        if (argc != 3) usage();
        sender(argv[2]);
    } else if (strcmp(which, "r") == 0){
        printf("RECEIVER");
        if (argc != 2) usage();
        receiver();
    } else {
        printf("Brokie %s!", argv[1]);
        usage();
    }
    return 0;
}
