/*************************************************************************

 distrib.cpp
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#include <iostream>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#include "distrib.h"
#include "utils.h"

using namespace std;

// the Distrib abstract class

Distrib::~Distrib()
{
}

int Distrib::type() { return 0; }
void Distrib::init() {}
void Distrib::nextState() {};
void Distrib::setSendTime( const struct timeval &tv )
{
	nextSend = tv;
};


// fix distribution

int Fix::type()
{
	return FIX;
}

void Fix::init()
{
	nextSize = size;
	period = ( unsigned long )( size * ( 8000000 / ( float )rate ) );  // inter packet time in us
	gettimeofday( &nextSend, NULL );
	nextSend.tv_sec++;
	nextSend.tv_usec = rand()%1000000;  // randomize the us value
}

void Fix::nextState()
{
		addTime( nextSend, period );  // increase nextSend with the period
}

int OnOff::type()
{
	return ONOFF;
}

void OnOff::init()
{
	gettimeofday( &nextSend, NULL );
	nextSend.tv_sec++;
	nextSend.tv_usec = rand()%1000000;
	nextSize = size1;
	start = nextSend;
	state = 1;
	if( rate1 == 0 ) period1 = 0;
	else period1 = ( unsigned long )( size1 * ( 8000000 / ( float )rate1 ) );  // in us
	if( rate2 == 0 ) period2 = 0;
	else period2 = ( unsigned long )( size2 * ( 8000000 / ( float )rate2 ) );  // in us
	if( time2 == 0 ) inf = -1;
	else inf = 0;
	if( period1 == 0 )  // we have to skip the first state
	{
		addTime( start, time1 );
		nextSend = start;
		nextSize = size2;
		state = 2;
	}
}
	
void OnOff::nextState()
{
	// calculate the next sending time

	if( state == 1 )
	{
		if( period1 == 0 )  // skip sending for time1 us
		{
			 addTime( start, time1 );
			 nextSend = start;
			 state = 2;
		}
		else
		{
			addTime( nextSend, period1 );
			if( ( unsigned long )timeDiff( nextSend, start ) > time1 )
			{
				state = 2;  // we have to transit to the second state
				addTime( start, time1 );
			}
		}
	}
	else
	{
		if( period2 == 0 )  // skip sending for time2 us
		{
			 addTime( start, time2 );
			 nextSend = start;
			 state = 1;
		}
		else
		{
			addTime( nextSend, period2 );
			if( !inf && ( ( unsigned long )timeDiff( nextSend, start ) > time2  ) )
			{
				state = 1;  // we have to transit to the first state
				addTime( start, time2 );
			}
		}
	}

	// calculate the next packet size

	if( state == 1 ) nextSize = size1;
	else nextSize = size2;
}

// the Poisson class
 
int Poisson::type()
{
	return POISSON;
}

void Poisson::init(bool deterministic)
{
	diff = 0;
	gettimeofday( &nextSend, NULL );
	if ( deterministic ) {
	  srand( 0 );
	} else {
	  srand( nextSend.tv_usec );
	}
	nextSend.tv_sec++;
	nextSend.tv_usec = rand()%1000000;  // randomize the us value
	nextSize = size;
}

void Poisson::nextState()
{
	diff = ( long )( -log( ( float )rand()/( float )RAND_MAX )*( float )avgint );
	addTime( nextSend, diff );
};
