/*************************************************************************

 nemu.cpp
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <netinet/in.h>
#include <netinet/tcp.h> 	
#include <netinet/if_ether.h>
#include <netpacket/packet.h>
#include <errno.h>

#include "defs.h"
#include "utils.h"
#include "impairs.h"

using namespace std;

// function declarations

void usage();

// the main function

int main( int argc, char **argv )
{
	double closs1 = 0;  // continuous loss
	double bloss1 = 0;  // bursty loss
	long blossl1 = 0;  // bursty loss length
	long blossp1 = 0;  // bursty loss period
	long blossm1 = 0;  // bursty loss max dropped frames
	long delay1 = 0;  // delay
	long cjitter1 = 0;  // continuous jitter
	long jitter1 = 0;  // jitter
	long jitteru1 = 0; // jitter up
	long jitterk1 = 0; // jitter keep
	double jitterl1 = 0; // jitter loss
	long jitterd1 = 0; // jitter down
	long jitterp1 = 0;  // jitter period
	char *filejitter1 = 0;
	double closs2 = 0;  // continuous loss
	double bloss2 = 0;  // bursty loss
	long blossl2 = 0;  // bursty loss length
	long blossp2 = 0;  // bursty loss period
	long blossm2 = 0;  // bursty loss max dropped frames
	long delay2 = 0;  // delay
	long cjitter2 = 0;  // continuous jitter
	long jitter2 = 0;  // jitter
	long jitteru2 = 0; // jitter up
	long jitterk2 = 0; // jitter keep
	double jitterl2 = 0; // jitter loss
	long jitterd2 = 0; // jitter down
	long jitterp2 = 0;  // jitter period
	char *filejitter2 = 0;
	int par_reset, par_cnt, par_clear, par_get;
	int par_closs1, par_bloss1, par_blossl1, par_blossp1, par_blossm1, par_cjitter1, par_filejitter1;
	int par_delay1, par_jitter1, par_jitteru1, par_jitterk1, par_jitterd1, par_jitterl1, par_jitterp1;
	int par_closs2, par_bloss2, par_blossl2, par_blossp2, par_blossm2, par_cjitter2, par_filejitter2;
	int par_delay2, par_jitter2, par_jitterl2, par_jitteru2, par_jitterk2, par_jitterd2, par_jitterp2;
	int smid;
	Impairs *impairs;
	
	cerr << "nemu " << VERSION << ", (C) 2006, Norbert Vegh\n";

	// process the command line arguments
	
	if( argc < 1 ) usage();
	par_reset = par_cnt = par_clear = par_get = 0;
	par_closs1 = par_bloss1 = par_blossl1 = par_blossp1 = par_blossm1 = par_delay1 = par_cjitter1 = 0;
	par_jitter1 = par_jitterl1 = par_jitteru1 = par_jitterk1 = par_jitterd1 = par_jitterp1 = par_filejitter1 = 0;
	par_closs2 = par_bloss2 = par_blossl2 = par_blossp2 = par_blossm2 = par_delay2 = par_cjitter2 = 0;
	par_jitter2 = par_jitterl2 = par_jitteru2 = par_jitterk2 = par_jitterd2 = par_jitterp2 = par_filejitter2 = 0;
	blossm1 = blossm2 = 0;
	
	while( argc > 1 )
	{
		if( !strcmp( "-reset", argv[1] ) )
		{
			par_reset = -1;
			argc--; argv++;
		}
		else if( !strcmp( "-cnt", argv[1] ) )
		{
			par_cnt = -1;
			argc--; argv++;
		}
		else if( !strcmp( "-clear", argv[1] ) )
		{
			par_clear = -1;
			argc--; argv++;
		}
		else if( !strcmp( "-get", argv[1] ) )
		{
			par_get = -1;
			argc--; argv++;
		}
		else if( !strcmp( "-1closs", argv[1] ) )
		{
			if( par_closs1 ) configError( "1closs parameter is duplicated" );
			if( argc == 1 ) usage();
			closs1 = atof( argv[2] );
			if( closs1 < 0 ) configError( "Invalid 1closs\n" );
			par_closs1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1bloss", argv[1] ) )
		{
			if( par_bloss1 ) configError( "1bloss parameter is duplicated" );
			if( argc == 1 ) usage();
			bloss1 = atof( argv[2] );
			if( bloss1 < 0 ) configError( "Invalid 1bloss\n" );
			par_bloss1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1blossl", argv[1] ) )
		{
			if( par_blossl1 ) configError( "1bloss parameter is duplicated" );
			if( argc == 1 ) usage();
			blossl1 = readTime( argv[2] );
			if( blossl1 < 0 ) configError( "Invalid 1bloss\n" );
			par_blossl1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1blossp", argv[1] ) )
		{
			if( par_blossp1 ) configError( "1blossp parameter is duplicated" );
			if( argc == 1 ) usage();
			blossp1 = readTime( argv[2] );
			if( blossp1 < 0 ) configError( "Invalid 1blossp\n" );
			par_blossp1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1blossm", argv[1] ) )
		{
			if( par_blossm1 ) configError( "1blossm parameter is duplicated" );
			if( argc == 1 ) usage();
			blossm1 = atoi( argv[2] );
			if( blossm1 < 0 ) configError( "Invalid 1blossm\n" );
			par_blossm1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1delay", argv[1] ) )
		{
			if( par_delay1 ) configError( "1delay parameter is duplicated" );
			if( argc == 1 ) usage();
			delay1 = readTime( argv[2] );
			if( delay1 < 0 ) configError( "Invalid 1delay\n" );
			par_delay1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1cjitter", argv[1] ) )
		{
			if( par_cjitter1 ) configError( "1cjitter parameter is duplicated" );
			if( argc == 1 ) usage();
			cjitter1 = readTime( argv[2] );
			if( cjitter1 < 0 ) configError( "Invalid 1cjitter\n" );
			par_cjitter1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1filejitter", argv[1] ) )
		{
			if( par_filejitter1 ) configError( "1filejitter parameter is duplicated" );
			if( argc == 1 ) usage();
			filejitter1 = argv[2];
			par_filejitter1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1jitter", argv[1] ) )
		{
			if( par_jitter1 ) configError( "1jitter parameter is duplicated" );
			if( argc == 1 ) usage();
			jitter1 = readTime( argv[2] );
			if( jitter1 < 0 ) configError( "Invalid 1jitter\n" );
			par_jitter1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1jitteru", argv[1] ) )
		{
			if( par_jitteru1 ) configError( "1jitteru parameter is duplicated" );
			if( argc == 1 ) usage();
			jitteru1 = readTime( argv[2] );
			if( jitteru1 < 0 ) configError( "Invalid 1jitteru\n" );
			par_jitteru1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1jitterk", argv[1] ) )
		{
			if( par_jitterk1 ) configError( "1jitterk parameter is duplicated" );
			if( argc == 1 ) usage();
			jitterk1 = readTime( argv[2] );
			if( jitterk1 < 0 ) configError( "Invalid 1jitterk\n" );
			par_jitterk1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1jitterl", argv[1] ) )
		{
			if( par_jitterl1 ) configError( "1jitterl parameter is duplicated" );
			if( argc == 1 ) usage();
			jitterl1 = atof( argv[2] );
			if( jitterl1 < 0 ) configError( "Invalid 1jitterl\n" );
			par_jitterl1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1jitterd", argv[1] ) )
		{
			if( par_jitterd1 ) configError( "1jitterd parameter is duplicated" );
			if( argc == 1 ) usage();
			jitterd1 = readTime( argv[2] );
			if( jitterd1 < 0 ) configError( "Invalid 1jitterd\n" );
			par_jitterd1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-1jitterp", argv[1] ) )
		{
			if( par_jitterp1 ) configError( "1jitterp parameter is duplicated" );
			if( argc == 1 ) usage();
			jitterp1 = readTime( argv[2] );
			if( jitterp1 < 0 ) configError( "Invalid 1jitterp\n" );
			par_jitterp1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2closs", argv[1] ) )
		{
			if( par_closs2 ) configError( "2closs parameter is duplicated" );
			if( argc == 1 ) usage();
			closs2 = atof( argv[2] );
			if( closs2 < 0 ) configError( "Invalid 2closs\n" );
			par_closs2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2bloss", argv[1] ) )
		{
			if( par_bloss2 ) configError( "2bloss parameter is duplicated" );
			if( argc == 1 ) usage();
			bloss2 = atof( argv[2] );
			if( bloss2 < 0 ) configError( "Invalid 2bloss\n" );
			par_bloss2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2blossl", argv[1] ) )
		{
			if( par_blossl2 ) configError( "2blossl parameter is duplicated" );
			if( argc == 1 ) usage();
			blossl2 = readTime( argv[2] );
			if( blossl2 < 0 ) configError( "Invalid 2blossl\n" );
			par_blossl2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2blossp", argv[1] ) )
		{
			if( par_blossp2 ) configError( "2blossp parameter is duplicated" );
			if( argc == 1 ) usage();
			blossp2 = readTime( argv[2] );
			if( blossp2 < 0 ) configError( "Invalid 2blossp\n" );
			par_blossp2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2blossm", argv[1] ) )
		{
			if( par_blossm2 ) configError( "2blossm parameter is duplicated" );
			if( argc == 1 ) usage();
			blossm2 = atoi( argv[2] );
			if( blossm2 < 0 ) configError( "Invalid 2blossm\n" );
			par_blossm2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2delay", argv[1] ) )
		{
			if( par_delay2 ) configError( "2delay parameter is duplicated" );
			if( argc == 1 ) usage();
			delay2 = readTime( argv[2] );
			if( delay2 < 0 ) configError( "Invalid 2delay\n" );
			par_delay2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2cjitter", argv[1] ) )
		{
			if( par_cjitter2 ) configError( "2cjitter parameter is duplicated" );
			if( argc == 1 ) usage();
			cjitter2 = readTime( argv[2] );
			if( cjitter2 < 0 ) configError( "Invalid 2cjitter\n" );
			par_cjitter2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2filejitter", argv[1] ) )
		{
			if( par_filejitter2 ) configError( "2filejitter parameter is duplicated" );
			if( argc == 1 ) usage();
			filejitter2 = argv[2];
			par_filejitter2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2jitter", argv[1] ) )
		{
			if( par_jitter2 ) configError( "2jitter parameter is duplicated" );
			if( argc == 1 ) usage();
			jitter2 = readTime( argv[2] );
			if( jitter2 < 0 ) configError( "Invalid 2jitter\n" );
			par_jitter2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2jitteru", argv[1] ) )
		{
			if( par_jitteru2 ) configError( "2jitteru parameter is duplicated" );
			if( argc == 1 ) usage();
			jitteru2 = readTime( argv[2] );
			if( jitteru2 < 0 ) configError( "Invalid 2jitteru\n" );
			par_jitteru2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2jitterk", argv[1] ) )
		{
			if( par_jitterk2 ) configError( "2jitterk parameter is duplicated" );
			if( argc == 1 ) usage();
			jitterk2 = readTime( argv[2] );
			if( jitterk2 < 0 ) configError( "Invalid 2jitterk\n" );
			par_jitterk2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2jitterl", argv[1] ) )
		{
			if( par_jitterl2 ) configError( "2jitterl parameter is duplicated" );
			if( argc == 1 ) usage();
			jitterl2 = atof( argv[2] );
			if( jitterl2 < 0 ) configError( "Invalid 2jitterl\n" );
			par_jitterl2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2jitterd", argv[1] ) )
		{
			if( par_jitterd2 ) configError( "2jitterd parameter is duplicated" );
			if( argc == 1 ) usage();
			jitterd2 = readTime( argv[2] );
			if( jitterd2 < 0 ) configError( "Invalid 2jitterd\n" );
			par_jitterd2 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2jitterp", argv[1] ) )
		{
			if( par_jitterp2 ) configError( "2jitterp parameter is duplicated" );
			if( argc == 1 ) usage();
			jitterp2 = readTime( argv[2] );
			if( jitterp2 < 0 ) configError( "Invalid 2jitterp\n" );
			par_jitterp2 = -1;
			argc -= 2; argv += 2;
		}
		else
		{
			cerr << "Invalid option: " << argv[1] << endl;
			return -1;
		}
	}
	
	// check parameters

	if( par_bloss1 && par_blossl1 && par_blossp1 ) {}
	else if( !par_bloss1 && !par_blossl1 && !par_blossp1 ) {}
	else
	{
		configError( "1bloss, 1blossl and 1blossp paramteres must be specified together\n" );
	}
	if( par_jitter1 && par_jitterl1 && par_jitterp1 ) {}
	else if( !par_jitter1 && !par_jitterl1 && !par_jitterp1 ) {}
	else
	{
		configError( "1jitter, 1jitterl and 1jitterp paramteres must be specified together\n" );
	}
	if( par_bloss2 && par_blossl2 && par_blossp2 ) {}
	else if( !par_bloss2 && !par_blossl2 && !par_blossp2 ) {}
	else
	{
		configError( "2bloss, 2blossl and 2blossp paramteres must be specified together\n" );
	}
	if( par_jitter2 && par_jitterl2 && par_jitterp2 ) {}
	else if( !par_jitter2 && !par_jitterl2 && !par_jitterp2 ) {}
	else
	{
		configError( "2jitter, 2jitterl and 2jitterp paramteres must be specified together\n" );
	}

	// create shared memory and initalize it
	
	smid = shmget( SMKEY, sizeof( Impairs ), 0 );
	if( smid == -1 )
	{
		perror( "Cannot get shared memory" );
		return -1;
	}
	impairs = ( Impairs * )shmat( smid, NULL, 0 );
	
	// now set the new values
	
	if( par_cnt )
	{
		cout << "1->2 Forwarded: " << impairs->total1-impairs->dropped1;
		cout << ", dropped: " << impairs->dropped1 << endl;
		cout << "2->1 Forwarded: " << impairs->total2-impairs->dropped2;
		cout << ", dropped: " << impairs->dropped2 << endl;
	}
	if( par_clear )
	{
		impairs->total1 = impairs->total2 = impairs->dropped1 = impairs->dropped2 = 0;
	}
	if( par_get )
	{
		cout << "1->2 delay: " << impairs->delay1 << endl;
		impairs->closs1.Dump( "1->2" );
		impairs->bloss1.Dump( "1->2" );
		impairs->jitter1.Dump( "1->2" );
		impairs->filejitter1.Dump( "1->2" );
		cout << "2->1 delay: " << impairs->delay2 << endl;
		impairs->closs2.Dump( "2->1" );
		impairs->bloss2.Dump( "2->1" );
		impairs->jitter2.Dump( "2->1" );
		impairs->filejitter2.Dump( "2->1" );
	}
	if( par_reset )
	{
		impairs->Init();
	}
	if( par_delay1 ) impairs->delay1 = delay1;
	if( par_delay2 ) impairs->delay2 = delay2;
	if( par_closs1 ) impairs->closs1.Init( closs1 );
	if( par_closs2 ) impairs->closs2.Init( closs2 );
	if( par_bloss1 ) impairs->bloss1.Init( bloss1, blossl1, blossp1, blossm1 );
	if( par_bloss2 ) impairs->bloss2.Init( bloss2, blossl2, blossp2, blossm2 );
	if( par_cjitter1 ) impairs->cjitter1.Init( cjitter1 );
	if( par_cjitter2 ) impairs->cjitter2.Init( cjitter2 );
	if( par_filejitter1 ) impairs->filejitter1.Init( filejitter1 );
	if( par_filejitter2 ) impairs->filejitter2.Init( filejitter2 );
	if( par_jitter1 ) impairs->jitter1.Init( jitter1, jitteru1, jitterk1, jitterl1, jitterd1, jitterp1 );
	if( par_jitter2 ) impairs->jitter2.Init( jitter2, jitteru2, jitterk2, jitterl2, jitterd2, jitterp2 );
}


void usage()
{
	cerr << "Command line arguments:\n";
	cerr << "[-reset]\n";
	cerr << "[-clear]\n";
	cerr << "[-cnt]\n";
	cerr << "[-get]\n";
	cerr << "[-1closs continuous_loss]\n";
	cerr << "[-1bloss bursty_loss]\n";
	cerr << "[-1blossm bursty_loss_max]\n";
	cerr << "[-1blossl bursty_loss_lenght]\n";
	cerr << "[-1blossp bursty_loss_period]\n";
	cerr << "[-1delay delay]\n";
	cerr << "[-1cjitter continuous_jitter]\n";
	cerr << "[-1filejitter file]\n";
	cerr << "[-1jitter jitter]\n";
	cerr << "[-1jitterl jitter_loss]\n";
	cerr << "[-1jitteru jitter_up]\n";
	cerr << "[-1jitterk jitter_keep]\n";
	cerr << "[-1jitterd jitter_down]\n";
	cerr << "[-1jitterp jitter_period]\n";
	cerr << "[-2closs continuous_loss]\n";
	cerr << "[-2bloss bursty_loss]\n";
	cerr << "[-2blossm bursty_loss_max]\n";
	cerr << "[-2blossl bursty_loss_lenght]\n";
	cerr << "[-2blossp bursty_loss_period]\n";
	cerr << "[-2delay delay]\n";
	cerr << "[-2cjitter continuous_jitter]\n";
	cerr << "[-2filejitter file]\n";
	cerr << "[-2jitter jitter]\n";
	cerr << "[-2jitterl jitter_loss]\n";
	cerr << "[-2jitteru jitter_up]\n";
	cerr << "[-2jitterk jitter_keep]\n";
	cerr << "[-2jitterd jitter_down]\n";
	cerr << "[-2jitterp jitter_period]\n";
	exit( -1 );
}
