/*************************************************************************

 nemud.cpp
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <time.h>
#include <sys/time.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h> 	
#include <netinet/if_ether.h>
#include <netpacket/packet.h>
#include <errno.h>

#include "defs.h"
#include "utils.h"
#include "linkedlist.t"
#include "impairs.h"

using namespace std;

// function declarations

void usage();

// global variables and definitions


struct Data
{
	struct timeval sendtime;
	int len;
	void *frame;
};

// the main function

int main( int argc, char **argv )
{
	void *framep;
	Data *datap;
	char buf[1600];
	int buflen = 1600;  // should accomodate any valid frames
	int sock1, sock2;
	int msglen;
	int d1, d2, fjitter;
	struct sockaddr_ll addrll;
	struct timeval now;
	char ifname1[IFNAMSIZ];
	char ifname2[IFNAMSIZ];
	int par_if1, par_if2;
	int smid;
	struct JittLoss jittloss;
	LinkedList <Data> frames1, frames2;
	Impairs *impairs;
	
	cerr << "nemud " << VERSION << ", (C) 2006, Norbert Vegh\n";

	// process the command line arguments
	
	if( argc != 5 ) usage();
	par_if1 = par_if2 = 0;
	
	while( argc > 1 )
	{
		if( !strcmp( "-1if", argv[1] ) )
		{
			if( par_if1 ) configError( "1if parameter is duplicated" );
			if( argc == 1 ) usage();
			strncpy( ifname1, argv[2], IFNAMSIZ );
			ifname1[IFNAMSIZ-1] = 0;
			par_if1 = -1;
			argc -= 2; argv += 2;
		}
		else if( !strcmp( "-2if", argv[1] ) )
		{
			if( par_if2 ) configError( "2if parameter is duplicated" );
			if( argc == 1 ) usage();
			strncpy( ifname2, argv[2], IFNAMSIZ );
			ifname2[IFNAMSIZ-1] = 0;
			par_if2 = -1;
			argc -= 2; argv += 2;
		}
		else
		{
			cerr << "Invalid option: " << argv[1] << endl;
			return -1;
		}
	}
	
	// check parameters
	
	if( par_if1 == 0 ) configError( "if1 parameter is missing\n" );
	if( par_if2 == 0 ) configError( "if2 parameter is missing\n" );
	
	// create shared memory and initalize it
	
	smid = shmget( SMKEY, sizeof( Impairs ), IPC_CREAT );
	if( smid == -1 )
	{
		perror( "Cannot create shared memory" );
		return -1;
	}
	impairs = ( Impairs * )shmat( smid, NULL, 0 );

	// create sockets

	sock1 = socket( PF_PACKET, SOCK_RAW, htons( ETH_P_ALL ) );  // packet socket to capture everything
	if( sock1 == -1 )
	{
		perror( "Cannot create the first packet socket for reading" );
		exit( -1 );
	}
	if( setPromiscMode( ifname1, -1 ) ) exit( -1 );
	if( getIfIndex( ifname1, addrll.sll_ifindex ) )
	{
		perror( "Cannot get first interface index" );
		exit( -1 );
	}
	addrll.sll_family = AF_PACKET;
	if( bind( sock1, ( struct sockaddr * )&addrll, sizeof( addrll ) ) )
	{
		perror( "Cannot bind socket to the first interface" );
		exit( -1 );
	}
	
	sock2 = socket( PF_PACKET, SOCK_RAW, htons( ETH_P_ALL ) );
	if( sock1 == -1 )
	{
		perror( "Cannot create the second packet socket for reading" );
		exit( -1 );
	}
	if( setPromiscMode( ifname2, -1 ) ) exit( -1 );
	if( getIfIndex( ifname2, addrll.sll_ifindex ) )
	{
		perror( "Cannot get second interface index" );
		exit( -1 );
	}
	if( bind( sock2, ( struct sockaddr * )&addrll, sizeof( addrll ) ) )
	{
		perror( "Cannot bind socket to the second interface" );
		exit( -1 );
	}

	// main read-send cycle

	while( -1 )
	{
		// try to receive on first interface
		
		msglen = recv( sock1, buf, buflen, MSG_DONTWAIT );
		if( msglen < 1 )  // reading error
		{
			if( errno != EAGAIN )
			{
				perror( "Reading error on the first interface\n" );
			}
		}
		else
		{
			gettimeofday( &now, NULL );
			impairs->total1++;
			fjitter = 0;
			if( impairs->filejitter1.exist )
			{
				fjitter = impairs->filejitter1.Value();
			}
			if( impairs->closs1.exist )
			{
				d1 = impairs->closs1.Drop();
			}
			else
			{
				d1 = 0;
			}
			if( impairs->bloss1.exist )
			{
				d2 = impairs->bloss1.Drop( now );
			}
			else
			{
				d2 = 0;
			}
			if( impairs->jitter1.exist )
			{
				jittloss = impairs->jitter1.Value( now );
			}
			else
			{
				jittloss.loss = 0;
				jittloss.jitter = 0;
			}
			if( d1 || d2 || jittloss.loss || ( fjitter == -1 ) )  // implement loss
			{
				impairs->dropped1++;
			}
			else
			{
				framep = malloc( msglen );
				if( framep == NULL )
				{
					cerr << "Cannot allocate memory\n";
					exit( -1 );
				}
				memcpy( framep, ( void * ) buf, msglen );
				datap = new Data;
				if( datap == 0 )
				{
					cerr << "Cannot allocate memory\n";
					exit( -1 );
				}
				datap->sendtime = now;
				if( impairs->delay1 != 0 ) addTime( datap->sendtime, impairs->delay1 );  // implement delay
				if( impairs->cjitter1.exist )
				{
					addTime( datap->sendtime, impairs->cjitter1.Value() );  // implement continuous jitter
				}
				if( jittloss.jitter ) addTime( datap->sendtime, jittloss.jitter );  // implement jitter
				if( fjitter > 0 )
				{
					addTime( datap->sendtime, fjitter );  // implement jitter from file
				}
				datap->frame = framep;
				datap->len = msglen;
				frames1.last();
				frames1.insertAfter( datap );
			}
		}
		
		// try to receive on second interface
		
		msglen = recv( sock2, buf, buflen, MSG_DONTWAIT );
		if( msglen < 1 )  // reading error
		{
			if( errno != EAGAIN )
			{
				perror( "Reading error on the second interface\n" );
			}
		}
		else
		{
			gettimeofday( &now, NULL );
			impairs->total2++;
			fjitter = 0;
			if( impairs->filejitter2.exist )
			{
				fjitter = impairs->filejitter2.Value();
			}
			if( impairs->closs2.exist )
			{
				d1 = impairs->closs2.Drop();
			}
			else
			{
				d1 = 0;
			}
			if( impairs->bloss2.exist )
			{
				d2 = impairs->bloss2.Drop( now );
			}
			else
			{
				d2 = 0;
			}
			if( impairs->jitter2.exist )
			{
				jittloss = impairs->jitter2.Value( now );
			}
			else
			{
				jittloss.loss = 0;
				jittloss.jitter = 0;
			}
			if( d1 || d2 || jittloss.loss || ( fjitter == -1 ) )  // implement loss
			{
				impairs->dropped2++;
			}
			else
			{
				framep = malloc( msglen );
				if( framep == NULL )
				{
					cerr << "Cannot allocate memory\n";
					exit( -1 );
				}
				memcpy( framep, ( void * ) buf, msglen );
				datap = new Data;
				if( datap == 0 )
				{
					cerr << "Cannot allocate memory\n";
					exit( -1 );
				}
				datap->sendtime = now;
				if( impairs->delay2 != 0 ) addTime( datap->sendtime, impairs->delay2 );  // implement delay
				if( impairs->cjitter2.exist )
				{
					addTime( datap->sendtime, impairs->cjitter2.Value() );  // implement continuous jitter
				}
				if( jittloss.jitter ) addTime( datap->sendtime, jittloss.jitter );  // implement jitter
				if( fjitter > 0 )
				{
					addTime( datap->sendtime, fjitter );  // implement jitter from file
				}
				datap->frame = framep;
				datap->len = msglen;
				frames2.last();
				frames2.insertAfter( datap );
			}
		}
		
		// now try to send if there is any on the first interface
		
		gettimeofday( &now, NULL );
		frames2.first();
		while( !frames2.endOfList() )
		{
			datap = frames2.getEntry();
			if( ( now > datap->sendtime ) )
			{
				if( send( sock1, datap->frame, datap->len, 0 ) < 0 )
				{
					perror( "Send error on first interface" );
				}
				else
				{
					free( datap->frame );
					frames2.remove();
					continue;
				}
			}
			break;
		}
		
		// now try to send if there is any on the second interface
		
		gettimeofday( &now, NULL );
		frames1.first();
		while( !frames1.endOfList() )
		{
			datap = frames1.getEntry();
			if( ( now > datap->sendtime ) )
			{
				if( send( sock2, datap->frame, datap->len, 0 ) < 0 )
				{
					perror( "Send error on first interface" );
				}
				else
				{
					free( datap->frame );
					frames1.remove();
					continue;
				}
			}
			break;
		}
	}
}

void usage()
{
	cerr << "Command line arguments:\n";
	cerr << "-if1 input_interface\n";
	cerr << "-if2 output_interface\n";
	exit( -1 );
}
