/*************************************************************************

 ngen.cpp
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h> 	
#include <netinet/if_ether.h> 	
#include <netinet/tcp.h> 	
#include <netpacket/packet.h> 	
#include <net/if_arp.h>
#include <pthread.h>	

#include "defs.h"
#include "utils.h"
#include "distrib.h"
#include "linkedlist.t"

using namespace std;

struct Stream
{
	pthread_t tid;  // thread id (for TCP streams)
	int pending;  // indicates the the TCP sender has not finished the sending
	int sock;  // the socket for TCP streams
	int lowLevel;  // delay measurement on this stream -> packet socket
	char dstMac[13];  // the destination mac address
	char outIfname[IFNAMSIZ];  // the out interface for this stream
	int ifIndex;
	struct sockaddr_ll addrll;
	unsigned long srcIp;  // network byte order
	IpRange *srcIpRange;
	unsigned long dstIp;  // network byte order
	IpRange *dstIpRange;
	int proto;
	int vlanid;
	int srcPort;  // host byte order
	PortRange *srcPortRange;
	int dstPort;  // host byte order
	PortRange *dstPortRange;
	int tos;
	int pbits;
	Distrib *dist;
	unsigned long seqNum;
};

void *tcpSender( void * );
void usage();

int main( int argc, char **argv )
{
	ifstream scriptf;
	char buf[1503];
	char *vlanbuf;
	char *ipbuf;
	unsigned char c;
	unsigned long *ubuf;
	unsigned long defaddr;
	int i, size, x, ret;
	int rawsock, packetsock, type;
	int rawstream, packetstream;
	struct timeval now, tmptime;
	struct sockaddr_in udpremote;
	struct sockaddr_in remote, local;
	int addrll_len, addrin_len;
	struct in_addr inaddr;
	sigset_t sigset;
	Stream *mystream;
	LinkedList <Stream> streamtb;  // stream table
	pthread_t tid;
	int par_srcip, par_dstip, par_proto, par_srcport, par_dstport, par_vlanid = 0;
	int par_rate, par_size, par_time, par_dist, par_tos, par_avgint, par_pbits;
	int par_rate2, par_size2, par_time2, par_dstmac, par_outif;

	cerr << "ngen " << VERSION << ", (C) 2006, Norbert Vegh\n";
	
	// initialize variables

	vlanbuf = buf;
	ipbuf = buf + 4;
	type = 0;
	mystream = 0;
	rawstream = packetstream = 0;
	rawsock = packetsock = 0;
	addrll_len = sizeof( struct sockaddr_ll );
	addrin_len = sizeof( struct sockaddr_in );
	
	// get the address of the default interface
	
	defaddr = getIfAddr( DEF_IF );

	// process the command line arguments

	if( ( argc > 1 ) && ( !strcmp( "-f", argv[1] ) ) )   // process the script file
	{
		if( argc != 3 ) usage();
		scriptf.open( argv[2], ios::in );
		if( !scriptf )
		{
			cerr << "Cannot open " << argv[2] << endl;
			return -1;
		}
		
		// now read the config file
		
		cerr << "Reading config...\n";
		if( nextWord( scriptf, buf, 256 ) ) configError( "Empty file" );
		if( strcmp( "streams", buf ) ) configError( "Keyword streams is expected instead of", buf );
		while( nextWord( scriptf, buf, 256 ) != -1 )  // while there is a new stream entry
		{
			cerr << "Stream #" << streamtb.size() + 1 << endl;
			if( streamtb.size() == MAX_STREAMS ) configError( "Too many streams, there can be maximum MAX_STREAMS" );
			mystream = new Stream;
			memset( mystream, 0, sizeof( mystream ) );
			if( mystream == 0 )
			{
				cerr << "Out of memory\n";
				exit( -1 );
			}
			if( strcmp( "{", buf ) ) configError( "{ is expected instead of", buf );
			
			// set the default stream parameters
			
			mystream->vlanid = -1;
			mystream->srcIp = defaddr;
			mystream->srcPort = DEF_SPORT;
			strncpy( mystream->outIfname, DEF_IF, IFNAMSIZ );
			mystream->outIfname[IFNAMSIZ-1] = 0;
			// now read the parameters of the stream
			
			par_srcip = par_dstip =  par_proto = par_srcport = par_dstport = par_vlanid = 0;
			par_rate = par_size = par_time = par_dist = par_tos = par_avgint = par_pbits = 0;
			par_rate2 = par_size2 = par_time2 = par_dstmac = par_outif = 0;
			while( -1 )
			{
				if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
				if( !strcmp( "srcip", buf ) )
				{
					if( par_srcip ) configError( "srcip parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "range", buf ) )
					{
						mystream->srcIpRange = new IpRange;
						if( mystream->srcIpRange == 0 )
						{
							cerr << "Cannot allocate memory\n";
							exit( -1 );
						}
						readIpRange( *mystream->srcIpRange, scriptf, buf, 256 );
					}
					else
					{
						if( inet_aton( buf, &inaddr ) == 0 ) configError( "Invalid srcip: ", buf );
						mystream->srcIp = inaddr.s_addr;
						mystream->srcIpRange = 0;
					}
					par_srcip = -1;
				}
				else if( !strcmp( "dstip", buf ) )
				{
					if( par_dstip ) configError( "dstip parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "range", buf ) )
					{
						mystream->dstIpRange = new IpRange;
						if( mystream->dstIpRange == 0 )
						{
							cerr << "Cannot allocate memory\n";
							exit( -1 );
						}
						readIpRange( *mystream->dstIpRange, scriptf, buf, 256 );
					}
					else
					{
						if( inet_aton( buf, &inaddr ) ==  0 ) configError( "Invalid dstip: ", buf );
						mystream->dstIp = inaddr.s_addr;
					}
					par_dstip = -1;
				}
				else if( !strcmp( "lowlevel", buf ) )
				{
					mystream->lowLevel = -1;
				}
				else if( !strcmp( "vlanid", buf ) )
				{
					if( par_vlanid ) configError( "vlanid parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					x = atoi( buf );
					if( x < 0 || x > 4095 ) configError( "Invalid VlanID: ", buf );
					mystream->vlanid = x;
					par_vlanid = -1;
				}
				else if( !strcmp( "pbits", buf ) )
				{
					if( par_pbits ) configError( "pbits parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					x = atoi( buf );
					if( x < 0 || x > 4095 ) configError( "Invalid pbits: ", buf );
					mystream->pbits = x;
					par_pbits = -1;
				}
				else if( !strcmp( "dstmac", buf ) )
				{
					if( par_dstmac ) configError( "dstmac parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					strncpy( mystream->dstMac, buf, 13 );
					mystream->outIfname[12] = 0;
					par_dstmac = -1;
				}
				else if( !strcmp( "outif", buf ) )
				{
					if( par_outif ) configError( "outif parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					strncpy( mystream->outIfname, buf, IFNAMSIZ );
					mystream->outIfname[IFNAMSIZ-1] = 0;
					par_outif = -1;
				}
				else if( !strcmp( "proto", buf ) )
				{
					if( par_proto ) configError( "proto parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "tcp", buf ) ) mystream->proto = IPPROTO_TCP;
					else if( !strcmp( "udp", buf ) )
					{
						mystream->proto = IPPROTO_UDP;
					}
					else configError( "Unknown protocol: ", buf );
					par_proto = -1;
				}
				else if( !strcmp( "tos", buf ) )
				{
					if( par_tos ) configError( "tos/prec parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					x = atoi( buf );
					if( x < 0 || x > 255 ) configError( "Invalid tos: ", buf );
					mystream->tos = x;
					par_tos = -1;
				}
				else if( !strcmp( "prec", buf ) )
				{
					if( par_tos ) configError( "tos/prec parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					x = atoi( buf );
					if( x < 0 || x > 7 ) configError(  "Invalid precedence: ", buf );
					x <<= 5;
					mystream->tos = x;
					par_tos = -1;
				}
				else if( !strcmp( "srcport", buf ) )
				{
					if( par_srcport ) configError( "srcport parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "range", buf ) )
					{
						mystream->srcPortRange = new PortRange;
						if( mystream->srcPortRange == 0 )
						{
							cerr << "Cannot allocate memory\n";
							exit( -1 );
						}
						readPortRange( *mystream->srcPortRange, scriptf, buf, 256 );
						mystream->srcPort = 1;
					}
					else
					{
						x = atoi( buf );
						if( x < 0 || x > 65535 ) configError( "Invalid srcport: ", buf );
						mystream->srcPort = x;
						mystream->srcPortRange = 0;
					}
					par_srcport = -1;
				}
				else if( !strcmp( "dstport", buf ) )
				{
					if( par_dstport ) configError( "dstport parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "range", buf ) )
					{
						mystream->dstPortRange = new PortRange;
						if( mystream->dstPortRange == 0 )
						{
							cerr << "Cannot allocate memory\n";
							exit( -1 );
						}
						readPortRange( *mystream->dstPortRange, scriptf, buf, 256 );
						mystream->dstPort = 1;
					}
					else
					{
						x = atoi( buf );
						if( x < 0 || x > 65535 ) configError( "Invalid dstport: ", buf );
						mystream->dstPort = x;
						mystream->dstPortRange = 0;
					}
					par_dstport = -1;
				}
				else if( !strcmp( "distrib", buf ) )
				{
					if( par_dist ) configError( "distrib parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "fix", buf ) ) type = FIX;  // fix distribution
					else if( !strcmp( "poisson", buf ) ) type = POISSON;  // poisson distribution
					else if( !strcmp( "on-off", buf ) ) type = ONOFF;  // onoff distribution
					else configError( "Unknown distribution", buf );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( strcmp( "{", buf ) ) configError( "{ is expected instead of ", buf );

					// now read the distribution parameters

					if( type == FIX )
					{
						Fix *fixp = new Fix;
						fixp->size = 1500;  // default packet size
						while( -1 )
						{
							if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
							if( !strcmp( "size", buf ) )
							{
								if( par_size ) configError( "size parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								x = atoi( buf );
								if( x < 46 || x > 1500 ) configError( "Invalid size: ", buf );
								fixp->size = x;
								par_size = -1;
							}
							else if( !strcmp( "rate", buf ) )
							{
								if( par_rate ) configError( "rate parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								x = readRate( buf );
								if( x < 1 || x > MAX_RATE ) configError( "Invalid rate: ", buf );
								fixp->rate = x;
								par_rate = -1;
							}
							else if( !strcmp( "}", buf ) ) break;
							else configError( "Invalid keyword: ", buf );
						}
						if( par_rate == 0 ) configError( "fix: rate parameter is missing" );
						mystream->dist = ( Distrib * )fixp;
					}
					else if( type == POISSON )
					{
						Poisson *poissonp = new Poisson;
						poissonp->size = 1500;  // default packet size
						while( -1 )
						{
							if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
							if( !strcmp( "size", buf ) )
							{
								if( par_size ) configError( "size parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								x = atoi( buf );
								if( x < 46 || x > 1500 ) configError( "Invalid size: ", buf );
								poissonp->size = x;
								par_size = -1;
							}
							else if( !strcmp( "avgint", buf ) )
							{
								if( par_avgint ) configError( "avgint parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								x = readTime( buf );
								if( x < 1 ) configError( "Invalid avgint: ", buf );
								poissonp->avgint = x;
								par_avgint = -1;
							}
							else if( !strcmp( "}", buf ) ) break;
							else configError( "Invalid keyword: ", buf );
						}
						if( par_avgint == 0 ) configError( "poisson: avgint parameter is missing" );
						mystream->dist = ( Distrib * )poissonp;
					}
					else if( type == ONOFF )
					{
						OnOff *onoffp = new OnOff;
						onoffp->size1 = 1500;
						onoffp->size2 = 1500;
						while( -1 )
						{
							if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
							if( !strcmp( "size1", buf ) )
							{
								if( par_size ) configError( "size1 parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								x = atoi( buf );
								if( x < 46 || x > 1500 ) configError( "Invalid size1: ", buf );
								onoffp->size1 = x;
								par_size = -1;
							}
							else if( !strcmp( "size2", buf ) )
							{
								if( par_size2 ) configError( "size2 parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								x = atoi( buf );
								if( x < 46 || x > 1500 ) configError( "Invalid size2: ", buf );
								onoffp->size2 = x;
								par_size2 = -1;
							}
							else if( !strcmp( "rate1", buf ) )
							{
								if( par_rate ) configError( "rate1 parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								x = readRate( buf );
								if( x < 0 || x > MAX_RATE ) configError( "Invalid rate1: ", buf );
								onoffp->rate1 = x;
								par_rate = -1;
							}
							else if( !strcmp( "rate2", buf ) )
							{
								if( par_rate2 ) configError( "rate2 parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								x = readRate( buf );
								if( x < 0 || x > MAX_RATE ) configError( "Invalid rate2: ", buf );
								onoffp->rate2 = x;
								par_rate2 = -1;
							}
							else if( !strcmp( "time1", buf ) )
							{
								if( par_time ) configError( "time1 parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								x = readTime( buf );
								if( x < 1 ) configError( "Invalid time1: ", buf );
								onoffp->time1 = x;
								par_time = -1;
							}
							else if( !strcmp( "time2", buf ) )
							{
								if( par_time2 ) configError( "time2 parameter is duplicated" );
								if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
								if( !strcmp( "infinite", buf ) ) onoffp->time2 = 0;
								else
								{
									x = readTime( buf );
									if( x < 1 ) configError( "Invalid time2: ", buf );
									onoffp->time2 = x;
								}
								par_time2 = -1;
							}
							else if( !strcmp( "}", buf ) ) break;
							else configError( "Invalid keyword: ", buf );
						}
						if( par_rate == 0 ) configError( "onoff: rate1 parameter is missing" );
						if( par_rate2 == 0 ) configError( "onoff: rate2 parameter is missing" );
						if( par_time == 0 ) configError( "onoff: time1 parameter is missing" );
						if( par_time2 == 0 ) configError( "onoff: time2 parameter is missing" );
						mystream->dist = ( Distrib * )onoffp;
					}
					par_dist = -1;
				}
				else if( !strcmp( "}", buf ) ) break;
				else configError( "Invalid keyword: ", buf );
			}
			if( par_dstip == 0 ) configError( "dstip parameter is missing" );
			if( par_proto == 0 ) configError( "proto parameter is missing" );
			if( par_dstport == 0 ) configError( "dstport parameter is missing" );
			if( !mystream->lowLevel && par_pbits ) configError( "pbits parameter is only for lowlevel mode" );
			if( !mystream->lowLevel && par_vlanid ) configError( "vlanid parameter is only for lowlevel mode" );
			if( mystream->proto == IPPROTO_TCP )
			{
				if( mystream->srcIpRange || mystream->dstIpRange || mystream->srcPortRange \
				|| mystream->dstPortRange ) configError( "Address and port ranges are not supported for TCP streams" ); 
				if( par_srcip ) configError( "srcip parameter is not supported for TCP streams" );
				if( mystream->lowLevel ) configError( "lowlevel mode is not supported for TCP streams" );
				if( par_dstmac ) configError( "dstmac parameter is not supported for TCP streams" );
				if( par_outif ) configError( "outif parameter is not supported for TCP streams" );
			}
			else
			{
				if( ( mystream->lowLevel || par_dstmac ) && !( mystream->lowLevel && par_dstmac ) )
				{
					configError( "delay, dstmac and outif parameters must be specified together" );
				}
				if( mystream->lowLevel && !par_srcip )
				{
					configError( "for delay measurements the srcip parameter must be defined, and must have a non-local address\n" );
				}
			}
			if( par_dist == 0 ) configError( "distrib parameter is missing" );
			streamtb.insertAfter( mystream );
		}
		if( streamtb.size() == 0 ) configError( "no streams" );
		cerr << "Done\n";
	}
	else  // no script -> process the command line arguments
	{
		if( argc < 6 || argc%2 != 1 ) usage();
		mystream = new Stream;
		if( mystream == 0 )
		{
			cerr << "Out of memory\n";
			exit( -1 );
		}
		memset( mystream, 0, sizeof( mystream ) );
		mystream->srcIp = defaddr;  // set default values
		mystream->srcPort = DEF_SPORT;  // set default ports
		mystream->dstPort = DEF_DPORT;
		Fix *fixp = new Fix;
		fixp->size = 1500;  // set the default packet size
		mystream->dist = ( Distrib * )fixp;
		
		// read the arguments
		
		par_srcip = par_dstip =  par_proto = par_srcport = par_dstport = 0;
		par_rate = par_size = par_tos = 0;
		while( argc != 1 )
		{
			if( !strcmp( "-srcip", argv[1] ) )
			{
				if( par_srcip ) configError( "srcip parameter is duplicated" );
				if( inet_aton( argv[2], &inaddr ) == 0 )
				{
					cerr << "Invalid source IP\n";
					return -1;
				}
				mystream->srcIp = inaddr.s_addr;
				par_srcip = -1;
			}
			else if( !strcmp( "-dstip", argv[1] ) )
			{
				if( par_dstip ) configError( "dstip parameter is duplicated" );
				if( inet_aton( argv[2], &inaddr ) == 0 )
				{
					cerr << "Invalid dest IP\n";
					return -1;
				}
				mystream->dstIp = inaddr.s_addr;
				par_dstip = -1;
			}
			else if( !strcmp( "-proto", argv[1] ) )
			{
				if( par_proto ) configError( "proto parameter is duplicated" );
				if( !strcmp( "tcp", argv[2] ) ) mystream->proto = IPPROTO_TCP;
				else if( !strcmp( "udp", argv[2] ) )
				{
					mystream->proto = IPPROTO_UDP;
				}
				else
				{
					cerr << "Unknown protocol\n";
					return -1;
				}
				par_proto = -1;
			}
			else if( !strcmp( "-srcport", argv[1] ) )
			{
				if( par_srcport ) configError( "srcport parameter is duplicated" );
				x = atoi( argv[2] );
				if( x < 0 || x > 65535 )
				{
					cerr << "Invalid source port\n";
					return -1;
				}
				mystream->srcPort = x;
				par_srcport = -1;
			}
			else if( !strcmp( "-dstport", argv[1] ) )
			{
				if( par_dstport ) configError( "dstport parameter is duplicated" );
				x = atoi( argv[2] );
				if( x < 0 || x > 65535 )
				{
					cerr << "Invalid destination port\n";
					return -1;
				}
				mystream->dstPort = x;
				par_dstport = -1;
			}
			else if( !strcmp( "-tos", argv[1] ) )
			{
				if( par_tos ) configError( "tos/prec parameter is duplicated" );
				x = atoi( argv[2] );
				if( x < 0 || x > 255 )
				{
					cerr << "Invalid tos value\n";
					return -1;
				}
				mystream->tos = x;
				par_tos = -1;
			}
			else if( !strcmp( "-prec", argv[1] ) )
			{
				if( par_tos ) configError( "tos/prec parameter is duplicated" );
				x = atoi( argv[2] );
				if( x < 0 || x > 7 )
				{
					cerr << "Invalid precedence port\n";
					return -1;
				}
				mystream->tos = x << 5;
				par_tos = -1;
			}
			else if( !strcmp( "-rate", argv[1] ) )
			{
				if( par_rate ) configError( "rate parameter is duplicated" );
				x = readRate( argv[2] );
				if( x < 1 || x > MAX_RATE )
				{
					cerr << "Invalid rate, must be between 1.." << MAX_RATE << endl;
					return -1;
				}
				fixp->rate = x;
				par_rate = -1;
			}
			else if( !strcmp( "-size", argv[1] ) )
			{
				if( par_size ) configError( "size parameter is duplicated" );
				x = atoi( argv[2] );
				if( x < 46 || x > 1500 || ( x > 1496 ) && ( par_vlanid ) )
				{
					cerr << "Invalid packet size, must be between 46..1500 (1496 for VLANs)\n";
					return -1;
				}
				fixp->size = x;
				par_size = -1;
			}
			else
			{
				cerr << "Invalid option: " << argv[1] << endl;
				return -1;
			}
			argc -= 2;
			argv += 2;
		}
		
		// check parameters
		
		if( par_dstip == 0 ) configError( "dstip parameter is missing\n" );
		if( par_proto == 0 ) configError( "proto parameter is missing\n" );
		if( par_rate == 0 ) configError( "rate parameter is missing\n" );
		if( mystream->proto == IPPROTO_TCP )
		{
			if( fixp->size < OVERHEAD + 12 )
			{
				cerr << "For TCP streams the packet size must be between " \
				<< OVERHEAD + 12 << " and 1500\n";
				return -1;
			}
			if( par_srcip )
			{
				cerr << "Cannot set srcip parameter for TCP streams\n";
			}
			if( par_vlanid )
			{
				cerr << "Cannot set vlandid parameter for TCP streams\n";
			}
		}
		streamtb.insertAfter( mystream );
	}
	
	// fill in the sockaddr_ll struct for all delay measuring streams
	
	streamtb.first();
	while( !streamtb.endOfList() )
	{
		mystream = streamtb.getEntry();
		streamtb.next();
		if( mystream->proto != IPPROTO_UDP ) continue;
		if( mystream->lowLevel == 0 )  // normal udp stream
		{
			rawstream = -1;
			continue;
		}
		packetstream = -1;
		mystream->addrll.sll_family = AF_PACKET;
//		mystream->addrll.sll_protocol = htons( 0x8100 );  // VLAN
		if( getIfIndex( mystream->outIfname, mystream->addrll.sll_ifindex ) )
		{
			exit( -1 );
		}
		mystream->addrll.sll_halen = 6;
		x = 0;
		for( i = 0; i < 6; i++ )
		{
			mystream->addrll.sll_addr[i] = hexConverter( mystream->dstMac[x] )*16 + hexConverter( mystream->dstMac[x+1] );
			x += 2;
		}
	}
	
	// create the packet socket
	
	buildIpHeader( ( struct iphdr * )( ipbuf ) );
		
	if( packetstream )
	{
		packetsock = socket( PF_PACKET, SOCK_DGRAM, htons( ETH_P_IP ) );
		if( packetsock == -1 )
		{
			perror( "Cannot create a packet socket" );
			return -1;
		}
	}

	// open the raw socket for normal udp streams

	if( rawstream )
	{
			rawsock = socket( PF_INET, SOCK_RAW, IPPROTO_RAW );
			if( rawsock == -1 )
			{
				perror( "Cannot create the raw socket" );
				return -1;
			}
			udpremote.sin_family = PF_INET;
			udpremote.sin_port = htons( IPPROTO_UDP );
	}
	
	// block SIGINT and SIGALRM, needed for sigwait

	sigemptyset( &sigset );
	sigaddset( &sigset, SIGINT );
	sigaddset( &sigset, SIGALRM );
	pthread_sigmask ( SIG_BLOCK, &sigset, NULL );

	// initialize the streams

	streamtb.first();
	while( !streamtb.endOfList() )
	{
		mystream = streamtb.getEntry();
		mystream->dist->init();  // init the distribution class
		if( mystream->proto != IPPROTO_TCP )
		{
			streamtb.next();
			continue;
		}
		
		// initialize tcp connections
		
		mystream->pending = -1;  // the sender is not ready
		mystream->sock = socket( PF_INET, SOCK_STREAM, 0 );
		if( mystream->sock == -1 )
		{
			perror( "Cannot open TCP socket" );
			exit( -1 );
		}
		x = -1;
		if( setsockopt( mystream->sock, SOL_SOCKET, SO_REUSEADDR, &x, sizeof( x ) ) )
		{
			perror( "Cannot set SO_REUSEADDR socket option" );
		}
		c = mystream->tos;
		if( setsockopt( mystream->sock, SOL_IP, IP_TOS, ( void * )&c, sizeof( c ) ) )
		{
			perror( "Cannot set tos for TCP socket" );
			exit( -1 );
		}
		x = -1;
		if( setsockopt( mystream->sock, SOL_TCP, TCP_NODELAY, ( void * )&x, sizeof( x ) ) )
		{
			perror( "Cannot set TCP_NODELAY socket option" );
			exit( -1 );
		}
		local.sin_port = htons( mystream->srcPort );
		local.sin_addr.s_addr = htonl( INADDR_ANY );
		if( bind( mystream->sock, ( struct sockaddr * )&local, sizeof( local ) ) == -1 )
		{
			perror( "Cannot bind TCP socket to local port:" );
			cerr << mystream->srcPort << endl;
			exit( -1 );
		}
		remote.sin_family = PF_INET;
		remote.sin_addr.s_addr = mystream->dstIp;
		remote.sin_port = htons( mystream->dstPort );
		if( connect( mystream->sock, ( struct sockaddr * )&remote, sizeof( remote ) ) )
		{
			perror( "Cannot connect TCP socket to:" );
			cerr << inet_ntoa( remote.sin_addr ) << " " << mystream->dstPort << endl;
			exit( -1 );
		}
		
		// create the thread for this stream
		
		if( pthread_create( &tid, NULL, tcpSender, ( void * )( mystream ) ) != 0 )
		{
			cerr << "Cannot create a TCP sender thread\n";
			exit( -1 );
		}
		streamtb.next();
	}

	// unblock SIGINT
	
	sigemptyset( &sigset );
	sigaddset( &sigset, SIGINT );
	pthread_sigmask ( SIG_UNBLOCK, &sigset, NULL );

	// traffic generation
	
	cerr << "Sending...\n";
	while( -1 )
	{
		gettimeofday( &now, NULL );
		if( streamtb.size() == 0 )
		{
			cerr << "No active streams\n";
			exit( -1 );
		}
		streamtb.first();
		while( !streamtb.endOfList() )  // search for a stream that has a packet to send
		{
			mystream = streamtb.getEntry();
			tmptime = mystream->dist->getSendTime();
			if( tmptime.tv_sec > now.tv_sec )
			{
				streamtb.next();
				continue;
			}
			if( ( tmptime.tv_sec == now.tv_sec ) && ( tmptime.tv_usec > now.tv_usec ) )
			{
				streamtb.next();
				continue;
			}

			if( mystream->proto == IPPROTO_UDP )  // UDP
			{
				size = mystream->dist->getSize();
				if( size < 46 ) size = 46;
				else if( size > 1500 ) size = 1500;
				if( mystream->srcIpRange != 0 ) mystream->srcIp = mystream->srcIpRange->getNextIp();
				if( mystream->dstIpRange != 0 ) mystream->dstIp = mystream->dstIpRange->getNextIp();
				if( mystream->srcPortRange != 0 ) mystream->srcPort = mystream->srcPortRange->getNextPort();
				if( mystream->dstPortRange != 0 ) mystream->dstPort = mystream->dstPortRange->getNextPort();
				modifyIpHeader( ( struct iphdr * )ipbuf, size, mystream->srcIp, mystream->dstIp, IPPROTO_UDP, mystream->tos );
				buildUdpHeader( ( struct udphdr * )( ipbuf+20 ), size-20, mystream->srcPort, mystream->dstPort );
				ubuf = ( unsigned long * )( ipbuf+28 );   // skip the ip and udp header
				ubuf[0] = htonl( mystream->seqNum );  // place the sequence number
				ubuf[1] = htonl( now.tv_sec );  // place timestamp
				ubuf[2] = htonl( now.tv_usec );
				if( mystream->lowLevel )
				{
					if( mystream->vlanid == -1 )
					{
						mystream->addrll.sll_protocol = htons( ETH_P_IP );  // ethernet
						computeChecksum( ( struct iphdr * )ipbuf );
						ret = sendto( packetsock, ipbuf, size, 0, ( struct sockaddr * )&( mystream->addrll ), addrll_len );
					}
					else
					{
						mystream->addrll.sll_protocol = htons( 0x8100 );  // VLAN
						setVlanHeader(	( struct vlanhdr * )( vlanbuf ), mystream->vlanid, mystream->pbits );
						computeChecksum( ( struct iphdr * )ipbuf );
						ret = sendto( packetsock, vlanbuf, size+4, 0, ( struct sockaddr * )&( mystream->addrll ), addrll_len );
					}
				}
				else
				{
					udpremote.sin_addr.s_addr = mystream->dstIp;
					ret = sendto( rawsock, ipbuf, size, 0, ( struct sockaddr * )&udpremote, addrin_len );
				}
				if( ret != -1 )
				{
					mystream->seqNum++;
				}
				else
				{
					perror( "Send error" );
				}
				mystream->dist->nextState();
			}
			else  // TCP
			{
				if( mystream->sock == -1 ) // the sender is broken
				{
					streamtb.remove();
					continue;
				}
				if( mystream->pending == -1 )
				{
					streamtb.next();
					continue;  // the sender is not ready
				}
				mystream->pending = -1;
				pthread_kill( mystream->tid, SIGALRM );  // signal to the thread
			}
			streamtb.next();
		}
	}
}

void *tcpSender( void * arg )
{
	char buf[1500];
	int n, size;
	unsigned long *ubuf;
	Stream *mystream;
	struct timeval now;
	struct sockaddr_in remote;
	sigset_t sigset;

	mystream = ( Stream * )arg;

	// traffic generation
	
	sigemptyset( &sigset );
	sigaddset( &sigset, SIGALRM );
	mystream->tid = pthread_self();
	mystream->pending = 0;  // the sender is ready
	while( -1 )
	{
		sigwait( &sigset, &n );  // wait for an ALRM signal
			
		// prepare and send the packet

		size = mystream->dist->getSize();
		if( size < OVERHEAD + 12 ) size = OVERHEAD + 12;  // 12 bytes seqnum + timestamp
		else if( size > 1500 ) size = 1500;
		size -= OVERHEAD;
		ubuf = ( unsigned long * )buf;
		ubuf[0] = htonl( mystream->seqNum );  // place the sequence number
		ubuf[1] = htonl( now.tv_sec );   // place timestamp
		ubuf[2] = htonl( now.tv_usec );
		if( send( mystream->sock, buf, size, MSG_NOSIGNAL ) == -1 )
		{
			remote.sin_addr.s_addr = mystream->dstIp;
			cerr << "TCP session to " << inet_ntoa( remote.sin_addr ) \
			<< " " << mystream->dstPort << " got broken\n";
			mystream->sock = -1;
			pthread_exit( NULL );
		}
		gettimeofday( &now, NULL );
		mystream->dist->setSendTime( now );  // sending time of the last packet
		mystream->dist->nextState();  // calculate next time and size
		mystream->seqNum++;
		mystream->pending = 0;  // clear pending state
	}
}

void usage()
{
	cerr << "Command line arguments:\n";
	cerr << "-proto udp|tcp\n";
	cerr << "-dstip dstip\n";
	cerr << "-rate rate : ( like 500k or 10M )\n";
	cerr << "[-size size] : in bytes\n";
	cerr << "[-srcip srcip]\n";
	cerr << "[-srcport srcport] : the default is " << DEF_SPORT << "\n";
	cerr << "[-dstport dstport] : the default is " << DEF_DPORT << "\n";
	cerr << "[-tos tos]\n";
	cerr << "[-prec precedence]\n";
	cerr << "-----or----\n-f file\n";
	exit( -1 );
}
