/*************************************************************************

 nrecv.cpp
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h> 	
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <pthread.h>

#include "defs.h"
#include "range.h"
#include "utils.h"
#include "rwlock.h"
#include "statistics.h"
#include "linkedlist.t"	

using namespace std;

// global variables and definitions

struct Interface
{
	char ifname[IFNAMSIZ];
	int ifnum;
	int promisc;
	int curs;  // cursor for each thread for the stream table
};

LinkedList <Interface> g_ifaces;

#define ID_SIZE 9

struct Stream
{
	int par_id;
	char id[ID_SIZE];
	int par_if;
	char ifname[IFNAMSIZ];
	int ifnum;
	int log;
	int logloss;
	unsigned long logdelay;
	int par_logdelay;
	int par_srcip;
	unsigned long srcIp;  // network byte order
	IpRange *srcIpRange;
	int par_dstip;
	unsigned long dstIp;  // network byte order
	IpRange *dstIpRange;
	int par_proto;
	int proto;
	int par_vlanid;
	int vlanid;
	int par_srcport;
	int srcPort;  // host byte order
	PortRange *srcPortRange;
	int par_dstport;
	int dstPort;  // host byte order
	PortRange *dstPortRange;
	int par_tos;
	int tos;
	int par_prec;
	int prec;
	int par_pbits;
	int pbits;
	int measureDelay;
	int measureJitter;
	int measurePercentile;
	int firstPacket;  // this is set to 0 when receiving the first packet
	unsigned long seqNum;
	struct Statistics *actualStat;
	struct Statistics *standbyStat;
	struct Statistics stat1;
	struct Statistics stat2;
	struct timeval firstTime;  // the arrival time of the first packet
	int terminated;  // indicates that the stream is broken
};

LinkedList <Stream> g_streamtb;
int g_writelog;
ofstream g_logfile;
RwLock g_rwlock;

// function declarations

void processPacket( struct timeval &tv, struct Stream *mystream, char *buf, int msglen );
void *udpReceiver( void *iface );
void *tcpReceiver( void *tcpstream );
void usage();

// the main function

int main( int argc, char **argv )
{
	ifstream scriptf;
	char buf[256], statfname[256], logfname[256];
	ostringstream ostream;
	ofstream statfile;
	int n, x, width;
	int udpstream;  // indicates the existence of udp streams
	int writestat;
	unsigned long sec, usec, timediff;
	unsigned long interval;   // statistical interval time
	struct Stream *mystream;
	struct Interface *myiface;
	struct Statistics *mystat;
	struct timeval lastStat;   // time of last statistical interval
	struct timeval now;
	struct tm *date;
	struct itimerval itimer;
	struct in_addr inaddr;
	pthread_t tid;
	sigset_t sigset;
	unsigned long delay, jitter, minDelay, maxDelay, perc;
	float rate, totalrate;
	unsigned long totalpackets, totallosses, totalmisords;
	int par_if, par_log, par_logfile, par_statfile, par_int;
	float p1, p2, p3;
	int par_p1, par_p2, par_p3;
	int measureDelay, measureJitter, measurePercentile, timestamp;

	
	cerr << "nrecv " << VERSION << ", (C) 2006, Norbert Vegh\n";

	// initialize some variables

	interval = 1000;   // the default statistical interval time is 1 sec
	statfname[0] = logfname[0] = 0;
	g_writelog = writestat = 0;
	delay = jitter = perc = 0;
	mystream = 0;
	udpstream = 0;
	par_if = par_log = par_logfile = par_statfile = par_int = 0;
	p1 = p2 = p3 = 0;
	par_p1 = par_p2 = par_p3 = 0;
	measureDelay = measureJitter = measurePercentile = timestamp = 0;
	width = 8;

	// process the command line arguments

	if( ( argc > 1 ) && ( !strcmp( "-f", argv[1] ) ) )   // process the script file
	{
		if( argc != 3 ) usage();
		scriptf.open( argv[2], ios::in );
		if( !scriptf )
		{
			cerr << "Cannot open " << argv[2] << endl;
			return -1;
		}

		// now read the config file
		
		cerr << "Reading config...\n";
		while( nextWord( scriptf, buf, 256 ) != -1 )  // while there is a new keyword
		{
			if( !strcmp( "interval", buf ) )
			{
				if( par_int ) configError( "interval parameter is duplicated" );
				if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
				x = readTime( buf ) / 1000;
				if( x < 1 ) configError( "interval is invalid" );
				interval = x;
				par_int = -1;
			}
			else if( !strcmp( "statfile", buf ) )
			{
				if( par_statfile ) configError( "statfile parameter is duplicated" );
				if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
				strcpy( statfname, buf );
				writestat = -1;
				par_statfile = -1;
			}
			else if( !strcmp( "timestamp", buf ) )
			{
				timestamp = -1;
			}
			else if( !strcmp( "logfile", buf ) )
			{
				if( par_log ) configError( "logfile parameter is duplicated" );
				if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
				strcpy( logfname, buf );
				par_logfile = -1;
			}
			else if( !strcmp( "log", buf ) )
			{
				if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
				g_writelog = -1;
				par_log = -1;
			}
			else if( !strcmp( "p1", buf ) )
			{
				if( par_p1 ) configError( "p1 parameter is duplicated" );
				if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
				p1 = ( float )atof( buf );
				if( p1 <= 0 ) configError( "Invalid percentile: ", buf );
				par_p1 = -1;
			}
			else if( !strcmp( "p2", buf ) )
			{
				if( par_p2 ) configError( "p2 parameter is duplicated" );
				if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
				p2 = ( float )atof( buf );
				if( p2 <= 0 ) configError( "Invalid percentile: ", buf );
				par_p2 = -1;
			}
			else if( !strcmp( "p3", buf ) )
			{
				if( par_p3 ) configError( "p3 parameter is duplicated" );
				if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
				p3 = ( float )atof( buf );
				if( p3 <= 0 ) configError( "Invalid percentile: ", buf );
				par_p3 = -1;
			}
			else if( !strcmp( "streams", buf ) ) break;
			else configError( "Invalid keyword: ", buf );
		}
		while( nextWord( scriptf, buf, 256 ) != -1 )  // while there is a new stream entry
		{
			cerr << "Stream #" << g_streamtb.size() + 1 << endl;
			if( g_streamtb.size() == MAX_STREAMS ) configError( "Too many streams, there can be maximum MAX_STREAMS" );
			mystream = new Stream;
			if( mystream == 0 )
			{
				cerr << "Cannot allocate memory\n";
				exit( -1 );
			}
			memset( mystream, 0, sizeof( mystream ) );
			strncpy( mystream->ifname, DEF_IF, IFNAMSIZ );  // the default interface
			mystream->ifname[IFNAMSIZ-1] = 0;
			if( strcmp( "{", buf ) ) configError( "{ is expected instead of", buf );
			
			// now read the parameters of the stream
			
			while( -1 )
			{
				if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
				if( !strcmp( "id", buf ) )
				{
					if( mystream->par_id ) configError( "id parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					strncpy( mystream->id, buf, ID_SIZE );
					mystream->id[ID_SIZE-1] = '\0';
					mystream->par_id = -1;
				}
				else if( !strcmp( "if", buf ) )
				{
					if( mystream->par_if ) configError( "interface parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					strncpy( mystream->ifname, buf, IFNAMSIZ );
					mystream->ifname[IFNAMSIZ-1] = 0;
					mystream->par_if = -1;
				}
				else if( !strcmp( "log", buf ) )
				{
					mystream->log = -1;
					par_log = -1;
				}
				else if( !strcmp( "logloss", buf ) )
				{
					mystream->logloss = -1;
					par_log = -1;
				}
				else if( !strcmp( "logdelay", buf ) )
				{
					if( mystream->par_logdelay ) configError( "logdelay parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					x = readTime( buf );
					if( x <= 0 ) configError( "Invalid logdelay: ", buf );
					mystream->logdelay = x;
					mystream->par_logdelay = -1;
					par_log = -1;
				}
				else if( !strcmp( "delay", buf ) )
				{
					mystream->measureDelay = -1;
					measureDelay = -1;
				}
				else if( !strcmp( "jitter", buf ) )
				{
					mystream->measureJitter = -1;
					measureJitter = -1;
				}
				else if( !strcmp( "percentile", buf ) )
				{
					mystream->measurePercentile = -1;
					measurePercentile = -1;
				}
				else if( !strcmp( "srcip", buf ) )
				{
					if( mystream->par_srcip ) configError( "srcip parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "range", buf ) )
					{
						mystream->srcIpRange = new IpRange;
						if( mystream->srcIpRange == 0 )
						{
							cerr << "Cannot allocate memory\n";
							exit( -1 );
						}
						readIpRange( *mystream->srcIpRange, scriptf, buf, 256 );
					}
					else
					{
						if( inet_aton( buf, &inaddr ) == 0 ) configError( "Invalid srcip: ", buf );
						mystream->srcIp = inaddr.s_addr;
					}
					mystream->par_srcip = -1;
				}
				else if( !strcmp( "dstip", buf ) )
				{
					if( mystream->par_dstip ) configError( "dstip parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "range", buf ) )
					{
						mystream->dstIpRange = new IpRange;
						if( mystream->dstIpRange == 0 )
						{
							cerr << "Cannot allocate memory\n";
							exit( -1 );
						}
						readIpRange( *mystream->dstIpRange, scriptf, buf, 256 );
					}
					else
					{
						if( inet_aton( buf, &inaddr ) == 0 ) configError( "Invalid dstip: ", buf );
						mystream->dstIp = inaddr.s_addr;
					}
					mystream->par_dstip = -1;
				}
				else if( !strcmp( "proto", buf ) )
				{
					if( mystream->par_proto ) configError( "proto parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "tcp", buf ) ) mystream->proto = IPPROTO_TCP;
					else if( !strcmp( "udp", buf ) )
					{
						udpstream = -1;
						mystream->proto = IPPROTO_UDP;
					}
					else configError( "Unknown protocol: ", buf );
					mystream->par_proto = -1;
				}
				else if( !strcmp( "tos", buf ) )
				{
					if( mystream->par_tos || mystream->par_prec ) configError( "tos/prec parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					x = atoi( buf );
					if( x < 0 || x > 255 ) configError( "Invalid tos: ", buf );
					mystream->tos = x;
					mystream->par_tos = -1;
				}
				else if( !strcmp( "prec", buf ) )
				{
					if( mystream->par_tos || mystream->par_prec ) configError( "tos/prec parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					x = atoi( buf );
					if( x < 0 || x > 7 ) configError( "Invalid precedence: ", buf );
					mystream->prec = x;
					mystream->par_prec = -1;
				}
				else if( !strcmp( "pbits", buf ) )
				{
					if( mystream->par_pbits ) configError( "pbits parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					x = atoi( buf );
					if( x < 0 || x > 7 ) configError( "Invalid pbits: ", buf );
					mystream->pbits = x;
					mystream->par_pbits = -1;
				}
				else if( !strcmp( "vlanid", buf ) )
				{
					if( mystream->par_vlanid ) configError( "vlanid parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					x = atoi( buf );
					if( x < 0 || x > 4095 ) configError( "Invalid vlanid: ", buf );
					mystream->vlanid = x;
					mystream->par_vlanid = -1;
				}
				else if( !strcmp( "srcport", buf ) )
				{
					if( mystream->par_srcport ) configError( "srcport parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "range", buf ) )
					{
						mystream->srcPortRange = new PortRange;
						if( mystream->srcPortRange == 0 )
						{
							cerr << "Cannot allocate memory\n";
							exit( -1 );
						}
						readPortRange( *mystream->srcPortRange, scriptf, buf, 256 );
					}
					else
					{
						x = atoi( buf );
						if( x < 0 || x > 65535 ) configError( "Invalid srcport: ", buf );
						mystream->srcPort = x;
					}
					mystream->par_srcport = -1;
				}
				else if( !strcmp( "dstport", buf ) )
				{
					if( mystream->par_dstport ) configError( "dstport parameter is duplicated" );
					if( nextWord( scriptf, buf, 256 ) ) configError( "Unexpected end of file" );
					if( !strcmp( "range", buf ) )
					{
						mystream->dstPortRange = new PortRange;
						if( mystream->dstPortRange == 0 )
						{
							cerr << "Cannot allocate memory\n";
							exit( -1 );
						}
						readPortRange( *mystream->dstPortRange, scriptf, buf, 256 );
					}
					else
					{
						x = atoi( buf );
						if( x < 0 || x > 65535 ) configError( "Invalid dstport: ", buf );
						mystream->dstPort = x;
					}
					mystream->par_dstport = -1;
				}
				else if( !strcmp( "}", buf ) ) break;
				else configError( "Invalid keyword: ", buf );
			}
			
			// check parameters
			
			if( par_log && !par_logfile ) configError( "logfile must be defined for logging" );
			if( mystream->par_proto == 0 ) configError( "proto parameter is missing" );
			if( mystream->proto == IPPROTO_TCP )
			{
				if( mystream->par_if ) configError( "interface parameter is not supported for TCP streams" );
				if( mystream->par_dstport == 0 ) configError( "dstport parameter is mandatory for TCP streams" );
				if( mystream->dstPortRange ) configError( "port range is not for TCP streams" );
				if( mystream->par_tos || mystream->par_prec ) configError( "tos/prec parameter is not supported for TCP streams" );
				if( mystream->par_srcport ) configError( "srcport parameter is not supported for TCP streams" );
				if( mystream->par_srcip ) configError( "srcip parameter is not supported for TCP streams" );
				if( mystream->par_dstip ) configError( "dstip parameter is not supported for TCP streams" );
			}
						
			// insert the stream into the list
			
			g_streamtb.insertAfter( mystream );
		}
		if( g_streamtb.size() == 0 ) configError( "no streams" );
		cerr << "Done\n";
	}	
	else  // no script -> process the command line arguments
	{
		if( argc < 3 ) usage();
		mystream = new Stream;
		if( mystream == 0 )
		{
			cerr << "Cannot allocate memory\n";
			exit( -1 );
		}
		memset( mystream, 0, sizeof( mystream ) );
		strcpy( mystream->id, "stream_0" );
		mystream->dstPort = DEF_DPORT;  // the default dest port
		strncpy( mystream->ifname, DEF_IF, IFNAMSIZ );  // the default interface
		mystream->ifname[IFNAMSIZ-1] = 0;
		while( argc > 1 )
		{
			if( !strcmp( "-delay", argv[1] ) )
			{
				mystream->measureDelay = -1;
				measureDelay = -1;
				argc--, argv++;
			}
			else if( !strcmp( "-jitter", argv[1] ) )
			{
				mystream->measureJitter = -1;
				measureJitter = -1;
				argc--, argv++;
			}
			else if( !strcmp( "-timestamp", argv[1] ) )
			{
				timestamp = -1;
				argc--, argv++;
			}
			else if( !strcmp( "-percentile", argv[1] ) )
			{
				mystream->measurePercentile = -1;
				measurePercentile = -1;
				argc--, argv++;
			}
			else if( !strcmp( "-p1", argv[1] ) )
			{
				if( par_p1 ) configError( "p1 parameter is duplicated" );
				if( argc == 2 ) usage();
				p1 = atof( argv[2] );
				if( p1 <= 0 ) configError( "Invalid percentile: ", buf );
				par_p1 = -1;
				argc -= 2, argv += 2;
			}
			else if( !strcmp( "-p2", argv[1] ) )
			{
				if( par_p2 ) configError( "p2 parameter is duplicated" );
				if( argc == 2 ) usage();
				p2 = atof( argv[2] );
				if( p2 <= 0 ) configError( "Invalid percentile: ", buf );
				par_p2 = -1;
				argc -= 2, argv += 2;
			}
			else if( !strcmp( "-p3", argv[1] ) )
			{
				if( par_p3 ) configError( "p3 parameter is duplicated" );
				if( argc == 2 ) usage();
				p3 = atof( argv[2] );
				if( p3 <= 0 ) configError( "Invalid percentile: ", buf );
				par_p3 = -1;
				argc -= 2, argv += 2;
			}
			else if( !strcmp( "-int", argv[1] ) )
			{
				if( par_int ) configError( "int parameter is duplicated" );
cerr << "argc: " << argc << endl;
				if( argc == 2 ) usage();
				interval = readTime( argv[2] ) / 1000;
				if( interval <= 0 ) configError( "Invalid time interval\n" );
				par_int = -1;
				argc -= 2, argv += 2;
			}
			else if( !strcmp( "-log", argv[1] ) )
			{
				if( par_log ) configError( "log parameter is duplicated" );
				if( argc == 2 ) usage();
				strcpy( logfname, argv[2] );
				g_writelog = -1;
				par_log = -1;
				argc -= 2, argv += 2;
			}
			else if( !strcmp( "-stat", argv[1] ) )
			{
				if( par_statfile ) configError( "stat parameter is duplicated" );
				if( argc == 2 ) usage();
				strcpy( statfname, argv[2] );
				writestat = -1;
				par_statfile = -1;
				argc -= 2, argv += 2;
			}
			else if( !strcmp( "-if", argv[1] ) )
			{
				if( mystream->par_if ) configError( "if parameter is duplicated" );
				if( argc == 2 ) usage();
				strncpy( mystream->ifname, argv[2], IFNAMSIZ );
				mystream->ifname[IFNAMSIZ-1] = 0;
				mystream->par_if = -1;
				argc -= 2; argv += 2;
			}
			else if( !strcmp( "-srcip", argv[1] ) )
			{
				if( mystream->par_srcip ) configError( "srcip parameter is duplicated" );
				if( argc == 2 ) usage();
				if( inet_aton( argv[2], &inaddr ) == 0 ) configError( "Invalid source IP\n" );
				mystream->srcIp = inaddr.s_addr;
				mystream->par_srcip = -1;
				argc -= 2; argv += 2;
			}
			else if( !strcmp( "-dstip", argv[1] ) )
			{
				if( mystream->par_dstip ) configError( "dstip parameter is duplicated" );
				if( argc == 2 ) usage();
				if( inet_aton( argv[2], &inaddr ) == 0 ) configError( "Invalid destination IP\n" );
				mystream->dstIp = inaddr.s_addr;
				mystream->par_dstip = -1;
				argc -= 2; argv += 2;
			}
			else if( !strcmp( "-proto", argv[1] ) )
			{
				if( mystream->par_proto ) configError( "proto parameter is duplicated" );
				if( argc == 2 ) usage();
				if( !strcmp( "tcp", argv[2] ) ) mystream->proto = IPPROTO_TCP;
				else if( !strcmp( "udp", argv[2] ) )
				{
					mystream->proto = IPPROTO_UDP;
					udpstream = -1;  // there is at least one UDP stream
				}
				else configError( "Unknown protocol\n" );
				mystream->par_proto = -1;
				argc -= 2; argv += 2;
			}
			else if( !strcmp( "-srcport", argv[1] ) )
			{
				if( mystream->par_srcport ) configError( "srcport parameter is duplicated" );
				if( argc == 2 ) usage();
				x = atoi( argv[2] );
				if( x < 0 || x > 65535 ) configError( "Invalid source port\n" );
				mystream->srcPort = x;
				mystream->par_srcport = -1;
				argc -= 2; argv += 2;
			}
			else if( !strcmp( "-dstport", argv[1] ) )
			{
				if( mystream->par_dstport ) configError( "dstport parameter is duplicated" );
				if( argc == 2 ) usage();
				x = atoi( argv[2] );
				if( x < 0 || x > 65535 ) configError( "Invalid destination port\n" );
				mystream->dstPort = x;
				mystream->par_dstport = -1;
				argc -= 2; argv += 2;
			}
			else if( !strcmp( "-tos", argv[1] ) )
			{
				if( mystream->par_tos || mystream->par_prec ) configError( "tos/prec parameter is duplicated" );
				if( argc == 2 ) usage();
				x = atoi( argv[2] );
				if( x < 0 || x > 255 ) configError( "Invalid tos value\n" );
				mystream->tos = x;
				mystream->par_tos = -1;
				argc -= 2; argv += 2;
			}
			else if( !strcmp( "-prec", argv[1] ) )
			{
				if( mystream->par_tos || mystream->par_prec ) configError( "tos/prec parameter is duplicated" );
				if( argc == 2 ) usage();
				x = atoi( argv[2] );
				if( x < 0 || x > 7 ) configError( "Invalid precedence port\n" );
				mystream->prec = x;
				mystream->par_prec = -1;
				argc -= 2; argv += 2;
			}
			else configError( "Invalid option: ", argv[1] );
		}

		// check parameters

		if( mystream->par_proto == 0 ) configError( "proto parameter is missing" );
		if( mystream->proto == IPPROTO_TCP )
		{
			if( mystream->par_if ) configError( "interface parameter is not suuported for TCP streams" );
			if( mystream->dstPortRange ) configError( "port range is not supported for TCP streams" );
			if( par_if ) configError( "if parameter is not supported for TCP streams" );
			if( mystream->par_tos || mystream->par_prec ) configError( "tos/prec parameter is not supported for TCP streams" );
			if( mystream->par_srcport ) configError( "srcport parameter is not supported for TCP streams" );
			if( mystream->par_srcip ) configError( "srcip parameter is not supported for TCP streams" );
			if( mystream->par_dstip ) configError( "dstip parameter is not supported for TCP streams" );
		}
		g_streamtb.insertAfter( mystream );
	}

	// open the stat and log files
	
	if( par_log )
	{
		g_logfile.open( logfname, ios::out );
		if( !g_logfile )
		{
			cerr << "Cannot open " << logfname << endl;
			return -1;
		}
	}
	if( writestat )
	{
		statfile.open( statfname, ios::out );
		if( !statfile )
		{
			cerr << "Cannot open file " << statfname << endl;
			exit( -1 );
		}
	}

	// create the interface list with all the ifs used in the stream config

	n = 0;  // interface number
	g_streamtb.first();
	while( !g_streamtb.endOfList() )
	{
		mystream = g_streamtb.getEntry();
		g_streamtb.next();
		x = 0;
		g_ifaces.first();
		while( !g_ifaces.endOfList() )
		{
			myiface = g_ifaces.getEntry();
			if( !strcmp( myiface->ifname, mystream->ifname ) )
			{
				x = -1;  // there is already an entry
				mystream->ifnum = myiface->ifnum;
				break;
			}
			g_ifaces.next();
		}
		if( x ) continue;  // nothing to do
		
		// add the this interface to the list
		
		myiface = new Interface;
		if( myiface == 0 )
		{
			cerr << "Cannot allocate memory\n";
			exit( -1 );
		}
		strncpy( myiface->ifname, mystream->ifname, IFNAMSIZ );
		myiface->ifnum = n;
		mystream->ifnum = n;
		myiface->curs = n+1;  // cursor 0 is used by the main thread
		g_ifaces.insertAfter( myiface );
		n++;
	}
	
	// initialization

	if( timestamp ) width = 26;
	g_streamtb.initCursors( g_ifaces.size() + 1 );
	// one cursor for each udp thread, and one for the main stream
	g_streamtb.first();
	while( !g_streamtb.endOfList() )
	{
		mystream = g_streamtb.getEntry();
		mystream->seqNum = 0;
		mystream->firstPacket = -1;
		mystream->actualStat = &( mystream->stat1 );
		mystream->standbyStat = &( mystream->stat2 );
		memset( mystream->actualStat, 0, sizeof( struct Statistics ) );
		memset( mystream->standbyStat, 0, sizeof( struct Statistics ) );
		g_streamtb.next();
	}
	gettimeofday( &lastStat, NULL );

	// block SIGINT and SIGALRM, needed for sigwait

	sigemptyset( &sigset );
	sigaddset( &sigset, SIGALRM );
	sigaddset( &sigset, SIGINT );
	pthread_sigmask ( SIG_BLOCK, &sigset, NULL );

	// init the log and stat file
	
	if( par_log )
	{
		g_logfile << setiosflags( ios::left ) << setw( 15 ) << "Time" << setw( 10 ) \
		<< "Stream" \
		<< setw( 8 ) << "Bytes" << setw( 11 ) << "Seq" << setw( 10 ) << "Delay" << endl;
	}
	
	ostream << setiosflags( ios::left );
	if( timestamp )
	{
		ostream << setw( 24 ) << "Timestamp";
	}
	ostream <<	setw( 8 )<< "Stream" \
	<< resetiosflags( ios::left ) << setw( 9 ) << "Packets" << setw( 9 ) \
	<< "Losses" << setw( 9 ) << "Misorder" << setw( 9 ) << "Rate";
	if( measureDelay )
	{
		ostream << setw( 9 ) << "Delay";
	}
	if( measureJitter )
	{
		ostream << setw( 9 ) << "mDelay" << setw( 9 ) << "MDelay" << setw( 9 ) << "Jitter";
	}
	if( measurePercentile )
	{
		if( p1 != 0 ) ostream << "p" << setiosflags( ios::left ) << setw( 8 ) << p1;
		if( p2 != 0 ) ostream << "p" << setiosflags( ios::left ) << setw( 8 ) << p2;
		if( p3 != 0 ) ostream << "p" << setiosflags( ios::left ) << setw( 8 ) << p3;
	}
	ostream << endl;
	cout << ostream.str();
	if( writestat ) statfile << ostream.str();
	// start the receiver threads
	
	g_streamtb.first();
	while( !g_streamtb.endOfList() )
	{
		mystream = g_streamtb.getEntry();
		g_streamtb.next();
		if( mystream->proto != IPPROTO_TCP ) continue;

		if( pthread_create( &tid, NULL, tcpReceiver, ( void * )( mystream ) ) != 0 )
		{
			cerr << "Cannot create a TCP receiver thread\n";
			return -1;
		}
	}
	if( udpstream )  // create one udp receiver for each interface
	{
		g_ifaces.first();
		while( !g_ifaces.endOfList() )
		{
			myiface = g_ifaces.getEntry();
			if( pthread_create( &tid, NULL, udpReceiver, ( void * )myiface ) )
			{
				cerr << "Cannot create the UDP receiver thread\n";
				return -1;
			}
			g_ifaces.next();
		}
	}
	
	// set the timer that will send SIGALRM signal periodically

	sec = interval/1000;
	usec = ( interval%1000 ) * 1000;
	itimer.it_interval.tv_sec = sec;
	itimer.it_interval.tv_usec = usec;
	itimer.it_value.tv_sec = sec;
	itimer.it_value.tv_usec = usec;
	if( setitimer( ITIMER_REAL, &itimer, NULL ) )
	{
		perror( "Cannot set the timer" );
		exit( -1 );
	}

	// wait for SIGALRM and SIGINT in infinite while cycle

	while( -1 )
	{
		sigwait( &sigset, &n );
		g_rwlock.writeLock();
		gettimeofday( &now, NULL );   // the starting time of the new statistical interval
		date = localtime( &now.tv_sec );
		if( g_streamtb.size() == 0 || n == SIGINT )  // quit
		{
			if( writestat )
			{
				statfile.close();
			}
		
			// set back the original promiscuos mode for all interfaces
			
			if( udpstream )
			{
				g_ifaces.first();
				while( !g_ifaces.endOfList() )
				{
					myiface = g_ifaces.getEntry();
					g_ifaces.next();
					if( !myiface->promisc )
					{
						setPromiscMode( myiface->ifname, 0 );
					}
				}
			}
			exit( 0 );
		}
		else  // print the stats
		{
			// Replace the actual and the standby containers

			g_streamtb.first();
			while( !g_streamtb.endOfList() )
			{
				mystream = g_streamtb.getEntry();
				g_streamtb.next();
				if( mystream->terminated ) continue;
				mystat = mystream->actualStat;
				mystream->actualStat = mystream->standbyStat;
				mystream->standbyStat = mystat;
			}
			g_rwlock.unlock();
			totalrate = totalpackets = totallosses = totalmisords = 0;
			g_streamtb.first();
			while( !g_streamtb.endOfList() )
			{
				mystream = g_streamtb.getEntry();
				mystat = mystream->standbyStat;
				timediff = timeDiff( now, lastStat );  // time from the previous stat interval
				rate = mystat->bytes * ( 8000 / ( float )timediff );  // in kbps
				totalrate += rate;
				totalpackets += mystat->packets;
				totallosses += mystat->losses;
				totalmisords += mystat->misorderings;
				minDelay = mystat->minDelay;
				maxDelay = mystat->maxDelay;
				if( mystream->measureJitter )
				{	
					jitter = maxDelay - minDelay;
				}
				if( mystream->measureDelay )
				{
					delay = ( unsigned long )( mystat->delay / mystat->packets );
				}

				// now print the statistics

				ostream.str( "" );

				if( timestamp )
				{
					ostream << setfill( '0' ) << setw( 2 ) << date->tm_year-100 << "/" \
					<< setw( 2 ) <<date->tm_mon+1 << "/" << setw( 2 ) << date->tm_mday << "." \
					<< setw( 2 ) << date->tm_hour << ":" << setw( 2 ) << date->tm_min << ":" \
					<< setw( 2 ) << date->tm_sec << " " << setfill( ' ' ) ;
				}
				ostream << setiosflags( ios::left ) << setw( 8 ) << mystream->id;
				ostream  << resetiosflags( ios::left ) << setw( 9 ) << mystat->packets \
				<< setw( 9 ) << mystat->losses << setw( 9 ) << mystat->misorderings \
				<< resetiosflags( ios::left ) << setiosflags( ios::fixed ) << setw( 7 );
				if( rate < 1000.0 ) ostream << setprecision( 1 ) << rate << " k";
				else ostream << setprecision( 1 ) << rate/1000.0 << " M";
				if( mystream->measureDelay )
				{
					ostream << setw( 7 );
					if( delay < 1000 ) ostream << delay << " u";
					else if( delay < 1000000 ) ostream << setprecision( 1 ) << ( float )delay/1000.0 << " m";
					else ostream << setprecision( 1 ) << ( float )delay/1000000.0 << " s";
					if( mystat->invalid ) ostream << " E";
				}
				if( mystream->measureJitter )
				{
					ostream << setw( 7 );
					if( minDelay < 1000 ) ostream << minDelay << " u";
					else if( minDelay < 1000000 )
					{
						ostream << setprecision( 1 ) << ( float )minDelay/1000.0 << " m";
					}
					else ostream << setprecision( 1 ) << ( float )minDelay/1000000.0 << " s";
					ostream << setw( 7 );
					if( maxDelay < 1000 ) ostream << maxDelay << " u";
					else if( maxDelay < 1000000 )
					{
						ostream << setprecision( 1 ) << ( float )maxDelay/1000.0 << " m";
					}
					else ostream << setprecision( 1 ) << ( float )maxDelay/1000000.0 << " s";
					ostream << setw( 7 );
					if( jitter < 1000 ) ostream << jitter << " u";
					else if( jitter < 1000000 )
					{
						ostream << setprecision( 1 ) << ( float )jitter/1000.0 << " m";
					}
					else ostream << setprecision( 1 ) << ( float )jitter/1000000.0 << " s";
				}
				if( mystream->measurePercentile )
				{
					ostream << setw( 7 );
					if( p1 != 0 )
					{
						ostream << setw( 7 );
						perc = mystat->getPercentile( p1 );
						if( perc < 1000 ) ostream << perc << " u";
						else if( perc < 1000000 )
						{
							ostream << setprecision( 1 ) << ( float )perc/1000.0 << " m";
						}
						else ostream << setprecision( 1 ) << ( float )perc/1000000.0 << " s";
					}
					if( p2 != 0 )
					{
						perc = mystat->getPercentile( p2 );
						ostream << setw( 7 );
						if( perc < 1000 ) ostream << perc << " u";
						else if( perc < 1000000 )
						{
							ostream << setprecision( 1 ) << ( float )perc/1000.0 << " m";
						}
						else ostream << setprecision( 1 ) << ( float )perc/1000000.0 << " s";
					}
					ostream << setw( 7 );
					if( p3 != 0 )
					{
						ostream << setw( 7 );
						perc = mystat->getPercentile( p3 );
						if( perc < 1000 ) ostream << perc << " u";
						else if( perc < 1000000 )
						{
							ostream << setprecision( 1 ) << ( float )perc/1000.0 << " m";
						}
						else ostream << setprecision( 1 ) << ( float )perc/1000000.0 << " s";
					}
				}
				ostream << endl;
				cout << ostream.str();
				if( writestat ) statfile << ostream.str();

				// now clear the standby stat
				
				mystat->clear();
//				memset( mystat, 0, sizeof( *mystat ) );
				g_streamtb.next();
			}
			lastStat = now;
			
			// now print the total stats
			
			ostream.str( "" );

			ostream << setiosflags( ios::left ) << setw( width ) << "TOTAL" \
			<< resetiosflags( ios::left ) << setw( 9 ) << totalpackets \
			<< setw( 9 ) << totallosses << setw( 9 ) << totalmisords \
			<< setprecision( 1 ) << setiosflags( ios::showpoint ) << setw( 7 );
			if( totalrate < 1000.0 ) ostream << totalrate << " k";
			else ostream << totalrate/1000.0 << " M";
			ostream << endl << setfill( '-' ) << setw( 85 ) << "-" << setfill( ' ' ) << endl;
			cout << ostream.str();
			if( writestat ) statfile << ostream.str();
		}
	}
}

// the tcp receiver threads

void *tcpReceiver( void *tcpstream )
{
	char buf[1600];
	int x, buflen, msglen;
	int sock, port, accsock;
	struct timeval intime;
	struct Stream *mystream;
	struct sockaddr_in local;
	
	buflen = 1600;  // MTU
	mystream = ( struct Stream * )tcpstream;
	accsock = socket( PF_INET, SOCK_STREAM, 0 );
	if( accsock == -1 )
	{
		perror( "Cannot open TCP socket" );
		exit( -1 );
	}
	x = -1;
	if( setsockopt( accsock, SOL_SOCKET, SO_REUSEADDR, &x, sizeof( x ) ) )
	{
		perror( "Cannot set SO_REUSEADDR socket option" );
	}
	x = -1;
	if( setsockopt( accsock, SOL_TCP, TCP_NODELAY, ( void * )&x, sizeof( x ) ) )
	{
		perror( "Cannot set TCP_NODELAY" );
		exit( -1 );
	}
	port = mystream->dstPort;
	local.sin_family = PF_INET;
	local.sin_port = htons( port );
	local.sin_addr.s_addr = htonl( INADDR_ANY );
	if( bind( accsock, ( struct sockaddr * )&local, sizeof( local ) ) == -1 )
	{
		perror( "Cannot bind socket" );
		exit( -1 );
	}
	if( listen( accsock, 256 ) == -1 )
	{
		perror( "Cannot listen on socket" );
		exit( -1 );
	}
	sock = accept( accsock, NULL, NULL );
	if( sock == -1 )
	{
		perror( "Cannot accept" );
		exit( -1 );
	}
	
	// now start receiving
	
	while( -1 )
	{
		msglen = recv( sock, buf, buflen, 0 );
		gettimeofday( &intime, NULL );
		if( msglen < 1 )  // reading error
		{
			cerr << "TCP session " << mystream->id << " got broken\n";
			mystream->terminated = -1;
			pthread_exit( NULL );
		}
		processPacket( intime, ( struct Stream * )tcpstream, buf, msglen );
	}
}

// the udp receiver thread

void *udpReceiver( void *iface )
{
	char buf[1500];
	char *ipbuf;
	int x;
	int curs, buflen, sock, msglen, vlanid;
	socklen_t addrlen;
	int srcport, dstport, tos, proto, pbits = 0;
	unsigned long srcip, dstip;
	struct timeval intime;
	struct Stream *mystream;
	struct Interface *myiface;
	struct iphdr *iph;
	struct udphdr *udph;
	struct sockaddr_ll addrll;
	struct vlanhdr *vlanh;
	struct sched_param sch_param;
	
	buflen = 1500;  // maximum packet size
	myiface = ( struct Interface * )iface;
	curs = myiface->curs;
	sock = socket( PF_PACKET, SOCK_DGRAM, htons( ETH_P_IP ) );  // packet socket to capture everything
	if( sock == -1 )
	{
		perror( "Cannot create the packet socket for the UDP streams" );
		exit( -1 );
	}

	// now get the flags for the interface and set promiscuous mode

	if( getPromiscMode( myiface->ifname, myiface->promisc ) ) exit( -1 );
	if( !myiface->promisc )  // we have to set promisc mode
	{
		if( setPromiscMode( myiface->ifname, -1 ) ) exit( -1 );
	}

	// now bind the socket to the interface
	
	x = 0;
	getIfIndex( myiface->ifname, x );
	addrll.sll_family = AF_PACKET;
	addrll.sll_protocol = htons( ETH_P_IP );
	addrll.sll_ifindex = x;
	
	if( bind( sock, ( struct sockaddr * )&addrll, sizeof( addrll ) ) )
	{
		perror( "Cannot bind UDP socket to the interface" );
		exit( -1 );
	}
	
	// now set the scheduling parameters

	sch_param.sched_priority = 99;
	if(  pthread_setschedparam( pthread_self(), SCHED_FIFO, &sch_param ) )
	{
		perror( "Cannot set priority scheduling" );
		exit( -1 );
	}
	
	// now start capturing
	
	while( -1 )
	{
		msglen = recvfrom( sock, buf, buflen, 0, ( struct sockaddr * )&addrll, &addrlen );
		gettimeofday( &intime, NULL );
		if( msglen < 1 )  // reading error
		{
			cerr << "Reading error on the UDP socket\n";
			pthread_exit( NULL );
		}
		if( addrll.sll_protocol == htons( ETH_P_IP ) )
		{
			vlanid = -1;
			ipbuf = buf;
		}
		else if( addrll.sll_protocol == htons( 0x8100 ) )
		{
			vlanh = ( struct vlanhdr *)&buf;
			pbits = vlanh->myunion.mystruct.prio;
			vlanh->myunion.mystruct.cfi = 0;
			vlanh->myunion.mystruct.prio = 0;
			vlanid = htons( vlanh->myunion.id );
			ipbuf = buf + 4;
		}
		else
		{
			continue;  // some other protocol
		}
		iph = ( struct iphdr * )ipbuf;
		proto = iph->protocol;
		if( proto != IPPROTO_UDP ) continue;
		srcip = iph->saddr;
		dstip = iph->daddr;
		tos = iph->tos;
		udph = ( struct udphdr * )( ipbuf + iph->ihl*4 );
		srcport = ntohs( udph->source );
		dstport = ntohs( udph->dest );
		g_streamtb.first( curs );
		while( !g_streamtb.endOfList( curs ) )  // look for the UDP streams
		{
			mystream = g_streamtb.getEntry( curs );
			g_streamtb.next( curs );
			if( ( vlanid != -1 ) && !mystream->par_vlanid ) continue;
			if( mystream->ifnum != myiface->ifnum ) continue;
			if( mystream->par_tos && ( tos != mystream->tos ) ) continue;
			if( mystream->par_prec && ( tos >> 5 != mystream->prec ) ) continue;
			if( mystream->par_pbits && ( pbits != mystream->pbits ) ) continue;
			if( mystream->par_vlanid && ( vlanid != mystream->vlanid ) ) continue;
			if( mystream->par_srcip )
			{
				if( mystream->srcIpRange == 0 )  // fix address
				{
					if( srcip != mystream->srcIp ) continue;
				}
				else  // address range
				{
					if( ntohl( srcip ) < mystream->srcIpRange->getFirstIp() || \
					ntohl( srcip ) > mystream->srcIpRange->getLastIp() ) continue; // not in range
					if( ( ntohl( srcip ) - mystream->srcIpRange->getFirstIp() ) % \
					mystream->srcIpRange->getIncrement() != 0 ) continue;  // not in range
				}
			}
			if( mystream->par_dstip )
			{
				if( mystream->dstIpRange == 0 )  // fix address
				{
					if( dstip != mystream->dstIp ) continue;
				}
				else  // address range
				{
					if( ntohl( dstip ) < mystream->dstIpRange->getFirstIp() || \
					ntohl( dstip ) > mystream->dstIpRange->getLastIp() ) continue; // not in range
					if( ( ntohl( dstip ) - mystream->dstIpRange->getFirstIp() ) % \
					mystream->dstIpRange->getIncrement() != 0 ) continue;  // not in range
				}
			}
			if( mystream->par_srcport )
			{
				if( mystream->srcPortRange == 0 )  // fix port
				{
					if( srcport != mystream->srcPort ) continue;
				}
				else  // port range
				{
					if( srcport < mystream->srcPortRange->getFirstPort() || \
					srcport > mystream->srcPortRange->getLastPort() ) continue; // not in range
					if( ( srcport - mystream->srcPortRange->getFirstPort() ) % \
					mystream->srcPortRange->getIncrement() != 0 ) continue;  // not in range
				}			
			}
			if( mystream->par_dstport )
			{
				if( mystream->dstPortRange == 0 )  // fix port
				{
					if( dstport != mystream->dstPort ) continue;
				}
				else  // port range
				{
					if( dstport < mystream->dstPortRange->getFirstPort() || \
					dstport > mystream->dstPortRange->getLastPort() ) continue; // not in range
					if( ( dstport - mystream->dstPortRange->getFirstPort() ) % \
					mystream->dstPortRange->getIncrement() != 0 ) continue;  // not in range
				}			
			}
			if( g_writelog )
			{
				if( !mystream->par_tos ) mystream->tos = tos;
			}
			processPacket( intime, mystream, ( char * )( udph ) + 8, msglen );
		}
	}
}	

void processPacket( struct timeval &intime, struct Stream *mystream, char *buf, int msglen )
{
	int bytes, loss;
	int invalid;
	unsigned long *ubuf;
	unsigned long seqnum, delay;
	struct timeval outtime, delaytime;
	struct tm *date;
	struct Statistics *mystat;
		
	// get the sequence number and timestamp from the packet

	invalid = 0;
	ubuf = ( unsigned long * )buf;
	seqnum = ntohl( ubuf[0] );
	if( mystream->proto == IPPROTO_UDP ) bytes = msglen;
	else bytes = msglen + OVERHEAD;  // add the TCP overhead

	if( mystream->firstPacket )
	{
		mystream->firstTime = intime;  // arrival time of the first packet
	}
	
	// calculate delay
	
	if( mystream->measureDelay || mystream->measureJitter )
	{
		outtime.tv_sec = ntohl( ubuf[1] );
		outtime.tv_usec = ntohl( ubuf[2] );
		delaytime = intime;
		subTime( delaytime, outtime );  // delay = in - out
		delay = delaytime.tv_sec*1000000 + delaytime.tv_usec;  // in us
		if( delaytime.tv_usec < 0 || delaytime.tv_sec < 0 )
		{
			invalid = -1;  // the delay is invalid
		}
	}
	else delay = 0;
	
	// print data to file

	date = localtime( &intime.tv_sec );
	if( g_writelog || mystream->log || ( mystream->logdelay && !invalid && ( delay > mystream->logdelay ) ) )
	{
		g_logfile << setfill( '0' ) << setw( 2 ) << date->tm_year-100 << "/" \
		<< setw( 2 ) << date->tm_mon+1 << "/" << setw( 2 ) << date->tm_mday << "." \
		<< setw( 2 ) << date->tm_hour << ":" << setw( 2 ) << date->tm_min << ":" \
		<< setw( 2 ) << date->tm_sec << "." << setw( 6 ) << intime.tv_usec << setfill( ' ' ) \
		<< resetiosflags( ios::left ) << setw( 10 ) << mystream->id << setw( 8 ) << bytes \
		<< setw( 11 ) << seqnum << setw( 8 );
		if( delay < 1000 ) g_logfile << delay << " u";
		else g_logfile << delay/1000 << " m";
		g_logfile << endl;
	}

	// lock the mutex and update the statistics

	loss = 0;
	g_rwlock.readLock();
	mystat = mystream->actualStat;
	mystat->bytes += bytes;
	if( seqnum == 0 ) mystream->firstPacket = -1;  // ngen has been restarted
	if( !mystream->firstPacket )  // it is not the first packet
	{
		if( mystream->proto == IPPROTO_UDP ) // no loss for TCP
		{
			loss = seqnum - mystream->seqNum - 1;
			if( loss < 0 )  // this is a misordered packet
			{
				mystat->misorderings++;
			}
			else if( loss > 0 )
			{
				mystat->losses +=  loss;
			}
		}
		if( seqnum > mystream->seqNum ) mystream->seqNum = seqnum;
	}
	else
	{
		mystream->seqNum = seqnum;
	}
	if( invalid )
	{
		mystat->invalid = -1;
	}
	else
	{
		if( mystream->measureDelay )
		{
			mystat->delay += delay;
			mystat->addDelay( delay );
		}
		if( mystream->measureJitter )
		{
			if( mystat->packets > 0 )
			{
				if( delay < mystat->minDelay ) mystat->minDelay = delay;
				if( delay > mystat->maxDelay ) mystat->maxDelay = delay;
			}
			else
			{
				mystat->minDelay = mystat->maxDelay = delay;
			}
		}
	}
	mystat->packets++;
	mystream->firstPacket = 0;
	g_rwlock.unlock();
	
	// now log if there was a loss
	
	if( ( loss > 0 ) && ( mystream->logloss ) )
	{
		g_logfile << setfill( '0' ) << setw( 2 ) << date->tm_year-100 << "/" \
		<< setw( 2 ) << date->tm_mon+1 << "/" << setw( 2 ) << date->tm_mday << "." \
		<< setw( 2 ) << date->tm_hour << ":" << setw( 2 ) << date->tm_min << ":" \
		<< setw( 2 ) << date->tm_sec << "." << setw( 6 ) << intime.tv_usec << setfill( ' ' ) \
		<< resetiosflags( ios::left ) \
		<< setw( 10 ) << mystream->id << setw( 8 ) << " LOSS" << setw( 11 ) << seqnum \
		<< setw( 8 ) << loss << endl;
	}
}

void usage()
{
	cerr << "Command line arguments:\n";
	cerr << "-proto udp|tcp\n";
	cerr << "[-if interface] : for UDP streams only]\n";
	cerr << "[-srcip srcip] : for UDP streams only\n";
	cerr << "[-dstip dstip] : for UDP streams only\n";
	cerr << "[-srcport srcport] : for UDP streams only\n";
	cerr << "[-dstport dstport] : for tcp the default is " << DEF_DPORT << "\n";
	cerr << "[-prec precedence]\n";
	cerr << "[-tos tos]\n";
	cerr << "[-delay] : to measure delay\n";
	cerr << "[-jitter] : measure jitter\n";
	cerr << "[-percentile] : measure percentile\n";
	cerr << "[-p1 percentile1] : specify the first percentile\n";
	cerr << "[-p2 percentile1] : specify the second percentile\n";
	cerr << "[-p3 percentile1] : specify the third percentile\n";
	cerr << "[-int statinterval] : '2' - 2 sec, '100m' - 100 ms, the default is : 1 second\n";
	cerr << "[-timestamp : to timestamp each statistics line\n";
	cerr << "[-stat statfile] : to save the statistics\n";
	cerr << "[-log logfile] : to log every packets\n";
	cerr << "-----or----\n-f file\n";
	exit( -1 );
}
