/*************************************************************************

 range.cpp
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "range.h"
#include "utils.h"

void IpRange::init( unsigned long firstip, unsigned long increment, long number )
{
	first = ntohl( firstip );  // convert to host byte order
	inc = ntohl( increment );
	num = number;
	last = first + ( num-1 )*inc;
	actual = first;
}

unsigned long IpRange::getNextIp()
{
	unsigned long ret;
	
	ret = actual;
	actual += inc;
	if( actual > last ) actual = first;
	return htonl( ret );  // convert back to network byte order
}


void PortRange::init( int firstport, int increment, int number )
{
	first = firstport;
	inc = increment;
	num = number;
	last = first + ( num-1 )*inc;
	actual = first;
}

int PortRange::getNextPort()
{
	int ret;
	
	ret = actual;
	actual += inc;
	if( actual > last ) actual = first;
	return ret;  // returns host byte order
}

// reads the IP range parameters from the config file, and puts it into 'range'

void readIpRange( IpRange &range, ifstream &scriptf, char *buf, int buflen )
{
	struct in_addr first, increment;
	long number;
	int par_first, par_inc, par_num;
	
	par_first = par_inc = par_num = 0;
	number = 0;
	if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
	if( strcmp( "{", buf ) ) configError( "{ is expected instead of ", buf );

	// now read the range parameters

	while( -1 )
	{
		if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
		if( !strcmp( "first", buf ) )
		{
			if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
			if( inet_aton( buf, &first ) == 0 ) configError( "Invalid first: ", buf );
			par_first = -1;
		}
		else if( !strcmp( "inc", buf ) )
		{
			if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
			if( inet_aton( buf, &increment ) == 0 ) configError( "Invalid increment: ", buf );
			par_inc = -1;
		}
		else if( !strcmp( "num", buf ) )
		{
			if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
			number = atoi( buf );
			if( number < 1 ) configError( "Invalid number: ", buf );
			par_num = -1;
		}
		else if( !strcmp( "}", buf ) ) break;
		else configError( "Invalid keyword: ", buf );
	}
	if( !par_first ) configError( "first parameter is missing\n" );
	if( !par_inc ) configError( "increment parameter is missing\n" );
	if( !par_num ) configError( "number parameter is missing\n" );
	range.init( first.s_addr, increment.s_addr, number );
}


// reads the port range parameters from the config file, and puts it into 'range'

void readPortRange( PortRange &range, ifstream &scriptf, char *buf, int buflen )
{
	int first, increment, number;
	int par_first, par_inc, par_num;
	
	first = increment = number = 0;
	par_first = par_inc = par_num = 0;
	if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
	if( strcmp( "{", buf ) ) configError( "{ is expected instead of ", buf );

	// now read the range parameters

	while( -1 )
	{
		if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
		if( !strcmp( "first", buf ) )
		{
			if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
			first = atoi( buf );
			if( first < 0 || first > 65535 ) configError( "Invalid first: ", buf );
			first = first;
			par_first = -1;
		}
		else if( !strcmp( "inc", buf ) )
		{
			if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
			increment = atoi( buf );
			if( increment < 0 ) configError( "Invalid increment: ", buf );
			increment = increment;
			par_inc = -1;
		}
		else if( !strcmp( "num", buf ) )
		{
			if( nextWord( scriptf, buf, buflen ) ) configError( "Unexpected end of file" );
			number = atoi( buf );
			if( number < 1 || number > 65535 ) configError( "Invalid number: ", buf );
			par_num = -1;
		}
		else if( !strcmp( "}", buf ) ) break;
		else configError( "Invalid keyword: ", buf );
	}
	if( !par_first ) configError( "first parameter is missing\n" );
	if( !par_inc ) configError( "increment parameter is missing\n" );
	if( !par_num ) configError( "number parameter is missing\n" );
	range.init( first, increment, number );
}
