/*************************************************************************

 rwlock.cpp
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#include <iostream>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "rwlock.h"

using namespace std;

RwLock::RwLock()
{
	pthread_mutex_init( &write_mutex, NULL );
	pthread_mutex_init( &read_mutex, NULL );
	sigemptyset( &sigset );
	sigaddset( &sigset, SIGUSR1 );
	pthread_sigmask ( SIG_BLOCK, &sigset, NULL );
	read_count = 0;
	write_waiting = 0;
}

RwLock::~RwLock()
{
	pthread_mutex_destroy( &write_mutex );
	pthread_mutex_destroy( &read_mutex );
}

void RwLock::readLock()
{
	pthread_mutex_lock( &write_mutex );
	pthread_mutex_lock( &read_mutex );
	pthread_mutex_unlock( &write_mutex );
	read_count++;
	pthread_mutex_unlock( &read_mutex );
}

void RwLock::writeLock()
{
	int n;
	
	pthread_mutex_lock( &write_mutex );
	pthread_mutex_lock( &read_mutex );
	tid = pthread_self();
	if( read_count != 0 )
	{
		write_waiting = -1;
		pthread_mutex_unlock( &read_mutex );
		sigwait( &sigset, &n );  // wait for SIGUSR1 signal
		return;
	}
	pthread_mutex_unlock( &read_mutex );
	return;
}

void RwLock::unlock()
{
	int count;
	
	pthread_mutex_lock( &read_mutex );
	if( pthread_self() == tid )  // it was a write lock
	{
		write_waiting = 0;
		tid = 0;
		pthread_mutex_unlock( &read_mutex );
		pthread_mutex_unlock( &write_mutex );
	}
	else  // it was a read lock
	{
		read_count--;
		count = read_count;
		pthread_mutex_unlock( &read_mutex );
		if( count < 0 )
		{
			cerr << "RwLock: the reader count is negative\n";
			exit( -1 );
		}
		if( ( count == 0 ) && write_waiting )
		{
			pthread_kill( tid, SIGUSR1 );  // wake up the sleepig thread
		}
	}
}
