/*************************************************************************

 rwlock.h
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/
 
#ifndef RWLOCK_H
#define RWLOCK_H 

/*
 This class implements a read-write lock. It works like recursive mutexes:
 it does not check that the thread unlocking the lock actually locked it.
 A thread can lock for read access the lock several times, but it has
 to unlock it exactly the same time. Double write lock results dead lock.
 The class uses the SIGUSR1 signal, so you should not use it in your
 code when using this rwlock.
*/

class RwLock
{
 public:
	RwLock();
	~RwLock();
	void readLock();
	void writeLock();
	void unlock();

 protected:
	sigset_t sigset;
	int write_waiting;  // indicates that a thread is waiting for write access
	pthread_t tid;  // the thread id of the waiting thread
	pthread_mutex_t write_mutex;
	pthread_mutex_t read_mutex;
	int read_count;  // the number of readers who locked the lock
};
 
#endif
                                                                                                   
