/*************************************************************************

 statistics.cpp
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#include <iostream>
#include <stdlib.h>
#include "statistics.h"

using namespace std;

Statistics::Statistics()
{
	packets = bytes = losses = misorderings = minDelay = maxDelay = invalid = 0;
	delay = 0;
	sorted = 0;
}

Statistics::~Statistics()
{
	arr.clear();
}

void Statistics::clear()
{
	arr.clear();
	packets = bytes = losses = misorderings = minDelay = maxDelay = invalid = 0;
	delay = 0;
	sorted = 0;
}

void Statistics::addDelay( unsigned long delay )
{
	arr.push_back( delay );
	sorted = 0;
}

unsigned long Statistics::getPercentile( float p )
{
	unsigned int i;
	
	if( packets <= 1 ) return 0;
	if( !sorted )
	{
		heapSort();
		sorted = -1;
	}
	i = ( int )( p*packets/100 );
	if( i >= packets - 1 )
	{
		return arr[packets-1];
	}
	else
	{
		return( ( arr[i] + arr[i-1] ) / 2 );
	}
}

void Statistics::heapSort()
{
	int i, temp;
	
	for( i = ( packets/2 ) - 1; i >= 0; i-- )
	{
		siftDown( i, packets );
	}
	
	for (i = packets - 1; i >= 1; i-- )
	{
		temp = arr[0];
		arr[0] = arr[i];
		arr[i] = temp;
		siftDown( 0, i-1 );
	}
}


void Statistics::siftDown( int root, int bottom )
{
	int done, maxChild, temp;
	
	done = 0;
	while( ( root*2 <= bottom ) && ( !done ) )
	{
		if( root*2 == bottom )
		{
			maxChild = root * 2;
		}
		else if( arr[root*2] > arr[root*2 + 1] )
		{
			maxChild = root*2;
		}
		else
		{
			maxChild = root*2 + 1;
		}
		if( arr[root] < arr[maxChild] )
		{
			temp = arr[root];
			arr[root] = arr[maxChild];
			arr[maxChild] = temp;
			root = maxChild;
		}
		else
		{
			done = 1;
		}
	}
}
