/*************************************************************************

 utils.cpp
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h> 	
#include <netinet/if_ether.h>
#include <pthread.h>
#include <ctype.h>
#include <limits.h>
#include <sys/ioctl.h>

#include "utils.h"

using namespace std;

const unsigned long ULONG_HALF = ULONG_MAX/2;


// prints out error messages

void configError( char *c1, char *c2 )
{
	if( c1 != NULL ) cerr << endl << c1;
	if( c2 != NULL ) cerr << " " << c2;
	cerr << endl;
	exit( -1 );
}


// reads the next word from the given file, skips the comment preceeded by '#'

int nextWord( ifstream &file, char *buf, int buflen )
{
	char c;
	int i;
	c = file.get();
	while( -1 )
	{
		while( isspace( c ) && !file.eof() )  // to skip the space characters
		{
			c = file.get();
		}
		if( file.eof() )
		{
			return -1;   // cannot read this element because of eof
		}
		if( c != '#' ) break;   // comment -> skip this line
		while( !file.eof() )
		{
			c = file.get();
			if( c  == '\n' ) break;
		}
		if( file.eof() ) return -1;
		c = file.get();
	}
	i = 0;
	buflen--;
	while( !isspace( c ) && !file.eof() && ( i < buflen) )
	{
		buf[i++] = c;
		c = file.get();
	}
	buf[i] = 0;
	return 0;
}


// read time parameter from the given buffer, and returns it in usec

long readTime( char *buf )
{
	char c;
	unsigned long l;
	
	l = atol( buf );
	c = buf[ strlen( buf ) - 1 ];
	if( c == 'u' )  // it is in us
	{
		return l;
	}
	else if( c == 'm' )  // it is in ms
	{
		return l * 1000;
	}
	else if( !isdigit( c ) )
	{
		return -1;
	}
	else
	{
		return l * 1000000;  // it is in sec
	}
}


// read rate parameter from the given buffer, and returns it in bit per sec

long readRate( char *buf )
{
	char c;
	unsigned long l;
	
	l = atol( buf );
	c = buf[ strlen( buf ) - 1 ];
	if( c == 'M' )  // it is in Mbps
	{
		return l * 1000000;
	}
	else if( c == 'k' )  // it is in kbps
	{
		return l * 1000;
	}
	else if( !isdigit( c ) )
	{
		return -1;
	}
	else
	{
		return l;  // it is in bps
	}
}


// gets the ip address for the interface

unsigned long getIfAddr( char *ifname )
{
	char buf[1024];
	int x, sock;
	struct ifreq ifdata;
	struct ifconf ifconfig;
	struct sockaddr_in local;
	unsigned long addr;
	
	addr = 0;
	sock = socket( PF_INET, SOCK_DGRAM, 0 );
	if( sock == -1 )
	{
		perror( "Cannot open UDP socket" );
		return 0;
	}
	ifconfig.ifc_len = 1500;
	ifconfig.ifc_buf = buf;
	if( ioctl( sock, SIOCGIFCONF, &ifconfig, sizeof( ifconfig ) ) )
	{
		perror( "Cannot get the address of the default interface" );
		return 0;
	}	
	x = 0;
	while( -1 )
	{
		if( x*sizeof( struct ifreq ) >= ( unsigned int )ifconfig.ifc_len ) break;
		ifdata = *( ifconfig.ifc_req + x );
		if( !strcmp( ifname, ifdata.ifr_name ) )  // we found it
		{
			memcpy( &local, &ifdata.ifr_addr, sizeof( ifdata.ifr_addr ) );
			addr = local.sin_addr.s_addr;
			break;
		}
		x++;
	}
	return addr;
}


// gets the index of the interface

int getIfIndex( char *ifname, int &index )
{
	int sock;
	struct ifreq ifdata;
	
	sock = socket( PF_INET, SOCK_DGRAM, 0 );
	if( sock == -1 )
	{
		perror( "Cannot open socket" );
		return -1;
	}
	strncpy( ifdata.ifr_name, ifname, IFNAMSIZ );
	ifdata.ifr_name[IFNAMSIZ] = 0;
	if( ioctl( sock, SIOCGIFINDEX, &ifdata, sizeof( ifdata ) ) )
	{
		perror( "Cannot get the interface index" );
		return -1;
	}
	index = ifdata.ifr_ifindex;
	return 0;
}


// gets the promisc mode of the interface

int getPromiscMode( char *ifname, int &mode )
{
	int sock;
	struct ifreq ifdata;
	
	sock = socket( PF_INET, SOCK_DGRAM, 0 );
	if( sock == -1 )
	{
		perror( "Cannot open socket" );
		return -1;
	}
	strncpy( ifdata.ifr_name, ifname, IFNAMSIZ );
	ifdata.ifr_name[IFNAMSIZ] = 0;
	if( ioctl( sock, SIOCGIFFLAGS, &ifdata, sizeof( ifdata ) ) )
	{
		perror( "Cannot get the interface flags" );
		return -1;
	}
	mode = ifdata.ifr_flags & IFF_PROMISC;
	return 0;
}


// sets the promisc mode of the interface

int setPromiscMode( char *ifname, int mode )
{
	int sock;
	struct ifreq ifdata;
	
	sock = socket( PF_INET, SOCK_DGRAM, 0 );
	if( sock == -1 )
	{
		perror( "Cannot open socket" );
		return -1;
	}
	strncpy( ifdata.ifr_name, ifname, IFNAMSIZ );
	ifdata.ifr_name[IFNAMSIZ] = 0;
	if( ioctl( sock, SIOCGIFFLAGS, &ifdata, sizeof( ifdata ) ) )
	{
		perror( "Cannot get the interface flags" );
		return -1;
	}
	if( mode )
	{
		ifdata.ifr_flags |= IFF_PROMISC;
	}
	else
	{
		ifdata.ifr_flags ^= IFF_PROMISC;
	}
	if( ioctl( sock, SIOCSIFFLAGS, &ifdata, sizeof( ifdata ) ) )
	{
		perror( "Cannot set promiscuous mode" );
		return -1;
	}
	return 0;
}


// fills in the static part of the IP header

void buildIpHeader( struct iphdr *iph )
{
	iph->version = 4;
	iph->ihl = 5;
	iph->id = 0;
	iph->frag_off = htonl( 0x4000 );  // don't fragment
	iph->ttl = 64;
}


// modifies the ip header

void modifyIpHeader( struct iphdr *iph, int length, unsigned long source, unsigned long dest, int proto, int tos )
{
	iph->tot_len = htons( length );
	iph->id = 0;
	iph->saddr = source;
	iph->daddr = dest;
	iph->protocol = proto;
	iph->tos = tos;
}



// computes ip checksum

void computeChecksum( struct iphdr *iph )
{
	int sum, count;
	unsigned short *hdr;
	
	hdr = ( unsigned short * )iph;
	iph->check = 0;
	sum = 0;
	count = 20;
	while( count > 1 )
	{
		sum += *hdr++;
		count -= 2;
	}
	if( count > 0 )	sum += *( unsigned char * )hdr;
	sum = ( sum >> 16 ) + ( sum & 0xffff );
	sum += ( sum >> 16 );
	iph->check = ( unsigned short )~sum;
}


// fills in the fields of the udp header

void buildUdpHeader( struct udphdr *udph, int length, unsigned short sport, unsigned short dport )
{
	udph->len = htons( length );
	udph->source = htons( sport );
	udph->dest = htons( dport );
	udph->check = 0;
}


// fills in the fields of the vlan header

void setVlanHeader( struct vlanhdr *vlanh, unsigned int vlanid, unsigned int prio )
{
	vlanh->myunion.id = 0;
	vlanh->myunion.mystruct.cfi = 0;
	vlanh->myunion.mystruct.prio = ( unsigned short )prio;
	vlanh->myunion.id |= htons( ( unsigned short )vlanid );
	vlanh->type = htons( ETH_P_IP );
}


// returns the difference in us between the two provided time value

long timeDiff( struct timeval &tv1, struct timeval &tv2 )
{
	return ( tv1.tv_sec - tv2.tv_sec ) * 1000000 + tv1.tv_usec - tv2.tv_usec;
}

/*
	Calculates the difference between the two tcp sequence numbers.
	A negative number means retransmission for sequence numbers.
	It also considers the overflow of the sequence number.
	It always considers the smaller distance between the numbers,
	e.g. if seqold is 1000, and seqnew is 4000000000, then
	it considers it as retransmission, rather than a new packet.
*/

long seqDiff( unsigned long seq1, unsigned long seq2 )
{
	if( seq1 == seq2 ) return 0;
	if( seq1 > seq2 )
	{
		if( seq1 - seq2 > ULONG_HALF )
		{
			return -( ULONG_MAX + 1 - seq1 + seq2 );
		}
		else
		{
			return seq1 - seq2;
		}
	}
	else
	{
		if( seq2 - seq1 > ULONG_HALF )
		{
			return ULONG_MAX + 1 - seq2 + seq1;
		}
		else
		{
			return seq1 - seq2;
		}
	}
}


// converts the hexadecimal character c into a decimal integer

int hexConverter( char c )
{
	switch( c )
	{
		case '0': return 0;
		case '1': return 1;
		case '2': return 2;
		case '3': return 3;
		case '4': return 4;
		case '5': return 5;
		case '6': return 6;
		case '7': return 7;
		case '8': return 8;
		case '9': return 9;
		case 'a': case 'A' : return 10;
		case 'b': case 'B' : return 11;
		case 'c': case 'C' : return 12;
		case 'd': case 'D' : return 13;
		case 'e': case 'E' : return 14;
		case 'f': case 'F' : return 15;
	}
	return -1;
}
