/*************************************************************************

 utils.h
 Copyright (C) 2006, Norbert Vegh
 Norbert Vegh, ntools@norvegh.com

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.

*************************************************************************/

#ifndef UTILS_H
#define UTILS_H

#include <iostream>
#include <fstream>
#include <string.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

#include "range.h"

using namespace std;

// this file contains some utility functions used by the other programs

struct vlanhdr
{
	union
	{
		struct
		{		
			unsigned short blank:4;
			unsigned short cfi:1;
			unsigned short prio:3;
		} mystruct;
		unsigned short id;
	} myunion;
	unsigned short type;
};

inline ostream &operator<<( ostream &os, timeval & tv );  // to print out time values
inline timeval operator+( timeval &tv1, timeval &tv2 );
inline timeval operator-( timeval &tv1, timeval &tv2 );
inline int operator<( timeval &tv1, timeval &tv2 );
inline int operator>( timeval &tv1, timeval &tv2 );
inline int operator<=( timeval &tv1, timeval &tv2 );
inline int operator>=( timeval &tv1, timeval &tv2 );

void configError( char *c1 = NULL, char *c2 = NULL );  // prints error messages
int nextWord( ifstream &file, char *buf, int buflen );
long readTime( char *buf );
long readRate( char *buf );
unsigned long getIfAddr( char *ifname );
int getIfIndex( char *ifname, int &index );
int getPromiscMode( char *ifname, int &mode );
int setPromiscMode( char *ifname, int mode );
void buildIpHeader( struct iphdr *iph );
void modifyIpHeader( struct iphdr *iph, int length, unsigned long source, unsigned long dest, int proto, int tos );
void setVlanHeader( struct vlanhdr *vlanh, unsigned int vlanid, unsigned int prio );
void computeChecksum( struct iphdr *iph );
void buildUdpHeader( struct udphdr *udph, int length, unsigned short sport, unsigned short dport );
long timeDiff( struct timeval &tv1, struct timeval &tv2 );
inline void subTime( struct timeval &tv1, struct timeval &tv2 );
inline void addTime( struct timeval &tv1, struct timeval &tv2 );
inline void addTime( struct timeval &tv1, unsigned long inc );
int compareSeq( unsigned long seq1, unsigned long seq2 );
long seqDiff( unsigned long seq1, unsigned long seq2 );
int hexConverter( char c );


// for printing out time values

inline ostream &operator<<( ostream &os, timeval & tv )
{
	os << tv.tv_sec << "." << tv.tv_usec;
	return os;
}

// subtracks the second time value from the first, stores the result in the first

inline void subTime( struct timeval &tv1, struct timeval &tv2 )
{
	tv1.tv_sec -= tv2.tv_sec;
	tv1.tv_usec -= tv2.tv_usec;
	if( tv1.tv_usec < 0 )
	{
		tv1.tv_sec--;
		tv1.tv_usec += 1000000;
	}
	else if( tv1.tv_usec >= 1000000 )
	{
		tv1.tv_sec++;
		tv1.tv_usec -= 1000000;
	}
}


// adds up the two time value, stores the result in the first

inline void addTime( struct timeval &tv1, struct timeval &tv2 )
{
	tv1.tv_sec += tv2.tv_sec;
	tv1.tv_usec += tv2.tv_usec;
	if( tv1.tv_usec > 1000000 )
	{
		tv1.tv_usec -= 1000000;
		tv1.tv_sec++;
	}
}


// adds 'inc' usec to the provided time value

inline void addTime( struct timeval &tv1, unsigned long inc )
{
	if( inc < 1000000 )
	{
		tv1.tv_usec += inc;
		if( tv1.tv_usec > 1000000 )
		{
			tv1.tv_usec -= 1000000;
			tv1.tv_sec++;
		}
	}
	else
	{
		tv1.tv_usec += inc;
		tv1.tv_sec += tv1.tv_usec/1000000;
		tv1.tv_usec = tv1.tv_usec%1000000;
	}
}

inline timeval operator+( timeval &tv1, timeval &tv2 )
{
	timeval ret;
	
	ret.tv_sec = tv1.tv_sec + tv2.tv_sec;
	ret.tv_usec = tv1.tv_usec + tv2.tv_usec;
	if( ret.tv_usec > 1000000 )
	{
		ret.tv_usec -= 1000000;
		ret.tv_sec++;
	}
	return ret;
}

inline timeval operator-( timeval &tv1, timeval &tv2 )
{
	timeval ret;
	
	ret.tv_sec = tv1.tv_sec - tv2.tv_sec;
	ret.tv_usec = tv1.tv_usec - tv2.tv_usec;
	if( ret.tv_usec < 0 )
	{
		ret.tv_usec += 1000000;
		ret.tv_sec--;
	}
	return ret;
}

inline int operator<( timeval &tv1, timeval &tv2 )
{
	if( tv1.tv_sec < tv2.tv_sec ) return -1;
	if( ( tv1.tv_sec == tv2.tv_sec ) && ( tv1.tv_usec < tv2.tv_usec ) ) return -1;
	return 0;
}

inline int operator<=( timeval &tv1, timeval &tv2 )
{
	if( tv1.tv_sec < tv2.tv_sec ) return -1;
	if( ( tv1.tv_sec == tv2.tv_sec ) && ( tv1.tv_usec <= tv2.tv_usec ) ) return -1;
	return 0;
}

inline int operator>( timeval &tv1, timeval &tv2 )
{
	if( tv1.tv_sec > tv2.tv_sec ) return -1;
	if( ( tv1.tv_sec == tv2.tv_sec ) && ( tv1.tv_usec > tv2.tv_usec ) ) return -1;
	return 0;
}

inline int operator>=( timeval &tv1, timeval &tv2 )
{
	if( tv1.tv_sec > tv2.tv_sec ) return -1;
	if( ( tv1.tv_sec == tv2.tv_sec ) && ( tv1.tv_usec >= tv2.tv_usec ) ) return -1;
	return 0;
}


#endif
