/*
 This file is part of pathload.

 pathload is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 pathload is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with pathload; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*-------------------------------------------------
   pathload : an end-to-end available bandwidth 
            estimation tool
   Author   : Manish Jain ( jain@cc.gatech.edu )
              Constantinos Dovrolis (dovrolis@cc.gatech.edu )
   Release  : Ver 1.3.2
   Support  : This work was supported by the SciDAC
            program of the US department 
--------------------------------------------------*/

/*
 * $Header: /net/cvs/bwtest/pathload/pathload_rcv.c,v 1.139 2006/05/19 22:30:13 jain Exp $
 */

#define LOCAL
#include "pathload_gbls.h"
#include "pathload_rcv.h"

/* modification not protected by MGRP defines */
int overhead = 20 + 8;	// IP + UDP

int pathload_probe_bytes = 0;
int pathload_probe_pkts = 0;

#ifdef MGRP
int adjust = DEFAULT_PATHLOAD_ADJUST_TYPE;
int tcp_port = TCPSND_PORT;
int udp_port = UDPRCV_PORT;
int use_kernel_probes = 0;
// only matters when using mgrp; communicated by sender
int allow_piggyback = 0;
int send_estimator_info = 0;
int use_estimator_rpc = 0;
int estimator_rpc_id = 0;
struct est_estimator estimator = { 0 };

char local_hostid[EST_INFO_IDLEN] = { 0 };
char remote_hostid[EST_INFO_IDLEN] = { 0 };

int faster_pathload = 0;
int debug_on = 0;

// to be used to adjust estimates because of piggybacking
long long last_stream_adj_bytes = 0;
long long last_stream_adj_duration_nsec = 0;
long long last_fleet_adj_bytes = 0;
long long last_fleet_adj_duration_nsec = 0;
#endif

// XXX hack so that we can send/recv the hostid as a ctr_code
void
send_hostid(char *hostid, int len)
{
    char ctr_buff[8];
    union hostid_hack hack;

    // truncates hostid to 4 bytes
    strncpy(hack.hostid_str, hostid, sizeof(hack.hostid_str));
    hack.hostid_str[sizeof(hack.hostid_str)-1] = 0;
    send_ctr_mesg(ctr_buff, hack.hostid_code);

    LOG(">>> sent hostid %s (original %s)", hack.hostid_str, hostid);
}

void
recv_hostid(int sock, char *hostid, int len)
{
    char ctr_buff[8];
    union hostid_hack hack;

    while ((hack.hostid_code = recv_ctr_mesg(sock, ctr_buff)) == -1);

    strncpy(hostid, hack.hostid_str, len);
    hostid[len-1] = 0;

    LOG(">>> received hostid %s (stored %s)", hack.hostid_str, hostid);
}

// 0,1 on sucess (indicates allow_piggyback flag)
// aborts on error
static inline int
recv_allow_piggyback(int sock)
{
        int rc = 0;
        int flag = 0;
        char ctr_buff[8];

        while ((rc = recv_ctr_mesg(sock, ctr_buff)) == -1);

        if ((((rc & CTR_CODE) >> 31) == 1) &&
            ((rc&0x7fffffff) == PIGGYBACK_ALLOWED)) {
                flag = 1;
                printf(">>> piggyback ALLOWED\n");

        } else if ((((rc & CTR_CODE) >> 31) == 1) &&
                   ((rc&0x7fffffff) == PIGGYBACK_NOTUSED)) {
                flag = 0;
                printf(">>> piggyback NOT allowed\n");

        } else {
                printf("pathload_snd did not send us allow_piggyback mode\n");
                close(sock);
                abort_failed();
        }

        return flag;
}


int main(l_int32 argc, char *argv[])
{
  extern char *optarg;
  struct hostent *host_snd;
  struct sockaddr_in snd_tcp_addr, rcv_udp_addr;
  struct utsname uts ;
  l_int32 ctr_code;
  l_int32 trend, prev_trend = 0;
  l_int32 opt_len, rcv_buff_sz, mss;
  l_int32 ret_val ; 
  l_int32 errflg=0;
  l_int32 file=0;
  char netlogfile[50],filename[50];
  char ctr_buff[8], myname[50], buff[26];
  char mode[4];
  l_int32 c ;
  struct itimerval expireat ;
  struct sigaction sigstruct ;
#ifdef MGRP
  int found = 0;
#endif

  slow=0;
  requested_delay = 0;
  interrupt_coalescence=0;
  bad_fleet_cs=0;
  num_stream = NUM_STREAM;
  stream_len = STREAM_LEN ;
  exp_flag = 1;
  num=0;
  snd_time_interval=0; 

  converged_gmx_rmx = 0 ;
  converged_gmn_rmn = 0 ;
  converged_rmn_rmx = 0 ;
  counter = 0 ;
  prev_actual_rate = 0;
  prev_req_rate = 0 ; 
  cur_actual_rate = 0 ;
  cur_req_rate = 0 ;
  gettimeofday(&exp_start_time, NULL);
  verbose=1;
  bw_resol=0;
  netlog=0;
  increase_stream_len=0;
  lower_bound=0;

  if ( argc == 1 ) errflg++ ;
#ifdef MGRP
  while ((c = getopt(argc, argv, "a:t:s:hw:vHqo:O:N:VxT:U:P:I:Fd")) != EOF)
#else
  while ((c = getopt(argc, argv, "t:s:hw:vHqo:O:N:V")) != EOF)
#endif
    switch (c) 
    {
      case 't':
        requested_delay = atoi(optarg);
        break;
      case 's':
        strcpy(hostname,optarg);
        break;
      case 'w':
        bw_resol = atof(optarg);
        break;
      case 'q':
        Verbose=0;
        verbose=0;
        break;
      case 'v':
        Verbose=1;
        break;
      case 'O':
        file=1;
        strcpy(filename,optarg);
        strcpy(mode,"a");
        break;
      case 'o':
        file=1;
        strcpy(filename,optarg);
        strcpy(mode,"w");
        break;
      case 'H':
      case 'h':
        help() ;
        break ;
      case 'N':
        netlog=1; 
        strcpy(netlogfile,optarg);
        break;
      case 'V':
        VVerbose = 1;
        break ;
#ifdef MGRP
      case 'T':
	found = sscanf(optarg, "%u", &tcp_port);
	if (found != 1) {
		fprintf(stderr, "invalid TCP send port '%s'", optarg);
		abort_failed();
	}
        break ;
      case 'U':
	found = sscanf(optarg, "%u", &udp_port);
	if (found != 1) {
		fprintf(stderr, "invalid UDP recv port '%s'", optarg);
		abort_failed();
	}
        break ;
      case 'a':
        adjust = pathload_get_adjust_type(optarg);
        break;
      case 'x':
        use_kernel_probes = 1;
        break ;

      // Estimator MODES
      //
      // no -P, -I, -F               default, no interaction with estimators
      //
      // -P<estimator> -I<rpc-ID>    Use estimator RPC to send results.
      //                             This mode is used when the estimator
      //                             initiates the pathload session.
      //
      //                             For example, when avbw2.c connects to
      //                             the estimator to request probing.
      //
      // -P<estimator> -F            Send INFO/availbw to the estimator with
      //                             the results. Used when pathload is invoked
      //                             directly (not through the estimator).
      //
      //                             This mode applies only to the receiver.

      // read an estimator address
      case 'P':
        est_parse_estimator(optarg, &estimator);
        break;
      // estimator_rpc
      case 'I':
        if (!estimator.valid) {
                fprintf(stderr, "error: option -P must precede -I\n");
                abort_failed();
        }
        found = sscanf(optarg, "%u", &estimator_rpc_id);
        if (found != 1) {
		fprintf(stderr, "error: invalid RPC id: '%s'\n", optarg);
		abort_failed();
        }
        use_estimator_rpc = 1;
        break;
      // estimator_info
      case 'F':
        if (!estimator.valid) {
                fprintf(stderr, "error: option -P must precede -F\n");
                abort_failed();
        }
        send_estimator_info = 1;
        break;
        
      case 'd':
        debug_on = 1;
        break;
#endif

      case '?':
        errflg++;
        break ;
    }
  if (errflg)
  {
    fprintf(stderr, "usage: pathload_rcv [-q|-v] [-o <filename>] [-N <filename>]\
[-w <bw_resol>] [-h|-H] -s <sender>\n");
    exit (0);
  }

#ifdef MGRP
  fprintf(stderr, "++++++++ starting pathload_rcv, rpcID %d\n", estimator_rpc_id);
  pathload_estimator_init();
#endif

  if (netlog)
    netlog_fp = fopen(netlogfile,"a");
  if (file)
    pathload_fp = fopen(filename,mode);
  else
    pathload_fp = fopen("pathload.log" , "a" ) ;
  fprintf(pathload_fp, "\n\n");

  if ((host_snd = gethostbyname(hostname)) == 0)
  {
    /* check if the user gave ipaddr */
    if ( ( snd_tcp_addr.sin_addr.s_addr = inet_addr(hostname) ) == -1 )
    {
      fprintf(stderr,"%s: unknown host\n", hostname);
      abort_failed();
    }
  }
  strncpy(buff, ctime(&(exp_start_time.tv_sec)), 24);
  buff[24] = '\0';
  bzero(myname,50);
  if ( gethostname(myname ,50 ) == -1 )
  {
    if ( uname(&uts) < 0 )
      strcpy(myname , "UNKNOWN") ;
    else
      strcpy(myname , uts.nodename) ;
  }
  if (verbose || Verbose)
    printf("\n\nReceiver %s starts measurements at sender %s on %s \n", myname , hostname, buff);
  fprintf(pathload_fp,"\n\nReceiver %s starts measurements at sender %s on %s \n", myname , hostname, buff);

#ifdef MGRP
  if (use_kernel_probes) {
	  printf("receiver: opening a MGRP socket explicitly\n");
	  if ((sock_udp = socket(AF_INET, SOCK_DGRAM, IPPROTO_MGRP)) < 0)
	  {
	    perror("ERROR :: failed to open DGRAM/MGRP socket:");
	    abort_failed();
	  }
  } else
#endif
  if ((sock_udp = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    perror("ERROR :: failed to open DGRAM socket:");
    abort_failed();
  }
  bzero((char *)&rcv_udp_addr, sizeof(rcv_udp_addr));
  rcv_udp_addr.sin_family = AF_INET;
  rcv_udp_addr.sin_addr.s_addr = INADDR_ANY;
#ifdef MGRP
  rcv_udp_addr.sin_port = htons(get_udp_port());
  overhead = get_udp_ip_overhead(sock_udp);
  if (udpex_socket(sock_udp)) {
	printf("** MGRP detected\n");
	if (use_kernel_probes) {
		printf("** Option -x asks for kernel probes: turn on timestamps\n");
		int on = 1;
		int rc = setsockopt(sock_udp, SOL_MGRP, UDPEX_TSTAMP,
					&on, sizeof(on));
		if (rc) {
			perror("turning on bhtp timestamps");
			abort_failed();
		}
                printf("** Adjusting estimates based on policy: %s\n",
                       pathload_get_adjust_policy(adjust));
	}
  } else if (use_kernel_probes) {
  	printf("Asking for -x kernel probes but socket is not UDPex socket\n");
	abort_failed();
  }
#else
  rcv_udp_addr.sin_port = htons(UDPRCV_PORT);
#endif
  if (bind(sock_udp, (struct sockaddr *)&rcv_udp_addr, sizeof(rcv_udp_addr)) < 0)
  {
    perror("ERROR :: failed to bind DGRAM socket:");
    abort_failed();
  }
  rcv_buff_sz = UDP_BUFFER_SZ;
  if (setsockopt(sock_udp, SOL_SOCKET, SO_RCVBUF, &rcv_buff_sz, sizeof(rcv_buff_sz)) < 0)
  {
    rcv_buff_sz/=2;
    if (setsockopt(sock_udp, SOL_SOCKET, SO_RCVBUF, &rcv_buff_sz, sizeof(rcv_buff_sz)) < 0)
    {
      printf("ERROR :: Unable to set socket buffer to %d .\n",rcv_buff_sz);
      abort_failed();
    }
  }
  opt_len = 1;
  if (setsockopt(sock_udp, SOL_SOCKET, SO_REUSEADDR, &opt_len, sizeof(opt_len)) < 0)
  {
    perror("setsockopt(sock_udp,SOL_SOCKET,SO_REUSEADDR):");
    abort_failed();
  }

  /* set up control channel */
  if ((sock_tcp = socket(AF_INET, SOCK_STREAM, 0)) < 0)
  {
    perror("socket(AF_INET, SOCK_STREAM):");
    abort_failed();
  }
  opt_len = 1;
  if (setsockopt(sock_tcp, SOL_SOCKET, SO_REUSEADDR, &opt_len, sizeof(opt_len)) < 0)
  {
    perror("setsockopt(sock_tcp,SOL_SOCKET,SO_REUSEADDR):");
    abort_failed();
  }
  bzero((char *)&snd_tcp_addr, sizeof(snd_tcp_addr));
  snd_tcp_addr.sin_family = AF_INET;
  memcpy((void *)&(snd_tcp_addr.sin_addr.s_addr), host_snd->h_addr, host_snd->h_length);
#ifdef MGRP
  snd_tcp_addr.sin_port = htons(get_tcp_port());
#else
  snd_tcp_addr.sin_port = htons(TCPSND_PORT);
#endif
  if (connect(sock_tcp, (struct sockaddr *)&snd_tcp_addr, sizeof(snd_tcp_addr)) < 0)
  {
    perror("Make sure that pathload_snd runs at sender:");
    abort_failed();
  }
  if (fcntl(sock_tcp, F_SETFL, O_NONBLOCK) < 0)
  {
    perror("fcntl:");
    abort_failed();
  }
  /*
    measure max_pkt_sz (based on TCP MSS).
    this is not accurate because it does not take into
    account MTU of intermediate routers.
  */
  opt_len = sizeof(mss);
  if (getsockopt(sock_tcp, IPPROTO_TCP, TCP_MAXSEG, (char*)&mss,(socklen_t *) &opt_len)<0)
  {
     perror("getsockopt(sock_tcp,IPPROTO_TCP,TCP_MAXSEG):");
     abort_failed();
  }
  rcv_max_pkt_sz = mss;
  if (rcv_max_pkt_sz == 0 || rcv_max_pkt_sz == 1448 ) 
    rcv_max_pkt_sz = 1472;   /* Make it Ethernet sized MTU */
  else
    rcv_max_pkt_sz = mss+12;

#ifdef MGRP
#ifdef DEBUG
    printf("DEBUG :: restricting original max_pkt_sz to %d from %d\n",
  	  rcv_max_pkt_sz, MGRP_MAX_TUN_UDP);
#endif
    /* Restrict the maximum packet size to leave space for bhtp overhead. */
    rcv_max_pkt_sz = MGRP_MAX_TUN_UDP;
#endif

  sigstruct.sa_handler = sig_alrm ;
  sigemptyset(&sigstruct.sa_mask);
  sigstruct.sa_flags = 0 ;
  #ifdef SA_INTERRUPT
    sigstruct.sa_flags |= SA_INTERRUPT ;
  #endif
  sigaction(SIGALRM , &sigstruct,NULL );
  expireat.it_value.tv_sec = 60 ; /* RECEIVER ABORTS TIME */
  expireat.it_value.tv_usec = 0 ;
  expireat.it_interval.tv_sec = 0 ; 
  expireat.it_interval.tv_usec = 0 ;
  setitimer(ITIMER_REAL, &expireat,NULL);

  /* exchange hostids */
  recv_hostid(sock_tcp, remote_hostid, sizeof(remote_hostid));
  send_hostid(local_hostid, sizeof(local_hostid));

  /* receive allow_piggyback mode */
  allow_piggyback = recv_allow_piggyback(sock_tcp);

  /* receive sender max_pkt_sz */
  while ((snd_max_pkt_sz = recv_ctr_mesg(sock_tcp, ctr_buff)) == -1);
  if ( snd_max_pkt_sz == -2 )
  {
    printf("pathload_snd did not respond for 60 sec\n");
    close(sock_tcp);
    abort_failed();
  }
  expireat.it_value.tv_sec = 0 ; 
  expireat.it_value.tv_usec = 0 ;
  expireat.it_interval.tv_sec = 0 ; 
  expireat.it_interval.tv_usec = 0 ;
  setitimer(ITIMER_REAL, &expireat,NULL);
  
  sigstruct.sa_handler = SIG_DFL ;
  sigemptyset(&sigstruct.sa_mask);
  sigstruct.sa_flags = 0 ;
  sigaction(SIGALRM , &sigstruct,NULL );


  send_ctr_mesg(ctr_buff, rcv_max_pkt_sz);
  max_pkt_sz = (rcv_max_pkt_sz < snd_max_pkt_sz) ? rcv_max_pkt_sz:snd_max_pkt_sz ;
  if (Verbose)
    printf("  Maximum packet size          :: %ld bytes\n",max_pkt_sz);
  fprintf(pathload_fp,"  Maximum packet size          :: %ld bytes\n",max_pkt_sz);
  rcv_latency = (l_int32) recvfrom_latency(rcv_udp_addr);
  while ((snd_latency = recv_ctr_mesg(sock_tcp, ctr_buff)) == -1);
  if (Verbose)
  {
    printf("  send latency @sndr           :: %ld usec\n",snd_latency);
    printf("  recv latency @rcvr           :: %ld usec\n",rcv_latency);
  }
  fprintf(pathload_fp,"  send latency @sndr           :: %ld usec\n",snd_latency);
  fprintf(pathload_fp,"  recv latency @rcvr           :: %ld usec\n",rcv_latency);
  min_time_interval=
      SCALE_FACTOR*((rcv_latency>snd_latency)?rcv_latency:snd_latency) ;
  min_time_interval = min_time_interval>MIN_TIME_INTERVAL?
      min_time_interval:MIN_TIME_INTERVAL;
  if (Verbose)
    printf("  Minimum packet spacing       :: %ld usec\n",min_time_interval );
  fprintf(pathload_fp,"  Minimum packet spacing       :: %ld usec\n",min_time_interval );
  max_rate = (max_pkt_sz+overhead) * 8. / min_time_interval ;
  min_rate = (MIN_PKT_SZ+overhead) * 8./ MAX_TIME_INTERVAL ;
  if(Verbose)
    printf("  Max rate(max_pktsz/min_time) :: %.2fMbps\n",max_rate);
  fprintf(pathload_fp,"  Max rate(max_pktsz/min_time) :: %.2fMbps\n",max_rate);

  /* Estimate ADR */
  adr = get_adr() ;
  if ( bw_resol == 0 && adr != 0 )
    bw_resol = .02*adr ;
  else if (bw_resol == 0 )
    bw_resol = 2 ; 
  if(Verbose)
    printf("  Grey bandwidth resolution    :: %.2f\n",grey_bw_resolution());
  fprintf(pathload_fp,"  Grey bandwidth resolution    :: %.2f\n",grey_bw_resolution());
  
  if (interrupt_coalescence)
  {
    bw_resol = .05*adr;
    if(verbose||Verbose)
      printf("  Interrupt coalescion detected\n");
    fprintf(pathload_fp,"  Interrupt coalescion detected\n");
  }
  
  if ( adr == 0 || adr > max_rate || adr < min_rate)
    tr = (max_rate+min_rate)/2.;
  else 
    tr = adr ;

#ifdef MGRP
	printf("rates max %f min %f tr %f\n", max_rate, min_rate, tr);
#endif
  /* Estimate the available bandwidth.*/
  transmission_rate = (l_uint32)rint(1000000 * tr);
  max_rate_flag = 0 ;
  min_rate_flag = 0 ; 
  fflush(pathload_fp);

  sigemptyset(&sigstruct.sa_mask);
  sigstruct.sa_handler = sig_alrm ;
  sigstruct.sa_flags = 0 ;
  #ifdef SA_INTERRUPT
    sigstruct.sa_flags |= SA_INTERRUPT ;
  #endif
  sigaction(SIGALRM , &sigstruct,NULL );

  if ( requested_delay )
  {
    expireat.it_value.tv_sec = requested_delay ; /* RECEIVER ABORTS TIME */
    expireat.it_value.tv_usec = 0 ;
    expireat.it_interval.tv_sec = 0 ; 
    expireat.it_interval.tv_usec = 0 ;
    setitimer(ITIMER_REAL, &expireat,NULL);
  }


  while (1)
  {
    DEBUGTS("%s:%d: %s", __FILE__, __LINE__, __FUNCTION__);
    if ( calc_param() == -1 )
    {
      ctr_code = TERMINATE | CTR_CODE;
      send_ctr_mesg(ctr_buff, ctr_code);
      terminate_gracefully(exp_start_time) ;
    }
    DEBUGTS("%s:%d: %s", __FILE__, __LINE__, __FUNCTION__);
    send_ctr_mesg(ctr_buff, transmission_rate);
    send_ctr_mesg(ctr_buff,cur_pkt_sz) ;
    if ( increase_stream_len )
      stream_len=3*STREAM_LEN;
    else
      stream_len = STREAM_LEN;
    send_ctr_mesg(ctr_buff,stream_len);
    send_ctr_mesg(ctr_buff,time_interval);
    ctr_code = SEND_FLEET | CTR_CODE ;
    send_ctr_mesg(ctr_buff, ctr_code);

    DEBUGTS("%s:%d: %s", __FILE__, __LINE__, __FUNCTION__);
    while (1)
    {
      ret_val = recv_ctr_mesg(sock_tcp, ctr_buff);
      if ((((ret_val & CTR_CODE) >> 31) == 1) &&    
           ((ret_val & 0x7fffffff) == RECV_FLEET )) 
        break ;
      else if ( (((ret_val & CTR_CODE) >> 31) == 1) &&    
                 ((ret_val & 0x7fffffff) == FINISHED_STREAM )) 
        ret_val = recv_ctr_mesg(sock_tcp, ctr_buff);
    }

    DEBUGTS("%s:%d: %s", __FILE__, __LINE__, __FUNCTION__);
    if (recv_fleet() == -1)
    {
      if ( !increase_stream_len )
      {
        trend = INCREASING;
        if ( exp_flag == 1 && prev_trend != 0 && prev_trend != trend)
          exp_flag = 0;
        prev_trend = trend;
        if (rate_adjustment(INCREASING) == -1)
          terminate_gracefully(exp_start_time);
      }
    }
    else
    {
      DEBUGTS("%s:%d: %s", __FILE__, __LINE__, __FUNCTION__);
      get_sending_rate() ;
      trend = aggregate_trend_result();
#ifdef MGRP_DEBUG
	printf("\n**** trend %d bad_fleet_cs %d retry_fleet_cnd_cs %d\n",
		trend, bad_fleet_cs, retry_fleet_cnt_cs);
#endif
      
      if ( trend == -1 && bad_fleet_cs && retry_fleet_cnt_cs >NUM_RETRY_CS )
        terminate_gracefully(exp_start_time) ;
      else if(( trend == -1 && bad_fleet_cs && retry_fleet_cnt_cs <= NUM_RETRY_CS )) /* repeat fleet with current rate. */
        continue ;

      if (trend != GREY)
      {
        if (exp_flag == 1 && prev_trend != 0 && prev_trend != trend)
          exp_flag = 0;
        prev_trend = trend;
      }

      if (rate_adjustment(trend) == -1)
        terminate_gracefully(exp_start_time);
    }
    fflush(pathload_fp);
    DEBUGTS("%s:%d: %s", __FILE__, __LINE__, __FUNCTION__);
  }

#ifdef MGRP
  pathload_estimator_done();
#endif
}
