#ifndef tcpdump_print_bhtp_h
#define tcpdump_print_bhtp_h

#include "mgrp/bhtp.h"

#define BHTP_MAX_SOURCES	100

struct bhtp_source_stats {
	__u32 addr;
	__u64 sndstamp_last;
	struct timeval tspkt_last;
};

struct bhtp_state {
	int sources;
	struct bhtp_source_stats stats[BHTP_MAX_SOURCES];
	struct timeval tspkt;
	struct {
		int active_riders;
		int syn;
		int synack;
		int fin;
		int rst;
	} tcp;
	struct {
		int tun;
		int tcp;
		int tstamp;
		int compact;
		int stats;
		int stun;
                int tsniff;
	} op;
};

void bhtp_print(const u_char *, u_int, const u_char *, int);
void pktrain_print(const u_char *);

extern struct bhtp_state bhtp;

/*
 * USAGE
 *
 *	tcpdump -P tun		Do full protocol parsing for TUNs
 *				Default: off, shows [tcp], [udp]
 *	
 *	tcpdump -P tcp		Do protocol parsing for TCP TUNs.
 *				Useful when reading tcpdumps of piggybacked
 *				packets. First TUN is usually the UDP prober.
 *	
 *	tcpdump -P tstamp	Show the absolute send and recv timestamps
 *				Default: off, show only relative timestamps
 *
 *	tcpdump -P compact	Show compact representation of TUNs
 *				and show reuse stats for probes.
 *	
 *	tcpdump -P stats	Show only reuse stats for probes, no TUNs.
 *
 *	tcpdump -P stun		List briefly all TUNs in BHTP packets
 *
 *	tcpdump -P tsniff	Sniff out TCP packets that pretend to be BHTP.
 *                              (due to an MGRP bug)
 *
 * HOWTO: to add an option
 *
 *	- Add a new variable to bhtp_state.op.XXX
 *	- Parse the new option in bhtp_parse_option()
 *	- Add USAGE information in bhtp_print_usage()
 */

static inline int
bhtp_parse_option(const char *option)
{
	int ok = 1;

	if (strcasecmp(option, "tun") == 0)
		bhtp.op.tun = 1;

	else if (strcasecmp(option, "tcp") == 0)
		bhtp.op.tcp = 1;

	else if (strcasecmp(option, "tstamp") == 0)
		bhtp.op.tstamp = 1;

	else if (strcasecmp(option, "compact") == 0)
		bhtp.op.compact = 1;

	else if (strcasecmp(option, "stats") == 0)
		bhtp.op.stats = 1;

	else if (strcasecmp(option, "stun") == 0)
		bhtp.op.stun = 1;

	else if (strcasecmp(option, "tsniff") == 0)
		bhtp.op.tsniff = 1;

	else
		ok = 0;

	return ok;
}

static inline int
bhtp_print_usage(FILE *fp)
{
	(void)fprintf(stderr,
"\t\t[ -P bhtp-option ]\n");
	(void)fprintf(stderr,
"\t\t\ttun\tDo full protocol parsing for TUNs\n");
	(void)fprintf(stderr,
"\t\t\ttcp\tDo protocol parsing only for TCP TUNs\n");
	(void)fprintf(stderr,
"\t\t\ttstamp\tShow absolute send and recv timestamps\n");
	(void)fprintf(stderr,
"\t\t\tcompact\tShow compact representation of TUNs\n");
	(void)fprintf(stderr,
"\t\t\tstats\tShow only reuse stats for probes, no TUNs\n");
	(void)fprintf(stderr,
"\t\t\tstun\tList briefly all TUNs in BHTP packets\n");
	(void)fprintf(stderr,
"\t\t\ttsniff\tSniff out TCP packets that pretend to be BHTP (len==20)\n");
}

#endif
